/**********************************************************************
 Freeciv - Copyright (C) 1996 - A Kjeldberg, L Gregersen, P Unold
 civpmap.c - Copyright (C) 2002 - R. Wetmore

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
***********************************************************************/

#define FC__CIVMAP_C

/**
 * civmap.c - core map operations
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "log.h"
#include "capability.h"
#include "registry.h"

#include "civmap_main.h"
#include "civtile.h"
#include "civmap.h"
#include "civmapgen.h"


/**
 * Remove working tiles from a civmap and clear its in memory reference.
 */
void civmap_tiles_free(CIVMAP *map)
{
  civtile_free((CIVTILE **)&map->tiles, map->xsize * map->ysize);
}

/**
 * (Re)initialize working tiles of a civmap in memory
 *
 * @return  reference the map object
 */
CIVMAP *civmap_tiles_init(CIVMAP *map)
{
  map->tiles = civtile_init((CIVTILE *)map->tiles, map->xsize*map->ysize, map);
  return map;
}

/**
 * Remove a civmap header structure from memory and clear its reference.
 */
void civmap_free(CIVMAP **map)
{
  if (NULL != *map) {
    civmap_tiles_free(*map);
    free(*map);
    *map = NULL;
  }
}

/**
 * (Re)initialize a civmap header structure in memory
 *
 * @return  reference to the new map object
 */
CIVMAP *civmap_init(CIVMAP *map)
{
  if (NULL != map)
    civmap_tiles_free(map);
  else
    map = fc_malloc(sizeof(CIVMAP));

  *map = CIVMAP_DEFAULT;
  return map;
}


/**
 * Merge a backing store map section into the civmap header structure 
 *
 * Note: loosely synced with server/savegame.c
 *       tile data is not loaded here, use civtile_load().
 */
int civmap_load(CIVMAP *map, struct section_file *fp)
{
  int  nn, val, val2;
  int  game_version, game_nplayers;
  char *savefile_options;
  enum server_states server_state;

  /* Key control options */
  logdebug_suppress_warning;
  game_version = secfile_lookup_int_default(fp, 0, "game.version");
  game_nplayers = secfile_lookup_int_default(fp, 0, "game.nplayers");
  savefile_options = secfile_lookup_str_default(fp, "", "savefile.options");
  server_state = (enum server_states)
    secfile_lookup_int_default(fp, PRE_GAME_STATE, "game.server_state");

  if (0 <= (val = secfile_lookup_int_default(fp, -1, "map.generator")))
    map->generator = val;
  if (0 != (val = secfile_lookup_int_default(fp, 0, "map.seed")))
    map->seed = val;
  if (0 <= (val = secfile_lookup_int_default(fp, -1, "map.maptype")))
    map->type = val;
  if (0 < (val = secfile_lookup_int_default(fp, 0, "map.landpercent")))
    map->landpercent = val;
  if (0 <= (val = secfile_lookup_int_default(fp, -1, "map.riches")))
    map->riches = val;
  if (0 <= (val = secfile_lookup_int_default(fp, -1, "map.huts")))
    map->huts = val;

  /* HACK ALERT: hardcoded ratios to handle the unwashed terrrain types */
  if (0 <= (val = secfile_lookup_int_default(fp, -1, "map.mountains")))
    map->MOUNTAINS = (val+2)/3, map->HILLS = val - map->MOUNTAINS;
  if (0 <= (val = secfile_lookup_int_default(fp, -1, "map.deserts")))
    map->DESERTS = val;
  if (0 <= (val = secfile_lookup_int_default(fp, -1, "map.forestsize")))
    map->JUNGLES = (val+3)/4, map->FORESTS = val - map->JUNGLES;
  if (0 <= (val = secfile_lookup_int_default(fp, -1, "map.swampsize")))
    map->OCEANS = (val+4)/5, map->SWAMPS = val - map->OCEANS;
  if (0 <= (val = secfile_lookup_int_default(fp, -1, "map.tundrasize")))
    map->ARCTIC = (val+2)/3, map->TUNDRA = val - map->ARCTIC;
  if (0 <= (val = secfile_lookup_int_default(fp, -1, "map.grasssize")))
    map->PLAINS = (val+1)/2, map->GRASSLAND = val - map->PLAINS;

  if (secfile_lookup_bool_default(fp, FALSE, "map.is_earth"))
    map->is_earth = TRUE;
  if (secfile_lookup_bool_default(fp, FALSE, "map.have_huts"))
    map->has_huts = TRUE;

  if (secfile_lookup_bool_default(fp, FALSE, "map.fixed_start_positions"))
    map->fixed_start_positions = TRUE;

  for( nn = 0;
       0 <= (val = secfile_lookup_int_default(fp, -1, "map.r%dsx", nn))
    && 0 <= (val2 = secfile_lookup_int_default(fp, -1, "map.r%dsy", nn));
       nn++ ) {
    map->start_positions[nn].x = val;
    map->start_positions[nn].y = val2;
  }

  /* map->num_start_positions is the "desired" number on load/gen */
  if (nn > 0 && nn < game_nplayers)
    nn = game_nplayers;
  map->num_start_positions = nn;

  {
    char *key1 = "map.width", *key2 = "map.height";

    if (!has_capability("startoptions", savefile_options)) {
      /* old versions saved with these names in PRE_GAME_STATE: */
      key1 = "map.xsize", key2 = "map.ysize";
    }

    if (0 < (val = secfile_lookup_int_default(fp, -1, key1)))
      map->xsize = val;
    if (0 < (val = secfile_lookup_int_default(fp, -1, key2)))
      map->ysize = val;
  }

  /* --- End of CIVMAP Header data load/merge --- */

  /* Tile data must be loaded explicitly. Usually it isn't needed. */

  return 0;
}

/**
 * Dump the civmap structure into a backing store map section 
 *
 * Note: loosely synced with server/savegame.c
 */
int civmap_save(const CIVMAP *map, struct section_file *fp)
{
  char options[512];
  int  nn;

  /* 
   * Map generation always leaves us in a pre-run game state.
   */
  secfile_insert_int(fp, (int) SELECT_RACES_STATE, "game.server_state");
  secfile_insert_bool(fp, 1, "game.is_new_game");
  
  /* Update the save_options */
  strncpy(options, SAVEFILE_OPTIONS, sizeof(options));
  if (map->num_start_positions > 0) {
    strncat(options, " startpos", sizeof(options));
  }
  if (map->has_specials) {
    strncat(options, " specials", sizeof(options));
  } else
  if (map->has_rivers_overlay) {
    strncat(options, " riversoverlay", sizeof(options));
  }
  secfile_insert_str(fp, options, "savefile.options");

  secfile_insert_int(fp, map->generator, "map.generator");
  secfile_insert_int(fp, map->seed, "map.seed");
  secfile_insert_int(fp, map->type, "map.maptype");
  secfile_insert_int(fp, map->landpercent, "map.landpercent");
  secfile_insert_int(fp, map->riches, "map.riches");
  secfile_insert_int(fp, map->huts, "map.huts");

  /* HACK ALERT: convert percentages for unwashed types for savegame */
  secfile_insert_int(fp, map->RIVERS,               "map.riverlength");
  secfile_insert_int(fp, map->DESERTS,              "map.deserts");
  secfile_insert_int(fp, map->TUNDRA+map->ARCTIC,   "map.tundrasize");
  secfile_insert_int(fp, map->MOUNTAINS+map->HILLS, "map.mountains");
  secfile_insert_int(fp, map->FORESTS+map->JUNGLES, "map.forestsize");
  secfile_insert_int(fp, map->SWAMPS+map->OCEANS,   "map.swampsize");
  secfile_insert_int(fp, map->GRASSLAND+map->PLAINS,"map.grasssize");

  secfile_insert_bool(fp, map->is_earth, "map.is_earth");
  secfile_insert_bool(fp, map->has_huts, "map.have_huts");

  secfile_insert_bool(fp, (map->num_start_positions > 0), "game.save_starts");
  if (map->num_start_positions > 0) {
    CIVMAP *pmap = (CIVMAP *)map;
    create_start_positions(pmap);
    secfile_insert_bool(fp,
      pmap->fixed_start_positions, "map.fixed_start_positions");
    for (nn = 0;  nn < pmap->num_start_positions; nn++) {
      secfile_insert_int(fp, pmap->start_positions[nn].x, "map.r%dsx", nn);
      secfile_insert_int(fp, pmap->start_positions[nn].y, "map.r%dsy", nn);
    }
  }

  /* TODO: fix these */
  secfile_insert_bool(fp,  0, "game.save_players");
  secfile_insert_bool(fp,  0, "game.private_map");
  secfile_insert_bool(fp,  0, "game.save_known");

  secfile_insert_int(fp, map->xsize, "map.width");
  secfile_insert_int(fp, map->ysize, "map.height");
  if (map->tiles)
    civtile_save(map, fp);

  map_printstats(map, 0);
  return 0;
}

/**
 * Returns the total number of (real) positions (or tiles) on the map.
 */
int civmap_map_num_tiles(const CIVMAP *pmap)
{
  return pmap->xsize * pmap->ysize;
}

/**
 * Returns the total extent of the x dimension on the standard map
 */
int civmap_std_xsize(const CIVMAP *pmap)
{
  int nn;

  if (civmap_get_type(pmap, MAP_TYPE_ISO)) 
    nn = (pmap->xsize+pmap->ysize/2);
  else
    nn = pmap->xsize;
  return nn;
}

/**
 * Returns the total extent of the y dimension on the standard map
 */
int civmap_std_ysize(const CIVMAP *pmap)
{
  int nn;

  if (civmap_get_type(pmap, MAP_TYPE_ISO)) 
    nn = (pmap->xsize+pmap->ysize/2);
  else
    nn = pmap->ysize;
  return nn;
}

/**
 * Topology function to find the vector which has the minimum "real"
 * distance between the map positions (x0, y0) and (x1, y1).  If there is
 * more than one vector with equal distance, no guarantee is made about
 * which is found.
 *
 * Real_distance is defined as the larger of the distances in the x and y
 * direction; since units can travel diagonally this is the "real" distance
 * a unit has to travel to get from point to point.
 *
 * sq_distance is defined to be the square of the trigonometric or
 * Pythagorean distance.
 *
 * city_distance is the city grid or Manhattan distance, i.e. sum of the
 * x and y distances with no diagonal moves allowed.
 *
 * The ranges of the return values are typically:
 *   -pmap->xsize/2 < dx <= pmap->xsize/2   i.e. wrapped coordinate
 *   -pmap->ysize   < dy <  pmap->ysize     i.e. non-wrapped coordinate
 *
 * TODO: this is really an optimized unnormalize, someday use the real thing.
 */
void civmap_distance(const CIVMAP *pmap, 
  int *dx, int *dy, int x0, int y0, int x1, int y1)
{
  if( civmap_get_type(pmap, MAP_TYPE_WRAPX|MAP_TYPE_WRAPY) ) {
    /* Wrapping is done in native coordinates */
    civmap_to_native_pos(pmap, &x0, &y0, x0, y0);
    civmap_to_native_pos(pmap, &x1, &y1, x1, y1);

    *dx = x1 - x0;
    if( civmap_has_wrap(pmap, MAP_TYPE_WRAPX) ) {
      while (*dx > pmap->xsize/2) 
        *dx -= pmap->xsize;
      while (*dx <= -pmap->xsize/2)
        *dx += pmap->xsize;
    }

    *dy = y1 - y0;
    if( civmap_has_wrap(pmap, MAP_TYPE_WRAPY) ) {
      while (*dy > pmap->ysize/2) 
        *dy -= pmap->ysize;
      while (*dy <= -pmap->ysize/2)
        *dy += pmap->ysize;
    }

    /* Convert the native delta back to standard map */
    y1 = y0 + *dy;
    x1 = x0 + *dx;
    native_to_civmap_pos(pmap, &x0, &y0, x0, y0);
    native_to_civmap_pos(pmap, &x1, &y1, x1, y1);
  }
  *dx = x1 - x0;
  *dy = y1 - y0;

  /**
   * standard axes lie along N-S (y') and W-E (x') lines (i.e. NW = (0,0) 
   * hex axes lie along NW-SE (y') and SW-NE (x') lines with y' scaled by 2 
   *
   * The dx and dy values are real distances, but no longer deltas in the 
   * (x0,y0) standard coordinate system after this point, rather (x',y').
   */
  if( civmap_get_type(pmap, MAP_TYPE_HEX)) {
    *dy += *dx;
    *dx -= (*dy > 0 ? *dy+1 : *dy)/2;
  }
}

/**
 * @return real_distance or MAX(abs(dx),abs(dy))
 */
int civmap_real_distance(const CIVMAP *pmap, int x0, int y0, int x1, int y1)
{
  int dx, dy;

  civmap_distance(pmap, &dx, &dy, x0, y0, x1, y1);
  dx = abs(dx), dy = abs(dy);
  return MAX(dx, dy);
}

/**
 * @return city_distance or abs(dx) + abs(dy)
 */
int civmap_city_distance(const CIVMAP *pmap, int x0, int y0, int x1, int y1)
{
  int dx, dy;

  civmap_distance(pmap, &dx, &dy, x0, y0, x1, y1);
  return abs(dx) + abs(dy);
}

/**
 * @return sq_distance or (dx*dx + dy*dy)
 */
int civmap_sq_distance(const CIVMAP *pmap, int x0, int y0, int x1, int y1)
{
  int dx, dy;

  civmap_distance(pmap, &dx, &dy, x0, y0, x1, y1);
  return (dx*dx + dy*dy);
}

/**
 * @return TRUE if the two map positions are adjacent
 */
bool civmap_is_adjc(const CIVMAP *pmap, int x0, int y0, int x1, int y1)
{
  return (civmap_real_distance(pmap, x0, y0, x1, y1) == 1);
}

/**
 * @return TRUE if the two map positions are equivalent (when normalized).
 * This function might be necessary ALOT of places...
 */
bool same_civmap_pos(const CIVMAP *pmap, int x1, int y1, int x2, int y2)
{
  /* Do full/slow normalized checks only if the simple case fails. */
  return
    (  ( x1 == x2 && y1 == y2 )
    || ( normalize_civmap_pos(pmap, &x1, &y1)
      && normalize_civmap_pos(pmap, &x2, &y2)
      && x1 == x2 && y1 == y2 ) );
}

/**
 * Convert from standard map to compact native coordinates.
 * Return value is always TRUE as this is a simple transformation.
 * Note: explicitly handle odd natural iso-x to force round down in native, 
 *       even when -ve. Shift of even *py is equivalent whether +ve or -ve, 
 *       /2 round to zero or div 2 round down.
 */
int civmap_to_native_pos(const CIVMAP *pmap, int *px, int *py, int mx, int my)
{
  if( civmap_get_type(pmap, MAP_TYPE_ISO) ) {
    *py= mx + my - pmap->xsize;
    *px= mx - (((*py & 1) + *py) >> 1);
  } else {
    *py= my, *px= mx;
  }
  return TRUE;
}

/**
 * Convert from compact native to standard map coordinates.
 * Return value is always TRUE as this is a simple transformation.
 * Note: explicitly recover odd natural iso-x bit by rounding up the y value,
 *       even when -ve. Shift of even y is equivalent whether +ve or -ve, 
 *       /2 round to zero or div 2 round down.
 */
int native_to_civmap_pos(const CIVMAP *pmap, int *pmx, int *pmy, int x, int y)
{
  if( civmap_get_type(pmap, MAP_TYPE_ISO) ) {
    *pmx= (((y & 1) + y) >> 1) + x;
    *pmy= y - *pmx + pmap->xsize;
  } else {
    *pmy= y, *pmx= x;
  }
  return TRUE;
}


/**
 * Tile coordinates are real if they can be converted to a valid normalized
 * form. Currently all wrappable coordinates, or those already within the
 * valid normal range are real.
 *
 * Filter conditions may be added here if they do not involve any transform
 * operations such as wrapping.
 *
 * When topologies potentially require both transformation and filtering to
 * test for realness, this will effectively involve a full normalize cycle.
 *
 * This is used as a fast check to avoid the full normalize cycle.
 */
bool is_real_civmap_pos(const CIVMAP *pmap, int x, int y)
{
  /* reality is best determined in native coordinates */
  civmap_to_native_pos(pmap, &x, &y, x, y);

  if( (civmap_has_wrap(pmap, MAP_TYPE_WRAPX) || RANGE_CHECK_0(pmap->xsize,x))
   && (civmap_has_wrap(pmap, MAP_TYPE_WRAPY) || RANGE_CHECK_0(pmap->ysize,y)) )
    return TRUE;
  return FALSE;
}

/**
 * Normalize the map position. Returns FALSE if the coordinates are not real,
 * Returns TRUE plus correct normalized coordinates if they are real.
 *
 * Coordinates should be updated only when real and unnormal. Any operation
 * on an unreal tile is illegal, and normalized tiles do not need updating.
 *
 * This is optimized specifically for 2-D rectangular maps. Use the
 * generalized unnormalize_map_pos with appropriate parameters otherwise.
 */
bool normalize_civmap_pos(const CIVMAP *pmap, int *px, int *py)
{
  int x, y, unnormal = 0;

  /* normalization is best determined in native coordinates */
  civmap_to_native_pos(pmap, &x, &y, *px, *py);

  /* Lazy realness check part 1 - fail if unnormal and cannot wrap. */
  if (!RANGE_CHECK_0(pmap->xsize, x)) {
    if (!civmap_has_wrap(pmap, MAP_TYPE_WRAPX))
      return FALSE;
    unnormal++;

    /*
     * In the normal case the while body doesn't execute, and on border
     * tiles this should happen only once, and occasionally twice.
     * Worry about (rare) multiple loops for safety, not performance.
     */
    while (x < 0)
      x += pmap->xsize;
    while (x >= pmap->xsize)
      x -= pmap->xsize;
  }

  /* Lazy realness check part 2 - fail if unnormal and cannot wrap. */
  if (!RANGE_CHECK_0(pmap->ysize, y)) {
    if (!civmap_has_wrap(pmap, MAP_TYPE_WRAPY))
      return FALSE;
    unnormal++;

    /*
     * In the normal case the while body doesn't execute, and on border
     * tiles this should happen only once, and occasionally twice.
     * Worry about (rare) multiple loops for safety, not performance.
     */
    while (y < 0)
      y += pmap->ysize;
    while (y >= pmap->ysize)
      y -= pmap->ysize;
  }

  /* Now map things back to standard map coordinates */
  if (unnormal)
    return native_to_civmap_pos(pmap, px, py, x, y);

  return TRUE;
}

/**
 * Generalized rectangular (un)normalize (in map coordinates)
 *
 * This finds the point in the (iso-)rectangular normalized set at
 * offset {x0,y0}. For isometric rectangles, the wrapping condition
 * test is given by a linear combination of the standard map coords.
 *
 * This still assumes a native rectangular map shape with x, y
 * wrapping conditions.
 *
 * Note: normalize_civmap_pos() is just this function called with
 *       {x0,y0} of the standard map origin and to_iso false.
 * Note: normalize_mapiso_pos() is just this function called with
 *       {x0,y0} of the standard mapiso origin and to_iso true.
 *
 * TODO: replace to_iso with target shape at next general shape extension.
 *       for the moment this is a bit overkill for the standard 4
 *       topologies in normal or iso views. 
 */
bool unnormalize_civmap_pos(const CIVMAP *pmap, 
  int *un_x, int *un_y, int x0, int y0, int to_iso)
{
  static int ij[2][2], uv[2][2], wraptype = 0;
  int nn, tt[2], vv[2], v0[2];

#define DOT(u,v)    (u[0]*v[0] + u[1]*v[1])
#define ADD(t,u,v)  (t[0]= u[0] + v[0], t[1]= u[1] + v[1])
#define SUB(t,u,v)  (t[0]= u[0] - v[0], t[1]= u[1] - v[1])

  /* unnormalize is only valid for real coordinates */
  if( !is_real_civmap_pos(pmap, *un_x, *un_y) )
    return FALSE;

  /* if no wrappable coordinates, just return is_real */
  if( (nn = civmap_get_type(pmap, MAP_TYPE_WRAPX|MAP_TYPE_WRAPY)) == 0 )
    return TRUE;

  /* Define the wrapping transformations (once and save) */
  if( nn != wraptype ) {
    wraptype = nn;
    switch(wraptype) {
    case MAP_TYPE_WRAPX:
      /* Only wraps in the standard map X direction */
      uv[0][0] = pmap->xsize;
      uv[0][1] = uv[1][0] = uv[1][1] = 0;
      break;
    case MAP_TYPE_WRAPY:
      /* Only wraps in the standard map Y direction */
      uv[0][0] = uv[0][1] = uv[1][0] = 0;
      uv[1][1] = pmap->ysize;
      break;
    case MAP_TYPE_WRAPX|MAP_TYPE_WRAPY:
      /* Wraps in both map X&Y directions */
      uv[0][0] = pmap->xsize;
      uv[0][1] = uv[1][0] = 0;
      uv[1][1] = pmap->ysize;
      break;
    default:
      /* No wrap - never used */
      uv[0][0] = uv[0][1] = uv[1][0] = uv[1][1] = 0;
      break;
    }

    /* set coordinate transformation vectors to standard identity */
    ij[0][0] = ij[1][1] = 1, ij[0][1] = ij[1][0] = 0;
  }

  /* unnormalization is best determined in native coordinates */
  civmap_to_native_pos(pmap, &vv[0], &vv[1], *un_x, *un_y);
  civmap_to_native_pos(pmap, &v0[0], &v0[1], x0, y0);

  /* pre-unnormalize - vv will have +ve offsets wrt v0 in native coordinates */
  if( civmap_has_wrap(pmap, MAP_TYPE_WRAPX) )
    vv[0] -= v0[0], 
    vv[0] = (vv[0]%pmap->xsize < 0 ? pmap->xsize : 0) + vv[0]%pmap->xsize,
    vv[0] += v0[0];
  if( civmap_has_wrap(pmap, MAP_TYPE_WRAPY) )
    vv[1] -= v0[1], 
    vv[1] = (vv[1]%pmap->ysize < 0 ? pmap->ysize : 0) + vv[1]%pmap->ysize,
    vv[1] += v0[1];

  /* This is a mongrel algorithm to deal with final fixes to pi/4 rotated
   * aka iso rectangluar window in a normal rectangular map or vice versa.
   * Note, if the target and native wrapping axes are aligned, this step
   * and/or code section is not needed. There are also many different
   * mappings possible, of which the one below is the most compact shape
   * aligned as much as possible along the rectangular target axes.
   *
   * Warning: this is almost impossible to understand without drawing am
   *          accurate graph of the two overlayed views and tricky then.
   *          Each operation effectively moves part of one view bounded
   *          by the appropriate lines through a wrapping translation.
   *          For a simple WRAPX map, it maps a native rectangle into a 
   *          parallelogram in one step by wrapping a triangular region.
   *          The pi/4 side now parallels one of the standard map axes.
   *          For a TORUS map, not all of the triangular shape above is
   *          taken, but bits are left for wrap in the other direction
   *          to clean up. The final shape is an "L" or "N" with the 
   *          letter strokes parallel to the standard map axes. Note "L"
   *          is just an "N" with one leg shrunk and added to the other
   *          reflecting the constraints imposed by the window origin.
   *          Because of the pre-normalization above, and carefully 
   *          chosen operations and order, only one final adjustment is
   *          needed in each direction.
   *
   * When adding new target shapes, only this sort of code should be
   * required, i.e. code that maps a native normal set into a defined
   * normal target shape set.
   */
  if( (to_iso != 0) != (civmap_get_type(pmap, MAP_TYPE_ISO) != 0) ) {
    int n0, n1, n2, n3;

    if( to_iso ) {
      ij[0][1] = -1, ij[1][0] =  1;

      /* x bounds in target view */
      n0 = DOT(v0, ij[0]);
      ADD(tt, v0, uv[0]);
      n2 = DOT(tt, ij[0]);

      /* y bounds in target view */
      n1 = DOT(v0, ij[1]);
      ADD(tt, v0, uv[1]);
      n3 = DOT(tt, ij[1]);

      if( civmap_has_wrap(pmap, MAP_TYPE_WRAPX) )
        while( DOT(vv, ij[0]) < n0
          && (!civmap_has_wrap(pmap, MAP_TYPE_WRAPY) || DOT(vv, ij[1]) < n3) )
          ADD(vv, vv, uv[0]);
      if( civmap_has_wrap(pmap, MAP_TYPE_WRAPY) )
        while( DOT(vv, ij[1]) >= n3 
          && (!civmap_has_wrap(pmap, MAP_TYPE_WRAPX) || DOT(vv, ij[0]) < n2) )
          SUB(vv, vv, uv[1]);
    } else { /* from iso */
      ij[0][1] =  1, ij[1][0] = -1;

      /* x bounds in target view */
      n0 = DOT(v0, ij[0]);
      ADD(tt, v0, uv[0]);
      n2 = DOT(tt, ij[0]);

      /* y bounds in target view */
      n1 = DOT(v0, ij[1]);
      ADD(tt, v0, uv[1]);
      n3 = DOT(tt, ij[1]);

      if( civmap_has_wrap(pmap, MAP_TYPE_WRAPY) )
        while( DOT(vv, ij[1]) < n1 
          && (!civmap_has_wrap(pmap, MAP_TYPE_WRAPX) || DOT(vv, ij[0]) < n2) )
          ADD(vv, vv, uv[1]);
      if( civmap_has_wrap(pmap, MAP_TYPE_WRAPX) )
        while( DOT(vv, ij[0]) >= n2 
          && (!civmap_has_wrap(pmap, MAP_TYPE_WRAPY) || DOT(vv, ij[1]) < n1) )
          SUB(vv, vv, uv[0]);
    }
  }

  /* put things back to standard map coordinates */
  native_to_civmap_pos(pmap, un_x, un_y, vv[0], vv[1]);
  return TRUE;
#undef  DOT
#undef  ADD
#undef  SUB
}

