/**********************************************************************
 Freeciv - Copyright (C) 1996 - A Kjeldberg, L Gregersen, P Unold
 civmap.h - Copyright (C) 2002 - R. Wetmore

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
***********************************************************************/
#ifndef FC__CIVMAP_H
#define FC__CIVMAP_H

/**
 * Civmap.h - basic Freeciv map header definitions and methods
 *
 *   This is loosely synced with FREECIV map.h. The hard compatiblity
 * constraint is the ability to load/save each other's savegame files.
 *
 *   The global map struct holds references to all collective map
 * parameters plus a reference to a vector of tiles for individual
 * properties of single map positions. Global concepts such as map
 * dimensions, topology, city extents or tile relationships as in
 * adjacency rules plus methods involving such concepts are accessed
 * through the map level.
 *
 *   Individual tile data is commmonly accessed by map position. A set
 * of convenience wrapper functions for retrieving tile data by map
 * position is defined in civmap_ex.h. Thus one can interchangeably use 
 *   civtile_get_food(CIVTILE *ptile)
 *     or
 *   civmap_get_food(CIVMAP *pmap, x, y)
 *     equivalent to
 *   civtile_get_food(civmap_get_tile(CIVMAP *pmap, x, y))
 * which hides the civmap_get_tile(CIVMAP *pmap, x, y) step for lots of
 * performance abuse.
 *
 *   In some cases the civmap flavours may perform additional work,
 * such as recomputing movecosts to adjacent tiles when updating the 
 * tile terrain. Civmap updates include global effects, civtile just
 * the local ones.
 *
 * Style Notes:
 *
 *   All civmap functions or accessors are prefaced with civmap_ 
 * followed by the name of the data value or method used. Similarly,
 * all civtile functions are prefaced by civtile_. If or when there
 * is an upgrade to C++ or another object oriented language, the
 * migration of such elements should be obvious.
 *
 *   Typically the hierarchy in civmap_tiles_load reflects a fuzzy
 * belief that tiles is the next level of object with a load method.
 * civmap_get_terrain on the contrary implies that terrain is a 
 * concept or simple data element of the civmap level.
 */

/* required dependencies */

#include "capability.h"
#include "registry.h"

#include "civterrain.h"
  
/* From common/game.h - used in load/save operations */

enum server_states { 
  PRE_GAME_STATE, 
  SELECT_RACES_STATE, 
  RUN_GAME_STATE,
  GAME_OVER_STATE
};

enum client_states {
  CLIENT_BOOT_STATE,
  CLIENT_PRE_GAME_STATE,
  CLIENT_SELECT_RACE_STATE,
  CLIENT_WAITING_FOR_GAME_START_STATE,
  CLIENT_GAME_RUNNING_STATE,
  CLIENT_GAME_OVER_STATE
};


/* From server/savegame.c - used in load/save operations */

/* The following should be removed when compatibility with
 * pre-1.13.0 savegames is broken: startoptions, spacerace2
 * and rulesets
 */
#define SAVEFILE_OPTIONS "startoptions spacerace2 rulesets"  \
" diplchance_percent worklists2 map_editor known32fix turn"  \
" attributes watchtower rulesetdir client_worklists"


/* From common/unit.h,map.h (vaguely) */

/*
 * MOVECOSTs - currently these values must fit in an unsigned char
 *
 *   These reflect the cost for units to move from one tile to the next
 * scaled so the smallest incremental value is integral. Thus if ROAD
 * movement is the smallest and is one third of a normal move the 
 * values for a ROAD move is 1 and SINGLE move 3.
 *   Land and sea moves share the movecost space, sea moves are thus
 * negative to distinguish them. This forces the restriction that no
 * tile-to-tile move can be both a land and a sea move. The only two
 * cases where this is an issue are city-to-city moves and river-to-
 * city or city-to-river when river movement is enabled. This also
 * blocks extension to COASTAL bridges/tunnels, i.e. the ROAD/RAIL
 * equivalent for land movement through coastal tiles.
 */
enum movecost_type {
  MOVECOST_MAX   = -1,           /* maximum move (==255), invalid move flag */
  MOVECOST_SEA   = -3,           /* ocean/coastal move must be negative     */
  MOVECOST_RAIL  =  0,           /* railroad move (instananeous)            */
  MOVECOST_ROAD  =  1,           /* road move (also IGTER move)             */
  MOVECOST_RIVER =  1,           /* river move (see rules for connectivity) */
  MOVECOST_AIR   =  3,           /* air move (equivalent to single move)    */
  MOVECOST_LAND  =  3,           /* land move (equivalent to single move)   */
  SINGLE_MOVE    =  3,           /* basic movement rate multiplier          */
};


/* common/map.h */
/**
 * Basic 2-D map location as in the standard internal cartesian form.
 *
 * TODO:  expand this to something like ...
 *
 * struct map_position {
 *   short x, y; / * internal standard 2-D map (cartesian) coordinates   * /
 *   short u, v; / * native 2-D coordinates (load/save and topology ops) * /
 *   int uv;     / * linear 1-D coordinates (internal memory access)     * /
 *   int type;   / * type index for topology and conversion routines     * /
 * };
 *
 * In general, new coordinates should be created and filled in during the
 * initial realness or normalization check and passed around as CIVPOINTs.
 * Access to internal, native or memory forms would be by simple lookup
 * with no additional conversion expense. Lazy normalization would 
 * initialize using only a single form. An exception would be caught and
 * the full normalization performed when accessed in another mode.
 */

struct map_position {
  int  x, y;
};

typedef struct map_position CIVPOINT;


/* common/city.h */
#define CITY_MAP_SIZE	5


/* common/nation.h */
#define MAX_NUM_NATIONS	63


/**
 * basic map header structure (plus map generation parameters)
 */
typedef struct civmap
{
  /* used during map generation only (TODO: abstract to separate struct?) */

  int generator;
  int seed;                     /* map generation random number seed      */
  int landpercent;              /* percentage allocated to land tiles     */
  int terrainpercent[16];	/* 16 basic terrain types                 */
  int riverpercent[16]; 	/* treats rivers as overlay/special_0     */
  int specialpercent[2][16];	/* 2 (3?) basic special modifiers         */
  int riches;                   /* maximum number of specials             */
  int huts;                     /* maximum number of native villages      */

  /* the remaining map parameters are used/updated during runtime play    */

  int xsize, ysize;		/* native map dimensions - 80x50          */
  int type;                     /* map type index                         */
  int maxcost;			/* more than the fastest unit moverate    */
  int num_start_positions;      /* active start positions (below)         */
  int num_continents;           /* summary statistic                      */
  int num_oceans;               /* summary statistic                      */

  bool is_earth;                /* TRUE if Denmark tile is marked         */
  bool tinyisles;               /* TRUE if 1x1 islands are allowed        */
  bool separatepoles;           /* TRUE if mandated polar passage         */
  bool has_huts;                /* summary statistic                      */
  bool has_specials;            /* CIV I vs CIV II terrain types          */
  bool has_rivers_overlay;      /* only applies if !have_specials         */
  bool fixed_start_positions;   /* do not select starting positions       */

  CIVPOINT start_positions[MAX_NUM_NATIONS+1];

  const CIVTERRAIN *terrain;    /* the actual runtime terrain types       */
  void *tiles;                  /* the actual runtime native map vector   */
} CIVMAP;

/*
 * Convenience/readability struct field mappings
*/
#define	ARCTIC  	terrainpercent[T_ARCTIC]
#define	DESERTS 	terrainpercent[T_DESERT]
#define	FORESTS 	terrainpercent[T_FOREST]
#define	GRASSLAND	terrainpercent[T_GRASSLAND]
#define	HILLS   	terrainpercent[T_HILLS]
#define	JUNGLES 	terrainpercent[T_JUNGLE]
#define	MOUNTAINS	terrainpercent[T_MOUNTAINS]
#define	OCEANS  	terrainpercent[T_OCEAN]
#define	PLAINS  	terrainpercent[T_PLAINS]
#define	RIVERS  	terrainpercent[T_RIVER]
#define	SWAMPS  	terrainpercent[T_SWAMP]
#define	TUNDRA  	terrainpercent[T_TUNDRA]


/* Hardcoded default settings */

static const int  COMPRESSED = 9;

static const char *const CIVMAPFILE = "civmap.sav.gz";
static const char *const CIVMAPLOG  = "civmap.log";
static const char *const CIVMAPRC   = ".civmaprc";

static const struct civmap CIVMAP_DEFAULT =
{
  1,           /* int generator;                                          */
  0,           /* int seed;                                               */
  33,          /* int landpercent;                                        */
  {  5,  5, 20, 20, 15,  5,  5,  0, 15, 25,  5, 5,  0, 0, 0, 0 },
               /* int terrainpercent[16];    * terrain "adfghjm.prstX123" */
  {  5,  5, 25, 25, 10, 25, 10,  0, 25,100, 50,10,  0, 0, 0, 0 },
               /* int riverpercent[16];      * treat rivers as special_0  */
  {{ 8,  8,  8, 20,  8,  8,  8,  0,  8,  0,  8, 8,  8, 8, 8, 8 },
  {  8,  8,  8,  8,  8,  8,  8,  0,  8,  0,  8, 8,  8, 8, 8, 8 }},
               /* int specialpercent[2][16]; * 2 basic special modifiers  */
  120,         /* int riches;                                             */
  50,          /* int huts;                                               */

  80, 50,      /* int xsize, ysize;          * native map dimensions      */
  1,           /* int type;                                               */
  0,           /* int maxcost;                                            */
  1,           /* int num_start_positions;                                */
  0,           /* int num_continents;                                     */
  0,           /* int num_oceans;                                         */

  FALSE,       /* bool is_earth;                                          */
  FALSE,       /* bool tinyisles;                                         */
  TRUE,        /* bool separatepoles;                                     */
  TRUE,        /* bool have_huts;                                         */
  TRUE,        /* bool have_specials;                                     */
  TRUE,        /* bool have_rivers_overlay;  * only if !have_specials     */
  FALSE,       /* bool fixed_start_positions;                             */

  { {0,0} },   /* struct map_position start_positions[MAX_NUM_NATIONS];   */
  NULL,        /* struct terrain *terrain;                                */
  NULL,        /* struct tile *tiles;                                     */
};


/* External definitions */

extern struct civmap *theMap;
extern int verbose;

void civmap_tiles_free(CIVMAP *map);
CIVMAP *civmap_tiles_init(CIVMAP *map);

void civmap_free(CIVMAP **map);
CIVMAP *civmap_init(CIVMAP *map);

int  civmap_load(CIVMAP *map, struct section_file *fp);
int  civmap_save(const CIVMAP *map, struct section_file *fp);

bool is_real_civmap_pos(const CIVMAP *map, int x, int y);
bool normalize_civmap_pos(const CIVMAP *map, int *x, int *y);
bool unnormalize_civmap_pos(const CIVMAP *map,
        int *ux, int *uy, int x0, int y0, int target_shape);

int  civmap_to_native_pos(const CIVMAP *map, int *nx, int *ny, int mx, int my);
int  native_to_civmap_pos(const CIVMAP *map, int *mx, int *my, int nx, int ny);

int  civmap_get_type(const CIVMAP *map, int maptype);
bool civmap_has_wrap(const CIVMAP *map, int wraptype);
bool civmap_is_type(const CIVMAP *map, int maptype);

int civmap_map_num_tiles(const CIVMAP *pmap);
int civmap_std_xsize(const CIVMAP *pmap);
int civmap_std_ysize(const CIVMAP *pmap);

void civmap_distance(const CIVMAP *pmap,
        int *dx, int *dy, int x0, int y0, int x1, int y1);
int civmap_real_distance(const CIVMAP *pmap, int x0, int y0, int x1, int y1);
int civmap_city_distance(const CIVMAP *pmap, int x0, int y0, int x1, int y1);
int civmap_sq_distance(const CIVMAP *pmap, int x0, int y0, int x1, int y1);

bool civmap_is_adjc(const CIVMAP *pmap, int x0, int y0, int x1, int y1);
bool same_civmap_pos(const CIVMAP *pmap, int x1, int y1, int x2, int y2);


/*  ======================================================================  */


/* The following are used extensively, so typically are optimized as macros.*/

/* Internal macro temporary storage - not multi-thread safe */
extern int _FC_MAP_N, _FC_MAP_X, _FC_MAP_Y;

#ifdef FC__CIVMAP_C            /* allocated in common/map.c */
int _FC_MAP_N, _FC_MAP_X, _FC_MAP_Y;
#endif


/* Map topology definitions and map dimension related macros                */

/*
 * The default maptype (1) reflects a cylindrical map wrapping at map->xsize
 * in standard cartesian coordinates with octagonal adjacency.
 * 
 * Hexagonal maps have no (0,4) directions, (NW,SE) or (N,S) iso, the map
 * distances are computed from ((dx-dy)/2,(dx+dy)) or (dx/2, dy), and the
 * "cartesian" directions for rivers are 6-fold vs 4-fold.
 */
#ifndef _map_type_defined
#define _map_type_defined
enum MAP_TYPE {
  MAP_TYPE_FLAT    = 0,  /* Note: not a WRAPTYPE                     */
  MAP_TYPE_WRAPX   = 1,
  MAP_TYPE_WRAPY   = 2,
  MAP_TYPE_TORUS   = 3,
  MAP_TYPE_ISO     = 4,  /* map is stored in isometric form          */
  MAP_TYPE_HEX     = 8,  /* map has hexagonal vs octagonal adjacency */
  MAP_TYPE_ISEARTH = 16, /* map == special type (Denmark flagged)    */
  MAP_TYPE_LAST    = 32
};
#endif

/* Some common fully supported map types */
#define	MAP_FLATEARTH           ( MAP_TYPE_FLAT  )
#define	MAP_EARTH               ( MAP_TYPE_WRAPX )
#define	MAP_URANUS              ( MAP_TYPE_WRAPY )
#define	MAP_DONUTWORLD          ( MAP_TYPE_TORUS )

#define	MAP_FLATEARTH_ISO       ( MAP_TYPE_FLAT  | MAP_TYPE_ISO )
#define	MAP_EARTH_ISO           ( MAP_TYPE_WRAPX | MAP_TYPE_ISO )
#define	MAP_URANUS_ISO          ( MAP_TYPE_WRAPY | MAP_TYPE_ISO )
#define	MAP_DONUTWORLD_ISO      ( MAP_TYPE_TORUS | MAP_TYPE_ISO )

/*
 * These two macros allow one to determine the current game map topology.
 * They are primarily used in the key map coordinate macros, but are used
 * in scattered code that needs to knwo this information.
 *
 * The argument to is_type is one of the enum MAP_TYPES.
 * The argument to has_wrap is one of the non-zero enum MAP_TYPES.
 */
#define civmap_is_type(MAP, TYPE)       ((MAP)->type == (TYPE))
#define civmap_get_type(MAP, MASK)      ((MAP)->type & (MASK))
#define civmap_has_wrap(MAP, WRAPTYPE)  \
  (civmap_get_type((MAP), (WRAPTYPE)) == (WRAPTYPE))


/*
 * These macros convert between native coordinates or a linear native index
 * and the standard internal civmap coordinates.
 * Currently, only standard and (compressed) ISO native forms are supported.
 * Both native forms normalize to a rectangular grid and are fully wrappable.
 */
#if !defined(FC__NO_CIVMAP_MACROS) && !defined(FC__CIVMAP_C)

/* Shift operations handle /2 as divide and round up for +ve or -ve cases.
 * Thus the natural iso_x coordinate is compressed with round down for native.
 * Note: operations carefully ordered for inplace conversion - return TRUE
 */
#define civmap_to_native_pos(MAP, PX, PY, MX, MY)     /* std map to native */ \
( (civmap_get_type((MAP), MAP_TYPE_ISO)                                       \
  ? (*(PY)= (MX) + (MY) - (MAP)->xsize,                                       \
     *(PX)= (MX) - (((*(PY) & 1) + *(PY)) >> 1))                              \
  : (*(PY)= (MY), *(PX)= (MX))), TRUE )

/* Shift operations handle /2 as divide and round up. This recovers the lost
 * natural iso_x bit, since the sum of iso_x + iso_y must be even.
 * Note: operations carefully ordered for inplace conversion - return TRUE
 */
#define native_to_civmap_pos(MAP, PMX, PMY, X, Y)     /* native to std map */ \
( (civmap_get_type((MAP), MAP_TYPE_ISO)                                       \
  ? (*(PMX)= ((((Y)&1)+(Y)) >> 1) + (X), *(PMY)= (Y) - *(PMX) + (MAP)->xsize) \
  : (*(PMY)= (Y), *(PMX)= (X))), TRUE )


/* Returns the index value (exception to the 2-D coordinate interface) */
#define native_pos_to_index(MAP, X, Y)               /* native to xy index */ \
( (MAP)->xsize * (Y) + (X) )

/* negative indicies are invalid, this at least keeps x +ve  - return TRUE */
#define index_to_native_pos(MAP, PX, PY, XY)         /* xy index to native */ \
( *(PY)= (XY)/(MAP)->xsize,                                                   \
  *(PX)= ((XY)%(MAP)->xsize < 0                                               \
       ? ((*(PY))--, (MAP)->xsize) : 0) + (XY)%(MAP)->xsize, TRUE )


/* Returns the index value (exception to the 2-D coordinate interface) */
#define civmap_pos_to_index(MAP, MX, MY)                /* map to xy index */ \
( _civmap_pos_to_index((MAP), &_FC_MAP_X, &_FC_MAP_Y, (MX), (MY)) )

/* Returns the index value exposing the native results in (*PX,*PY) */
#define _civmap_pos_to_index(MAP, PX, PY, MX, MY)       /* map to xy index */ \
( civmap_to_native_pos((MAP), (PX), (PY), (MX), (MY)),                        \
  native_pos_to_index((MAP), *(PX), *(PY)) )

/* negative indicies are invalid, this at least keeps x +ve  - return TRUE */
#define index_to_civmap_pos(MAP, PMX, PMY, PX, PY, XY)  /* xy index to map */ \
( index_to_native_pos((MAP), (PX), (PY), (XY))                                \
  && native_to_civmap_pos((MAP), (PMX), (PMY), *(PX), *(PY)) )


/*
 * The following sets of macros or functions control the game's internal
 * map navigation based on the currently allowed topological map types.
 * _<flavours> are internal working algorithms.
 *
 * is_real_civmap_pos(map,x,y) is TRUE if the position(x,y) in map *can* be
 *   mapped to a real tile. This is typically checked in native coordinates.
 *
 * normalize_civmap_pos(map,px,py,x,y) remaps any wrappable positions so
 *   if TRUE, position(*px,*py) is *guaranteed* to point to a real tile.
 *   If FALSE, (x,y) are unreal and should be untouched, i.e. no side effects.
 *   Wrapping is a complex vector operation if not done in native coordinates.
 *
 * _civmap_normalize() does a partial normalize on a single coordinate.
 *   It is a utility macro to simplify normalize_civmap_pos() and is not
 *   intended for general use.
 */
#define is_real_civmap_pos(MAP, MX,MY)                                        \
 _is_real_civmap_pos((MAP), &_FC_MAP_X,&_FC_MAP_Y, (MX),(MY))

/* This exposes the native coordinate (partial) results in (PX,PY). */
#define _is_real_civmap_pos(MAP, PX,PY, MX,MY)                                \
( civmap_to_native_pos((MAP), (PX),(PY), (MX),(MY)),                          \
  ( civmap_has_wrap((MAP), MAP_TYPE_WRAPX)                                    \
    || RANGE_CHECK_0((MAP)->xsize, *(PX)) ) &&                                \
  ( civmap_has_wrap((MAP), MAP_TYPE_WRAPY)                                    \
    || RANGE_CHECK_0((MAP)->ysize, *(PY)) ) )

#define normalize_civmap_pos(MAP, PMX,PMY)                                    \
 _normalize_civmap_pos((MAP), &_FC_MAP_X,&_FC_MAP_Y, (PMX),(PMY))

/* This exposes the native coordinate (partial) results in (PX,PY). */
#define _normalize_civmap_pos(MAP, PX,PY, PMX,PMY)                            \
( civmap_to_native_pos((MAP), (PX),(PY), *(PMX),*(PMY)),                      \
  (  _civmap_normalize((MAP), MAP_TYPE_WRAPX, (MAP)->xsize, (PX))             \
  && _civmap_normalize((MAP), MAP_TYPE_WRAPY, (MAP)->ysize, (PY))             \
  && native_to_civmap_pos((MAP), (PMX),(PMY), *(PX),*(PY)) ) )

/* Return FALSE if unreal, or TRUE and a *PPOS result guaranteed in range. */
#define _civmap_normalize(MAP, WRAP, BASE, PPOS)                              \
( RANGE_CHECK_0((BASE),*(PPOS))                                               \
  || ( civmap_has_wrap((MAP), (WRAP))                                         \
    && (*(PPOS)= (*(PPOS)%(BASE) < 0 ? (BASE) : 0) + *(PPOS)%(BASE), TRUE) ) )
#endif

/*
 * A "border position" is any one that has adjacent positions that are
 * unreal or not normalized. Alternatively, it is not "internal".
 *
 * civmap_is_border_pos(map, x,y) and
 * civmap_is_border2_pos(map, x,y) are TRUE if map position(x,y) is on the
 *   outer border of the native rectangular map or is outside the map. They
 *   are useful as a fast check for tile validity inside adjacent loops, 
 *   i.e. to skip the normalization step. In the second version the border
 *   is extended to a depth of two tiles.
 *
 * civmap_is_borderN_pos(MAP,N,X,Y) is a macro that treats border depth
 *   as a parameter, i.e. within "N" tiles of the real border. N can be
 *   considered a standard map distance, thus may be more than N tiles in
 *   some topologies.  Negative values are converted to positive distance.
 *   If the distance is zero or MAP is NULL, no check is made and the 
 *   result is FALSE. 
 */
#define civmap_is_borderN_pos(MAP, NN, MX, MY)                                \
( ((_FC_MAP_N = (NN)) != 0) && (MAP) &&                                       \
  ( _FC_MAP_N = (_FC_MAP_N < 0 ? -_FC_MAP_N : _FC_MAP_N),                     \
    _FC_MAP_N+= (civmap_get_type((MAP), MAP_TYPE_ISO) ? _FC_MAP_N+1 : 0),     \
    civmap_to_native_pos((MAP), &_FC_MAP_X, &_FC_MAP_Y, (MX), (MY)),          \
    (  !(_FC_MAP_N <= _FC_MAP_Y && _FC_MAP_Y < (MAP)->ysize-_FC_MAP_N)        \
    || !(_FC_MAP_N <= _FC_MAP_X && _FC_MAP_X < (MAP)->xsize-_FC_MAP_N) ) ) )

#define civmap_is_border_pos(MAP, MX, MY)                                     \
  civmap_is_borderN_pos((MAP), 1, (MX), (MY))

#define civmap_is_border2_pos(MAP, MX, MY)                                    \
  civmap_is_borderN_pos((MAP), 2, (MX), (MY))

/**********************************************************************/

/*
 * Data driven offset arrays when used to compute neighboring tiles, can
 * greatly simplify internal direction handling and adjacent loops.
 *
 * Using rDir = 0, 1, ... 7 and
 *   x1 = x + DIR_DX2[rDir];
 *   y1 = y + DIR_DY2[rDir];
 * will give you the tiles in clockwise rotation order as shown below.
 * 
 * -------  
 * |0|1|2|  NW (== 0) is the origin
 * |-+-+-|
 * |7| |3|
 * |-+-+-|
 * |6|5|4|
 * -------
 *
 * -------  
 * |0|1|2|  This is the array order used for example, by GUI tiling,
 * |-+-+-|  If the direction enums are given by the previous numbers,
 * |3| |4|  then a simple translation can be done to iterate in any
 * |-+-+-|  order such as aDir = 0, 1, ... 7 order at right.
 * |5|6|7|    DIR_ator[] = { 0, 1, 2, 7, 3, 6, 5, 4 }
 * -------    rDir = DIR_ator[aDir];
 * 
 * Array order is primarily used to store graphics in a continuous NW-SE
 * direction where overlaps are updated only on the leading edge.
 *
 * Data driven is both efficient in the inner loop as data lookup rather
 * than code is used to control the flow. It is also highly extensible to
 * new cases with no code changes other than substituion of the data
 * controls. For example, different topologies may have different
 * adjacent ordering or offsets. The same loop code with a topology
 * aware lookup to set its data controls will run any existing (and
 * as yet unimplemented) topologies without code change.
 *
 * The single core_iterate loop below, manages all standard map, iso
 * map and city map adjacent iterations currently used by freeciv.
*/

/**
 * Macros for moving around the 1st and 2nd square in NWSE directions
 *
 * There are 8n directions or edge tiles in the n'th square outwards,
 * (2n+1)^2 => 1, 9, 25 sized squares, (2n+1)^2 - (2n-1)^2= 8n in each.
 *
 * The Delta macros step to direction "n", wrapping at high/low ends.
 *
 * Iterate macros below use a single indexed while loop based on the
 * following known ordering, with access to the index (i.e. _nn), or
 * using the "_dir_iterate" form to provide a custom index.
 * Example: if _nn= 0, you are NW of your start.
 *
 * All allocate the user's chosen x,y,dir variables within the (core) loop.
 * All treat arguments other than x,y,dir as single use expressions.
 * All use a single loop on the index value _nn or supplied dir.
 * All blocks will be correctly wrapped in the x dimension at the
 * map->xsize boundary, and off map y values will be skipped. With a
 * properly set normalize_civmap_pos, this will be true for a select
 * set of other map topologies (see above). This is implemented by
 * running within an inner if clause, so "} else {" can be used to
 * process invalid map positions.
 *
 * Depends on map->xsize, map->ysize and normalize_civmap_pos.
 *
 * TODO: extend this to the third square or 7x7 to handle most local ops
*/

/* Extern definitions valid globally                                  */
extern const int DIR_off[], DIR_dx[], DIR_dy[];
extern const int DIR_ator[], DIR_rtoa[], DIR_ctor[], DIR_htor[];
extern const char *DIR_dn[];

/* Special case for adjacent directions in ring 1 */
#define dir_get_name(dir)   (DIR_dn+1 + (dir))


/* Actual variable allocations happen once in civmap.c                */
#ifdef FC__CIVMAP_C

/* Ring offset calculations for optimized lookup  (2n-1)^2            */
const int DIR_off[] = {
   0,  1,  9,  25, 
};

/* Array of offsets starting at the NW corner and moving clockwise,   */
/* including the origin. Note, first 1, 9, and 25 == square iteration */
/* WARNING: Code below depends on the clockwise order starting at NW. */
const int DIR_dx[] = {
    0,
   -1,  0,  1,  1,  1,  0, -1, -1,
   -2, -1,  0,  1,  2,  2,  2,  2,  2,  1,  0, -1, -2, -2, -2, -2
};
const int DIR_dy[] = {
    0,
   -1, -1, -1,  0,  1,  1,  1,  0,
   -2, -2, -2, -2, -2, -1,  0,  1,  2,  2,  2,  2,  2,  1,  0, -1
};

/* Compact names for directions                                       */
const char *DIR_dn[] = {
   "X",
   "NW", "N", "NE", "E", "SE", "S", "SW", "W",
   "NW2", "NNW", "N2", "NNE", "NE2", "ENE", "E2", "ESE",
   "SE2", "SSE", "S2", "SSW", "SW2", "WSW", "W2", "WNW"
};


/* array to rotational order mapping, rDir = DIR_ator[aDir]           */
const int DIR_ator[] = {
    0,
    1,  2,  3,  8,  4,  7,  6,  5,
    9, 10, 11, 12, 13, 24, 14, 23, 15, 22, 16, 21, 20, 19, 18, 17,
};

/* rotational to array order mapping, aDir = DIR_rtoa[rDir]           */
const int DIR_rtoa[] = {
    0,
    1,  2,  3,  5,  8,  7,  6,  4,
    9, 10, 11, 12, 13, 15, 17, 19, 24, 23, 22, 21, 20, 18, 16, 14,
};


/* city order to rotational mapping, rDir = DIR_ctor[cDir]            */
const int DIR_ctor[] = {
    0,                                                 /* centre      */
    2,  4,  6,  8,                                     /* cartesian 1 */
    1,  3,  5,  7,                                     /* diagonal  1 */
   11, 15, 19, 23,                                     /* cartesian 2 */
   10, 12, 14, 16, 18, 20, 22, 24,                     /* knights   2 */
    9, 13, 17, 21,                            /* diagonal  2 - unused */
};


/* octagonal to hex mapping, rDir = DIR_htor[hDir]                    */
/* hex_map distances, NW-SE are ring 2, WNW-NNW ring 3 and NW2 ring 4 */
const int DIR_hex[] = {
    0,                                                 /* centre      */
    2,  3,  4,  6,  7,  8,                             /* ring 1      */
   11, 12, 13, 14, 15,  5, 19, 20, 21, 22, 23,  1,     /* ring 2      */
   16, 18, 24, 10,                                     /* unused      */
   17,  9,                                             /* unused      */
};
#endif /* FC__CIVMAP_C */


/**
 * Direction enum for rotation order adjacent rings
 * Absolute positions and ring offset are listed for convenience.
 */
enum DirectionR {
  DIR_rUNKNOWN        = -1, /* not-a-direction */

  DIR_rCENTRE         =  0, /* centre tile */

  DIR_rNORTHWEST      =  0, /* first adjacent ring at offset 1 */
  DIR_rNORTH          =  1,
  DIR_rNORTHEAST      =  2,
  DIR_rEAST           =  3,
  DIR_rSOUTHEAST      =  4,
  DIR_rSOUTH          =  5,
  DIR_rSOUTHWEST      =  6,
  DIR_rWEST           =  7,

  DIR_rNORTHWEST2     =  0, /* second ring at offset 9 */
  DIR_rNORTHNORTHWEST =  1,
  DIR_rNORTH2         =  2,
  DIR_rNORTHNORTHEAST =  3,
  DIR_rNORTHEAST2     =  4,
  DIR_rEASTNORTHEAST  =  5,
  DIR_rEAST2          =  6,
  DIR_rEASTSOUTHEAST  =  7,
  DIR_rSOUTHEAST2     =  8,
  DIR_rSOUTHSOUTHEAST =  9,
  DIR_rSOUTH2         = 10,
  DIR_rSOUTHSOUTHWEST = 11,
  DIR_rSOUTHWEST2     = 12,
  DIR_rWESTSOUTHWEST  = 13,
  DIR_rWEST2          = 14,
  DIR_rWESTNORTHWEST  = 15,

  DIR_UNKNOWN         = -1, /* not-a-direction */

  DIR_CENTRE          =  0, /* centre tile */

  DIR_NORTHWEST       =  1, /* first adjacent ring at offset 1 */
  DIR_NORTH           =  2,
  DIR_NORTHEAST       =  3,
  DIR_EAST            =  4,
  DIR_SOUTHEAST       =  5,
  DIR_SOUTH           =  6,
  DIR_SOUTHWEST       =  7,
  DIR_WEST            =  8,

  DIR_NORTHWEST2      =  9, /* second ring at offset 9 */
  DIR_NORTHNORTHWEST  = 10,
  DIR_NORTH2          = 11,
  DIR_NORTHNORTHEAST  = 12,
  DIR_NORTHEAST2      = 13,
  DIR_EASTNORTHEAST   = 14,
  DIR_EAST2           = 15,
  DIR_EASTSOUTHEAST   = 16,
  DIR_SOUTHEAST2      = 17,
  DIR_SOUTHSOUTHEAST  = 18,
  DIR_SOUTH2          = 19,
  DIR_SOUTHSOUTHWEST  = 20,
  DIR_SOUTHWEST2      = 21,
  DIR_WESTSOUTHWEST   = 22,
  DIR_WEST2           = 23,
  DIR_WESTNORTHWEST   = 24
};


/**
 * Adjacent (or ring 1) utility macros 
 * - Cartesian tests, checked direction, reverse and bit reps
 */
#define DIR_IS_CARDINAL(dir) ((dir)&1)
#define DIR_CHECKED(dir)     ((dir)&7)
#define DIR_REVERSE(dir)     (dir^4)
#define DIR_BIT(dir)         (1<<DIR_CHECKED(dir))

#define DIR_CCW(dir)         DIR_CHECKED(dir-1)
#define DIR_CW(dir)          DIR_CHECKED(dir+1)

#define DIR_CCW_N(dir, n)    DIR_CHECKED(dir-n)
#define DIR_CW_N(dir, n)     DIR_CHECKED(dir+n)


/* These macros are used to step to adjacent tiles and normalize */
#define _map_step(px, py, dir)             \
( *(px) = DIR_dx[1+(dir)],                 \
  *(py) = DIR_dy[1+(dir)] )

#define civmap_step(map, px, py, x0, y0, dir, is_border)  \
( _map_step((px), (py), (dir)),                \
  *(px) += (x0), *(py) += (y0),                \
  (!(is_border) || normalize_civmap_pos((map), (px),(py))) )


/* These are the generalized forms for ring "n" */
#define _map_stepN(px, py, dir, nn)        \
( *(px) = DIR_dx[DIR_off[nn]+(dir)],       \
  *(py) = DIR_dy[DIR_off[nn]+(dir)] )

#define civmap_stepN(map, px, py, x0, y0, dir, nn, is_border) \
( _map_stepN((px), (py), (dir), (nn)),     \
  *(px) += (x0), *(py) += (y0),            \
  (!(is_border) || normalize_civmap_pos((map), (px),(py))) )


/**
 * Core adjacent iteration loop 
 *
 * Most users need only use the canned iterators below. Details here
 * are for those intrepid coders wanting to add new iterators.
 *
 * This returns a set of {x, y, dir} values based on the control settings,
 * (x, y) a normalized map position, dir in absolute form (e.g. DIR_WEST).
 * dir can be converted to/from a ring offset using DIR_off[ring]. Ring
 * "0" is both the center tile, and the global origin for multi ring dirs.
 * Unreal positions can be retrieved from the "} else {" clause.
 *
 * Controls:
 * map      - is a pointer to the map environment (dimensions, topology)
 * x0,y0    - is the centre map position
 * dtor     - maps an arbitrary direction set to absolute r-directions
 * n0,n1,ni - are the start, end and increment for the dtor index loop
 * nr       - is the number of the furthest tile ring for border checks
 *            if (x0,y0) is within nr of the border, normalization is on
 *            nr -ve runs as ring "0", for global/multi-ring iteration.
 *            nr +ve returns dir offsets relative to the start of ring nr.
 *
 * Notes:                                                     
 * 0)  x, y and dir are allocated and accessible within the loop only.
 *     Users should *not* use any other internal locals in their code.
 * 1)  reference arguments accessed once by local assignment  
 * 2)  border guard avoids full map checks except where an increment 
 *     could force an out of bounds condition. A NULL map or 0 distance
 *     turns off all border checks (see the code/macro definition).
 * 3)  if nr is negative, the border check is done and nr is set to 0
 *     i.e. directions are absolute and not ring offsets (see citymap).
 * 4)  combination of dtor and loop controls allows *any* iteration set
 *     or ordering with output {x,y,dir} still in standard map form.
 * 5)  when constant args are used, a good optimizing compiler will 
 *     completely eliminate significant chunks of macro codepaths.
 * 6)  this is a singly nested loop - "break" works!
 */
#define core_iterate(map, dtor, nr, n0, n1, ni, x0, y0, x, y, dir) {   \
  const CIVMAP *_map = (map);                                          \
  const int *_dtor = (dtor);                                           \
  int _dir = (nr);                                                     \
  const int  _n1 = (n1), _ni = (ni), _x0 = (x0), _y0 = (y0);           \
  const bool _border = civmap_is_borderN_pos(_map, _dir, _x0, _y0);    \
  const int  _nr = (_dir < 0 ? 0 : _dir);                              \
                                                                       \
  for( _dir = (n0); _dir < _n1; _dir += _ni ) {                        \
    int x, y, dir;                                                     \
    dir = (_dtor ? _dtor[DIR_off[_nr]+_dir]-DIR_off[nr] : _dir);       \
    if( civmap_stepN((map), &x, &y, _x0, _y0, dir, _nr, _border) ) {

#define core_iterate_end                                               \
    }                                                                  \
  }                                                                    \
}


/**
 * TODO: add tests and loop controls for hexagonal map iteration 
 *       there is only one cartesian/adjc form for the hex case. 
 * core_iterate((map), DIR_htor, 1,  7, 1, 1, (x0), (y0), x, y, _nn)
 * core_iterate((map), DIR_htor, 7, 19, 2, 2, (x0), (y0), x, y, _nn)
 * core_iterate((map), DIR_htor, 8, 19, 2, 2, (x0), (y0), x, y, _nn)
 * core_iterate((map), DIR_htor, 7, 19, 1, 2, (x0), (y0), x, y, _nn)
 * Note: 
 * 1)  hexagonal cities fill the full ring 2 set, right?      
 * 2)  backwards compatible 1st ring dir's, need to go absolute for hex
 */

/* Define { dtor, n0, n1, ni, nr } by iterator and topology with */
/* index function to select ???                                  */

/* Cartesian N, E, S, W iteration */
#define cartesian_adjc_iterate(map, x0, y0, x, y)                      \
  core_iterate((map), NULL, 1, 1, 8, 2, (x0), (y0), x, y, _nn)

#define cartesian_adjc_iterate_end                                     \
  core_iterate_end

/* Cartesian diagonal iteration NW, NE, SE, SW */
#define cartesian_diagonal_iterate(map, x0, y0, x, y)                  \
  core_iterate((map), NULL, 1, 0, 8, 2, (x0), (y0), x, y, _nn)

#define cartesian_diagonal_iterate_end                                 \
  core_iterate_end


/* full 1st ring iteration NW, N, ... W */
#define adjc_iterate(map, x0, y0, x, y)                                \
  core_iterate((map), NULL, 1, 0, 8, 1, (x0), (y0), x, y, _nn)

#define adjc_iterate_end                                               \
  core_iterate_end

/* full 1st ring iteration array order, i.e. NW -> SE */
#define adjc_array_iterate(map, x0, y0, x, y)                          \
  core_iterate((map), DIR_ator, 1, 0, 8, 1, (x0), (y0), x, y, _nn)

#define adjc_array_iterate_end                                         \
  core_iterate_end


/* Cartesian N, E, S, W iteration: 2nd ring*/
#define cartesian_adjc_iterate2(map, x0, y0, x, y)                     \
  core_iterate((map), NULL, 2, 2, 16, 4, (x0), (y0), x, y, _nn)

#define cartesian_adjc_iterate2_end                                    \
  core_iterate_end

/* Cartesian diagonal iteration: 2nd ring */
#define cartesian_diagonal_iterate2(map, x0, y0, x, y)                 \
  core_iterate((map), NULL, 2, 0, 16, 4, (x0), (y0), x, y, _nn)

#define cartesian_diagonal_iterate2_end                                \
  core_iterate_end

/* Cartesian knights iteration: 2nd ring */
#define cartesian_knights_iterate2(map, x0, y0, x, y)                  \
  core_iterate((map), NULL, 2, 1, 16, 2, (x0), (y0), x, y, _nn)

#define cartesian_knights_iterate2_end                                 \
  core_iterate_end


/* full iteration: 2nd ring */
#define adjc_iterate2(map, x0, y0, x, y)                               \
  core_iterate((map), NULL, 2, 0, 16, 1, (x0), (y0), x, y, _nn)

#define adjc_iterate2_end                                              \
  core_iterate_end


/* Cartesian N, E, S, W iteration */
#define cartesian_adjc_dir_iterate(map, x0, y0, x, y, dir)             \
  core_iterate((map), NULL, 1, 1, 8, 2, (x0), (y0), x, y, dir)

#define cartesian_adjc_dir_iterate_end                                 \
  core_iterate_end

/* Cartesian diagonal iteration NW, NE, SE, SW */
#define cartesian_diagonal_dir_iterate(map, x0, y0, x, y, dir)         \
  core_iterate((map), NULL, 1, 0, 8, 2, (x0), (y0), x, y, dir)

#define cartesian_diagonal_dir_iterate_end                             \
  core_iterate_end


/* full 1st ring iteration NW, N, ... W */
#define adjc_dir_iterate(map, x0, y0, x, y, dir)                       \
  core_iterate((map), NULL, 1, 0, 8, 1, (x0), (y0), x, y, dir)

#define adjc_dir_iterate_end                                           \
  core_iterate_end

/* full 1st ring iteration array order, i.e. NW -> SE */
#define adjc_array_dir_iterate(map, x0, y0, x, y, dir)                 \
  core_iterate((map), DIR_ator, 1, 0, 8, 1, (x0), (y0), x, y, dir)

#define adjc_array_dir_iterate_end                                     \
  core_iterate_end


/* Cartesian N, E, S, W iteration: 2nd ring */
#define cartesian_adjc_dir_iterate2(map, x0, y0, x, y, dir)            \
  core_iterate((map), NULL, 2, 2, 16, 4, (x0), (y0), x, y, dir)

#define cartesian_adjc_dir_iterate2_end                                \
  core_iterate_end

/* Cartesian diagonal iteration: 2nd ring */
#define cartesian_diagonal_dir_iterate2(map, x0, y0, x, y, dir)        \
  core_iterate((map), NULL, 2, 0, 16, 4, (x0), (y0), x, y, dir)

#define cartesian_diagonal_dir_iterate2_end                            \
  core_iterate_end

/* Cartesian knights iteration: 2nd ring */
#define cartesian_knights_dir_iterate2(map, x0, y0, x, y, dir)         \
  core_iterate((map), NULL, 2, 1, 16, 2, (x0), (y0), x, y, dir)

#define cartesian_knights_dir_iterate2_end                             \
  core_iterate_end


/* full iteration: 2nd ring */
#define adjc_dir_iterate2(map, x0, y0, x, y, dir)                      \
  core_iterate((map), NULL, 2, 0, 16, 1, (x0), (y0), x, y, dir)

#define adjc_dir_iterate2_end                                          \
  core_iterate_end


/* citymap checked iteration over rings 0, 1, 2: nr = -2 for checks */
/* Note: map_stepN with nr = 0 is just a dir_step.                  */
#define citymap_checked_iterate(map, x0, y0, cx, cy, mx, my) {         \
  int cx, cy, _xo= CITY_MAP_SIZE/2, _yo= CITY_MAP_SIZE/2;              \
  core_iterate((map), DIR_ctor, -2, 0, 21, 1, (x0), (y0), mx, my, _nn) \
    civmap_stepN(_map, &cx, &cy, _xo, _yo, _nn, 0, 0);

#define citymap_checked_iterate_end                                    \
  core_iterate_end                                                     \
}

/* citymap iteration over rings 1, 2 (i.e. omit city centre)        */
/* Note: map_stepN with nr = 0 is just a dir_step.                  */
#define citymap_checked_iterate_1(map, x0, y0, cx, cy, mx, my) {       \
  int cx, cy, _xo= CITY_MAP_SIZE/2, _yo= CITY_MAP_SIZE/2;              \
  core_iterate((map), DIR_ctor, -2, 1, 21, 1, (x0), (y0), mx, my, _nn) \
    civmap_stepN(_map, &cx, &cy, _xo, _yo, _nn, 0, 0);

#define citymap_checked_iterate_1_end                                  \
  core_iterate_end                                                     \
}


/* citymap iteration over rings 0, 1, 2: nr = 0 -> no checks          */
/* Technically this is no longer a map iteration - no map dependency. */
#define city_map_iterate(cx, cy) {                                     \
  const int _xo= CITY_MAP_SIZE/2, _yo= CITY_MAP_SIZE/2;                \
  core_iterate(NULL, DIR_ctor, 0, 0, 21, 1, _xo, _yo, cx, cy, _nn)

#define city_map_iterate_end                                           \
  core_iterate_end                                                     \
}

/* citymap iteration over rings 1, 2 (i.e. omit city centre)          */
/* Technically this is no longer a map iteration - no map dependency. */
#define city_map_iterate_1(cx, cy) {                                   \
  const int _xo= CITY_MAP_SIZE/2, _yo= CITY_MAP_SIZE/2;                \
  core_iterate(NULL, DIR_ctor, 0, 1, 21, 1, _xo, _yo, cx, cy, _nn)

#define city_map_iterate_1_end                                         \
  core_iterate_end                                                     \
}

/* citymap radius iteration over rings 0, 1, 2 - centered on (0,0)    */
/* Technically this is no longer a map iteration - no map dependency. */
#define city_radius_iterate(cx, cy)                                    \
  core_iterate(NULL, DIR_ctor, 0, 0, 21, 1, 0, 0, cx, cy, _nn)

#define city_radius_iterate_end                                        \
  core_iterate_end


/* citymap array iteration over rings 0, 1, 2: nr = 0 -> no checks    */
#define city_map_array_checked_iterate(map, x0, y0, cx, cy, mx, my) {  \
  int cx, cy, _xo= CITY_MAP_SIZE/2, _yo= CITY_MAP_SIZE/2;              \
  core_iterate((map), DIR_cator, -2, 0, 21, 1, (x0), (y0), mx, my, _nn)\
    civmap_stepN(_map, &cx, &cy, _xo, _yo, _nn, 0, 0);

#define city_map_array_checked_iterate_end                             \
  core_iterate_end                                                     \
}

/* citymap array iteration over rings 0, 1, 2: nr = 0 -> no checks    */
/* Technically this is no longer a map iteration - no map dependency. */
#define city_array_iterate(cx, cy)                                     \
  const int _xo= CITY_MAP_SIZE/2, _yo= CITY_MAP_SIZE/2;                \
  core_iterate(NULL, DIR_cator, 0, 0, 21, 1, _xo, _yo, cx, cy, _nn)

#define city_map_iterate_1_end                                         \
  core_iterate_end                                                     \
}


/* square iteration over rings 0, 1, 2 */
#define square_iterate2(map, x0, y0, x, y, dir)                        \
  core_iterate((map), NULL, -2, 0, 16, 1, (x0), (y0), x, y, dir)

#define square_iterate2_end                                            \
  core_iterate_end


/* 
 * Generic 2n+1 square array iteration (within map distance n)
 *
 * Square iteration is mostly done to map a fixed distance local area.
 * Since distance is calculated in standard map coordinates, even for
 * isometric maps, results must be checked for realness + normalized.
 *
 * A negative distance or map NULL turns off all map normalization
 * checks, as when iterating a 3x3 citymap display including corners.
 * Thus this becomes a simple loop.
 *   square_iterate(NULL, CITY_MAP_SIZE/2, CITY_MAP_SIZE/2, 2, cx, cy)
 *
 * This also uses a single loop (break works).
 */
#define square_iterate(map, x0, y0, nn, x1, y1) {                      \
  const CIVMAP *_map = (map);                                          \
  int x1, y1 = (nn), _x0 = (x0), _y0 = (y0);                           \
  const bool _border = (y1 > 0 && civmap_is_borderN_pos(_map, y1, _x0, _y0)); \
  const int  _nn = (y1 < 0 ? -y1 : y1);                                \
  const int  _yn = _y0 + _nn, _xn = _x0 + _nn;                         \
  _y0 = _y0 - _nn, _x0 = _x0 - _nn - 1;                                \
  while( 1 ) {                                                         \
    if( ++_x0 > _xn ) {                                                \
      if( ++_y0 > _yn )                                                \
        break;                                                         \
      _x0 = _xn - _nn - _nn;                                           \
    }                                                                  \
    x1 = _x0, y1 = _y0;                                                \
    if( !_border || normalize_civmap_pos(_map, &x1, &y1) ) {

#define square_iterate_end                                             \
    }                                                                  \
  }                                                                    \
}


/**
 * Optimized whole map iterate (counts down the linear index with
 * conversion to standard map coordinates).
 *
 * It uses a single loop (on the linear offset xy = map->xsize*y + x).
 */
#define whole_civmap_iterate(map, mx, my)                              \
  whole_civmap_xy_iterate(map, _xy, _x, _y, mx, my)

#define whole_civmap_iterate_end                                       \
  whole_civmap_xy_iterate_end


/* Expose native and index coordinates as well as standard map (x,y). */
#define whole_civmap_xy_iterate(map, xy, xn, yn, mx, my) {             \
  int mx, my, xn, yn, xy= map->ysize*map->xsize;                       \
  while( --xy >= 0 ) {                                                 \
    if( index_to_civmap_pos(map, &mx, &my, &xn, &yn, xy) ) {

#define whole_civmap_xy_iterate_end                                    \
    }                                                                  \
  }                                                                    \
}


/*
 * block iteration backgrounder
 *
 * First, block iteration is valid *only* in native coordinates.
 *
 *  +------+
 *  | 1+++ |       - flat-like coords always reduce to this
 *  | ++++ |         if point coordinates or deltas are reversed
 *  | ++++ |         they are flipped, rather than aborting.
 *  | +++2 |
 *  +------+
 *
 *  +------+
 *  |++  1+|++     - x-wrap coords straddle the xsize wall
 *  |++  ++|++       there is still only one logical region
 *  |++  ++|++       when the storage wraps in x as well.
 *  |+2  ++|+2
 *  +------+
 *
 *  +------+
 *  | +++2 |       - y-wrap coords straddle the ysize wall
 *  |      |         with 2 regions to iterate
 *  |      |
 *  | 1+++ |
 *  +------+
 *    +++2
 *
 *  +------+
 *  |+2 ---|       - xy-wrap coords exist in 4 quarter regions
 *  |      |         with 2 regions to iterate,
 *  |-- 1++|--
 *  |-- +++|--
 *  +------+--
 *      --- +2
 *
 * block_iterate(map, x,y,x0,y0,dx,dy) loops the user variables x and y
 * block_xy_iterate(map, xy,x,y,x0,y0,dx,dy) loops user variables x, y, xy
 *   over the tiles in the rectangular block with corner x0, y0 and
 *   dimensions dx, dy. Reasonable argument checking is done.
 *
 * block_iterate2(map, x,y,x0,y0,x1,y1) loops the user variables x and y
 * block_xy_iterate2(map, xy,x,y,x0,y0,x1,y1) loops user variables x, y, xy
 *   over the tiles in the rectangular block defined by the two points
 *   x0, y0 and x1, y1. Reasonable argument checking is done.
 *
 * fastblock_iterate(xy,x,y,xy1,xy0,xsize,xs,xm,ys,ym) is the unchecked 
 *   core loop from point xy1 to xy0 with skips of xs whenever x equals xm, 
 *   and skip of ys plus region skip reset whenever xy < ym.
 *
 * *** WARNING: If you cannot figure out what block_iterate does,
 *              do not use fastblock_iterate directly.
 *
 * All allocate the user's chosen x, y, xy variables within the (core) loop.
 * All treat arguments other than x, y, xy as single use expressions.
 * All use a single loop on the linear offset xy = map-xsize*y + x.
 * All blocks will be correctly wrapped in the x or y dimension at the
 *   appropriate map size boundary as required by map_type, or off map
 *   values will be skipped.
 * All values are real and normalized by default (i.e. within the native
 *   space). There is thus no need for expensive checking.
 * All values are ordered in reverse array order (for GUI tiling).
 *
 * Depends on map->xsize, map->ysize, civmap_has_wrap()
 */
#define fastblock_iterate(xy, x, y, xy1, xy0, xsize, xs, xm, ys, ym) {     \
  int x;                                                                   \
  int y;                                                                   \
  int xy = (xy1);                                                          \
  int _xy0 = (xy0);                                                        \
  int _xsize = (xsize);                                                    \
  int _xm = (xm);                                                          \
  int _xs = (xs);                                                          \
  int _ym = (ym);                                                          \
  int _ys = (ys);                                                          \
                                                                           \
  /* loop from upper xy1 to lower xy0 point, skip xs at xm and ys at ym */ \
  /* printf("fastblock: %d, %d, %d, %d. %d. %d \n",                        \
       xy, _xy0, _xs, _xm, _ys, _ym);  for debug */                        \
  while (xy-- > _xy0) {                                                    \
    if (xy < _ym) {                                                        \
      /* skip to region 2 - onetime only */                                \
      _ym = 0;                                                             \
      xy -= _ys;                                                           \
      if (xy < 0)                                                          \
        break;                                                             \
    }                                                                      \
    y = xy / _xsize;                                                       \
    x = xy % _xsize;                                                       \
    if (x == _xm) {                                                        \
      /* skip within a line */                                             \
      xy -= _xs;                                                           \
      continue;                                                            \
    }                                                                      \
    /* printf("fastblock: %d. %d. %d\n", xy, x, y); for debug */

#define fastblock_iterate_end                                              \
  }                                                                        \
}


/* 
 * Special cases for multiple map topologies has probably pushed this to
 * its conceptual limits, macros should never exceed 1 page in size :-).
 */
#define block_iterate(map, x, y, x0, y0, dx, dy)                              \
  block_xy_iterate(map, _xy, x, y, (x0), (y0), (dx), (dy))

#define block_iterate_end                                                     \
  block_xy_iterate_end


#define block_xy_iterate(map, xy, x, y, x0, y0, dx, dy) {                     \
  const int _xo = (x0), _yo = (y0);                                           \
  block_xy_iterate2(map, xy, x, y, _xo, _yo, (_xo+(dx)-1), (_yo+(dy)-1))

#define block_xy_iterate_end                                                  \
  block_xy_iterate2_end                                                       \
}


#define block_iterate2(map, x, y, x0, y0, x1, y1)                             \
  block_xy_iterate2(map, _xy, x, y, (x0), (y0), (x1), (y1))

#define block_iterate2_end                                                    \
  block_xy_iterate2_end


#define block_xy_iterate2(map, xy, x, y, x0, y0, x1, y1) {                    \
  const CIVMAP *_map = (map);                                                 \
  const int _usize = _map->xsize, _vsize = _map->ysize;                       \
  int _um= -1, _us= -1, _vm= -1, _vs= -1;                                     \
  int _x0= (x0), _x1= (x1), _y0= (y0), _y1= (y1);                             \
  /* printf("block_iterate2  %d, %d, %d, %d\n", _x0, _y0, _x1, _y1); */       \
  /* Y dimension                                                           */ \
  if( _y0 > _y1 ) _y1^= _y0, _y0^= _y1, _y1^= _y0;      /* fix -ve delta Y */ \
  if( civmap_has_wrap(_map, MAP_TYPE_WRAPY) ) {               /* WRAP in Y */ \
    /* _y1= y1+1 simplifies logic here                                     */ \
    ++_y1, _y1= WRAP_0(_vsize,_y1), _y0= WRAP_0(_vsize,_y0);   /* sanitize */ \
    if( _y0 == _y1 ) {                                                        \
      _y0= 0, _y1= _vsize;                            /* full map, no skip */ \
    } else                                                                    \
    if( _y0 > _y1 )                           /* 2 regions, top and bottom */ \
      _vm= _y0*_usize, _vs=(_y0-_y1)*_usize, _y1= _vsize, _y0= 0;             \
    _y1--;                                                                    \
  } else {                                                    /* FLAT in Y */ \
    _y1= CLIP(0,_y1,_vsize-1), _y0= CLIP(0,_y0,_vsize-1);      /* sanitize */ \
  }                                                                           \
  /* X dimension                                                           */ \
  if( _x0 > _x1 ) _x1^= _x0, _x0^= _x1, _x1^= _x0;      /* fix -ve delta X */ \
  if( civmap_has_wrap(_map, MAP_TYPE_WRAPX) ) {               /* WRAP in X */ \
    /* _x1+1 compensates for initial loop decrement, simplifies logic here */ \
    ++_x1, _x1= WRAP_0(_usize,_x1), _x0= WRAP_0(_usize,_x0);   /* sanitize */ \
    if( _x0 == _x1 ) {                                                        \
      _x0= 0, _x1= _usize;                            /* full map, no skip */ \
    } else                                                                    \
    if( _x0 > _x1 ) {                           /* start top, skip midline */ \
      _us= _x0-_x1-1, _um= _x0-1, _x0= 0, _x1= _usize;         /* full map */ \
    } else {                           /* start midline, skip low and wrap */ \
      _us= _x0-_x1 + _usize-1;                                                \
      if( -1 < _us )                                  /* full map, no skip */ \
        _um= (_x0 <= 0 ? _usize-1 : x0-1);           /* mark skip, 1st OOB */ \
    }                                                                         \
  } else {                                                    /* FLAT in X */ \
    _x1= CLIP(0,_x1,_usize-1), _x0= CLIP(0,_x0,_usize-1);      /* sanitize */ \
    _x1++;                        /* compensate for initial loop decrement */ \
    _us= _x0-_x1 + _usize-1;           /* start midline, skip low and wrap */ \
    if( -1 < _us )                                    /* full map, no skip */ \
      _um= (_x0 <= 0 ? _usize-1 : x0-1);             /* mark skip, 1st OOB */ \
  }                                                                           \
  /* printf("block_iterate2: %d, %d, %d, %d, %d, %d, %d, %d\n",               \
    _x0, _y0, _x1, _y1, _us, _um, _vs, _vm);                 */               \
  _y1= _usize*_y1+_x1, _y0= _usize*_y0+_x0;                                   \
  /* Note: _args carefully do not overlap with fastblock_iterate tempvars  */ \
  fastblock_iterate(xy, x, y, _y1, _y0, _usize, _us, _um, _vs, _vm)

#define block_xy_iterate2_end                                                 \
  fastblock_iterate_end                                                       \
}


/*                     :
 * +-++-------------+  :  To the left is the pictorial N-ccw walk
 * |6||5  5  5  5  5|  :  of an outward spiral, as in walk -1y, -1x,
 * | |+-++-------++-+  :  +2y, +2x, ... changing the sign of the xy
 * |6||4||3  3  3||5|  :  increment and the length of the walk at
 * | || |+-++-++-+| |  :  the beginning of each y sequence.
 * |6||4||2||1||3||5|  :
 * | || || |+-+| || |  :  The walk spins when y runs off map, skips
 * |6||4||2||1||3||5|  :  y-walk when length > map->xsize and exits
 * | || |+-++-+| || |  :  when there are no map squares untouched.
 * |6||4||2  2||3||5|  :  Step increment below precedes the walk
 * | |+-++----++-+| |  :  sequence reset otherwise everything would
 * |6||4  4  4  4||5|  :  be shifted around by 1 finishing each leg
 * +-++----------++-+  :  1 into the next square.
 *                     :
 * MACRO Usage:
 *  spiral_outward(map, x, y, x0, y0, NWSE, USE_00, condition)
 *    <user code at coordinates (x,y)>
 *  spiral_outward_end
 *
 * Where:
 * map        - pointer to the map environment (topology, dimensions, etc)
 * x,y        - user supplied index variables allocated within the loop
 * x0,y0      - initial coordinates
 * NWSE       - spiral rotation and start direction
 *              N(1),W(2),S(3),E(4), -ve values for clockwise rotation
 * USE_00     - boolean, set TRUE if x,y should be included in the walk
 * condition  - loop termination condition, e.g. (--dist > 0) or (1)
 *
 * Dependencies:
 * map->xsize, map->ysize and civmap_has_wrap()
*/
#define spiral_outward(map, x, y, x0, y0, NWSE, USE_00, COND)                \
{                                                                            \
  CIVMAP *_map = (map);                                                      \
  const int _xsize = _map->xsize, _ysize = _map->ysize;                      \
  int x = (x0), y = (y0), _dir = (NWSE);                                     \
  int _last, _ccw, _inc, _new, _cnt = 0, _len = 0;                           \
                                                                             \
  _new = (!civmap_has_wrap(_map, MAP_TYPE_WRAPY)      /* largest diameter */ \
       ? (_ysize - y > y ? _ysize - y : y) * 2 : _ysize);                    \
  _last = (!civmap_has_wrap(_map, MAP_TYPE_WRAPX)                            \
        ? (_xsize - x > x ? _xsize - x : x) * 2 : _xsize);                   \
  if (_last < _new)                                                          \
    _last = _new;                                                            \
                                                                             \
  _ccw = (_dir < 0) ? (_dir = -_dir, 0) : 1;     /* parse NWSE, set start */ \
  _inc = (_dir > 2) ? 1 : -1;                    /* NW corner is at 0,0   */ \
  _new = _dir &= 1;                              /* remember starting dir */ \
  _ccw ^= _dir;                               /* x->y clockwise, y->x ccw */ \
                                                                             \
  if (USE_00) {                         /* backup for USE_00 special case */ \
    if ((_dir ^ _ccw) == _new)                                               \
      _inc = -_inc;                                                          \
    if (_dir ^= 1)                                                           \
      y -= _inc;                                                             \
    else                                                                     \
      x -= _inc;                                                             \
    _len--;                                                                  \
  }                                                                          \
                                                                             \
  while (COND) {                                                             \
    if (_dir)                            /* actually the trailing step of */ \
      y += _inc;                         /* last cycle into a new square  */ \
    else                                                                     \
      x += _inc;                                                             \
                                                                             \
    if (--_cnt < 0) {   /* Reset _dir _len _inc _cnt after walking a side */ \
      if ((_dir ^= 1) != _new) {                             /* flip _dir */ \
        if (_len >= _last)    /* 1st _len completes a square, exit on 2nd */ \
          break;                                                             \
      } else                                                                 \
        ++_len;                               /* new cycle every 2nd pass */ \
                                                                             \
      if ((_dir ^ _ccw) == _new)                                             \
        _inc = -_inc;                         /* ccw is in phase y, out x */ \
                                                                             \
      _cnt = _len;                                                           \
      if (_dir) {                                  /* fixup cnt: y=1, x=0 */ \
        if( civmap_has_wrap(_map, MAP_TYPE_WRAPY) && _len >= _ysize)         \
          _cnt = _ysize - 1;            /* new y, watch for wrapwall */      \
        /* TODO: if WrapX && len > xsize/2 OR !WrapX && unsigned x >= xsize  \
                 y+= _cnt*_inc; _cnt = 0; continue;  *** Optimization */     \
      } else {                                                               \
        if( civmap_has_wrap(_map, MAP_TYPE_WRAPX) && _len >= _xsize)         \
          _cnt = _xsize - 1;            /* new x, watch for wrapwall */      \
        /* TODO: if WrapY && len > ysize/2 OR !WrapY && unsigned y >= ysize  \
                 x+= _cnt*_inc; _cnt = 0; continue;  *** Optimization */     \
      }                                                                      \
    }                                                                        \
    /* printf("spiral_outward: %3,3d,%-3.3d \n", x, y); for debug */         \
    if (normalize_civmap_pos(_map, &x, &y)) {

#define spiral_outward_end                                                   \
    }                                                                        \
  }                                                                          \
}


/* Replacement for the original iterate_outward macro which looped
 * thru top/bottom/left/right of increasing squares from the start
 * values outwards to a given distance.
 *
 * Note:
 *   Spiral_outward in addition to providing a more varied walk
 * and set of termination conditions, is also a single while block
 * from which user code can break out. This applies to the
 * replacement macro here as well. The normal iteration step uses
 * fewer active varibles and is probably marginally faster under
 * most compilers.
 *   Usage of iterate_outward should be considered deprecated.
*/
#define iterate_outward(map, x0, y0, dist, x, y)                   \
{                                                                  \
  int _dist= (dist)*4;  /* 4 sides to a square walk */             \
  spiral_outward(map, x, y, x0, y0, 1, 1, (_cnt != 0 || _dist-- >= 0))

#define iterate_outward_end                                        \
  spiral_outward_end                                               \
}

#endif  /* FC__CIVMAP_H */

