/**********************************************************************
 Freeciv - Copyright (C) 1996 - A Kjeldberg, L Gregersen, P Unold
 civmap_ex.c - Copyright (C) 2002 - R. Wetmore

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
***********************************************************************/

#define FC__CIVMAP_EX_C

/**
 * civmap_ex.c - extended map operations
 *
 * Extended functions for accessing and updating CIVMAP positions,
 * in particular CIVTILE and CIVTERRAIN accessor convenience methods.
 *
 * Handles global effects such as tile connectivity and adjacent
 * update operations.
 *
 * The rough ordering is one of increasing complexity.
 * 1)  simple (often convenience) accessor methods
 * 2)  more complex but still single position accessor methods
 * 3)  complex init and reset methods
 * 4)  runtime actions and updates
 * 5)  miscellaneous (multitile) info and summary methods
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "log.h"
#include "mem.h"
#include "rand.h"
#include "support.h"

#include "civmap_ex.h"


/***********************************************************************
 Return a pointer to the tile at the normalized map position (x,y).
 This is used extensively and is typically optimized to a macro.
***********************************************************************/
#undef civmap_get_tile
CIVTILE *civmap_get_tile(const CIVMAP *pmap, int x, int y)
{
  return (((CIVTILE *)pmap->tiles) + civmap_pos_to_index(pmap, x, y));
}

/***********************************************************************
 Return a reference to the city at this position, or NULL if none.
***********************************************************************/
struct city *civmap_get_city(const CIVMAP *pmap, int x, int y)
{
  return civmap_get_tile(pmap, x, y)->city;
}

/***********************************************************************
 Reset the reference to the city at this position.
***********************************************************************/
void civmap_set_city(const CIVMAP *pmap, int x, int y, struct city *pcity)
{
  civmap_get_tile(pmap, x, y)->city = pcity;
}

/***********************************************************************
 Return a reference to the city working this position, or NULL if none.
***********************************************************************/
struct city *civmap_get_worked(const CIVMAP *pmap, int x, int y)
{
  return civmap_get_tile(pmap, x, y)->worked;
}

/***********************************************************************
 Reset the reference to the city working this position.
***********************************************************************/
void civmap_set_worked(const CIVMAP *pmap, int x, int y, struct city *pcity)
{
  civmap_get_tile(pmap, x, y)->worked = pcity;
}

/***********************************************************************
 Convenience access method for maps.
***********************************************************************/
short civmap_get_continent(const CIVMAP *pmap, int x, int y)
{
  return civmap_get_tile(pmap, x, y)->continent;
}

/***********************************************************************
 Convenience access method for maps.
***********************************************************************/
void civmap_set_continent(const CIVMAP *pmap, int x, int y, int val)
{
  civmap_get_tile(pmap, x, y)->continent = val;
}

/***********************************************************************
 Convenience access method for maps.
***********************************************************************/
short civmap_get_ocean(const CIVMAP *pmap, int x, int y)
{
  return civmap_get_tile(pmap, x, y)->ocean;
}

/***********************************************************************
 Convenience access method for maps.
***********************************************************************/
void civmap_set_ocean(const CIVMAP *pmap, int x, int y, int val)
{
  civmap_get_tile(pmap, x, y)->ocean = val;
}


/***********************************************************************
 Return a basic goodness value. Convenience method for map positions
***********************************************************************/
int civmap_goodness(const CIVMAP *pmap, int x, int y)
{
  return civtile_goodness(civmap_get_tile(pmap, x, y));
}

/***********************************************************************
 Return the basic food value. Convenience method for map positions
***********************************************************************/
int civmap_get_food_base(const CIVMAP *pmap, int x, int y)
{
  return civtile_get_food_base(civmap_get_tile(pmap, x, y));
}

/***********************************************************************
 Return the basic shield value. Convenience method for map positions
***********************************************************************/
int civmap_get_shield_base(const CIVMAP *pmap, int x, int y)
{
  return civtile_get_shield_base(civmap_get_tile(pmap, x, y));
}

/***********************************************************************
 Return the basic trade value. Convenience method for map positions
***********************************************************************/
int civmap_get_trade_base(const CIVMAP *pmap, int x, int y)
{
  return civtile_get_trade_base(civmap_get_tile(pmap, x, y));
}

/***********************************************************************
 Return a (static) string with a tile's food/prod/trade
***********************************************************************/
const char *civmap_get_fpt_text(const CIVMAP *pmap, int x, int y)
{
  return civtile_get_fpt_text(civmap_get_tile(pmap, x, y));
}

/***********************************************************************
 Return a (static) string with terrain info; Convenience method for maps.
 eg: "Hills"
 eg: "Hills (Coals)"
 eg: "Hills (Coals) [Pollution]"
***********************************************************************/
const char *civmap_get_info_text(const CIVMAP *pmap, int x, int y)
{
  return civtile_get_info_text(civmap_get_tile(pmap, x, y));
}

/***********************************************************************
 Convenience access method for maps.
***********************************************************************/
int civmap_road_time(const CIVMAP *pmap, int x, int y)
{
  return civtile_road_time(civmap_get_tile(pmap, x, y));
}

/***********************************************************************
 Convenience access method for maps.
***********************************************************************/
int civmap_railroad_time(const CIVMAP *pmap, int x, int y)
{
  return civtile_railroad_time(civmap_get_tile(pmap, x, y));
}

/***********************************************************************
 Convenience access method for maps.
***********************************************************************/
int civmap_mine_time(const CIVMAP *pmap, int x, int y)
{
  return civtile_mine_time(civmap_get_tile(pmap, x, y));
}

/***********************************************************************
 Convenience access method for maps.
***********************************************************************/
int civmap_irrigation_time(const CIVMAP *pmap, int x, int y)
{
  return civtile_irrigation_time(civmap_get_tile(pmap, x, y));
}

/***********************************************************************
 Convenience access method for maps.
***********************************************************************/
int civmap_transform_time(const CIVMAP *pmap, int x, int y)
{
  return civtile_transform_time(civmap_get_tile(pmap, x, y));
}

/***********************************************************************
 Convenience access method for maps.
***********************************************************************/
int civmap_fortress_time(const CIVMAP *pmap, int x, int y)
{
  return civtile_fortress_time(civmap_get_tile(pmap, x, y));
}

/***********************************************************************
 Convenience access method for maps.
***********************************************************************/
int civmap_airbase_time(const CIVMAP *pmap, int x, int y)
{
  return civtile_airbase_time(civmap_get_tile(pmap, x, y));
}

/***********************************************************************
 Convenience access method for maps.
***********************************************************************/
int civmap_pollution_time(const CIVMAP *pmap, int x, int y)
{
  return civtile_pollution_time(civmap_get_tile(pmap, x, y));
}

/***********************************************************************
 Convenience access method for maps.
***********************************************************************/
int civmap_fallout_time(const CIVMAP *pmap, int x, int y)
{
  return civtile_fallout_time(civmap_get_tile(pmap, x, y));
}

#ifdef	FC__UNIT_H
/***********************************************************************
 Time to complete the given activity on the given map position.
***********************************************************************/
int civmap_activity_time(const CIVMAP *pmap, int x, int y, 
  enum unit_activity activity)
{
  return civtile_activity_time(civmap_get_tile(pmap, x, y), activity);
}
#endif	FC__UNIT_H


/***********************************************************************
 Return the special set at the given map position.
 Convenience access method for maps.
***********************************************************************/
enum special_type civmap_get_specials(const CIVMAP *pmap, int x, int y)
{
  return civmap_get_tile(pmap, x, y)->special;
}

/***********************************************************************
 Returns TRUE iff the given map position has the given special set.
***********************************************************************/
bool civmap_has_specials(const CIVMAP *pmap, int x, int y, 
  enum special_type spe)
{
  return civtile_has_specials(civmap_get_tile(pmap, x, y), spe);
}

/***********************************************************************
 Update the special set at the given map position and handle resets.
 Convenience access method for maps.
***********************************************************************/
void civmap_set_specials(const CIVMAP *pmap, int x, int y, 
  enum special_type spe)
{
  civmap_get_tile(pmap, x, y)->special = spe;

  if ((spe & (S_ROAD | S_RAILROAD)) != 0)
    civmap_movecosts_reset(pmap, x, y);
}

/***********************************************************************
 Add the special at the given map position and handle resets.
 Convenience access method for maps.
***********************************************************************/
void civmap_special_add(const CIVMAP *pmap, int x, int y, enum special_type spe)
{
  civtile_special_add(civmap_get_tile(pmap, x, y), spe);

  if ((spe & (S_ROAD | S_RAILROAD)) != 0)
    civmap_movecosts_reset(pmap, x, y);
}

/***********************************************************************
 Delete the special at the given map position and handle resets.
 Convenience access method for maps.
***********************************************************************/
void civmap_special_delete(const CIVMAP *pmap, int x, int y, enum special_type spe)
{
  civtile_special_delete(civmap_get_tile(pmap, x, y), spe);

  if ((spe & (S_ROAD | S_RAILROAD)) != 0)
    civmap_movecosts_reset(pmap, x, y);
}

/***********************************************************************
 Delete the special set at the given map position and handle resets.
 Convenience access method for maps.
***********************************************************************/
void civmap_infrastructure_delete(const CIVMAP *pmap, int x, int y)
{
  civtile_infrastructure_delete(civmap_get_tile(pmap, x, y));
  civmap_movecosts_reset(pmap, x, y);
}

/***********************************************************************
 Delete the special set at the given map position.
 Convenience access method for maps.
***********************************************************************/
void civmap_dirtiness_delete(const CIVMAP *pmap, int x, int y)
{
  civtile_dirtiness_delete(civmap_get_tile(pmap, x, y));
}


/***********************************************************************
 Convenience access method for maps.
***********************************************************************/
enum terrain_type civmap_get_terrain(const CIVMAP *pmap, int x, int y)
{
  return civmap_get_tile(pmap, x, y)->terrain;
}
/***********************************************************************
 Convenience access method for maps.
***********************************************************************/
bool civmap_has_terrain(const CIVMAP *pmap, int x, int y, enum terrain_type ter)
{
  return civtile_has_terrain(civmap_get_tile(pmap, x, y), ter);
}

/***********************************************************************
 Convenience access method for maps.
***********************************************************************/
void civmap_set_terrain(const CIVMAP *pmap, int x, int y, enum terrain_type ter)
{
  enum terrain_type tt = civmap_get_tile(pmap, x, y)->terrain;
  civtile_set_terrain(civmap_get_tile(pmap, x, y), ter);
  civmap_terrain_changed(pmap, x, y, tt);
}


/***********************************************************************
 The basic cost to move punit from tile t1 to tile t2.
 That is, tile_movecosts(), with pre-calculated tile pointers;
 the tiles are assumed to be adjacent, and the is_cardinal
 flag is used only for the river bonus.

 May also be used with punit==NULL for unit-independent results,
 in which case punit specific tests are not done.
***********************************************************************/
static int tile_movecosts(struct unit *punit, 
  CIVTILE *t1, CIVTILE *t2, bool is_cardinal)
{
#ifdef FC__UNIT_H
  /* This does not adequately handle naval units, right??? */
  if (punit && !is_ground_unit(punit))
    return SINGLE_MOVE;
#endif

  if (civtile_has_specials(t1, S_RAILROAD)
   && civtile_has_specials(t2, S_RAILROAD))
    return MOVECOST_RAIL;

#ifdef FC__UNIT_H
  if (punit && unit_flag(punit, F_IGTER))
    return (SINGLE_MOVE+2)/3;
#endif

  if (civtile_has_specials(t1, S_ROAD)
   && civtile_has_specials(t2, S_ROAD))
    return MOVECOST_ROAD;

  if ((civtile_has_specials(t1, S_RIVER)
    && civtile_has_specials(t2, S_RIVER))) {
    switch (t2->global->terrain->river_move_mode) {
    case RMV_NORMAL:
      break;
    case RMV_FAST_STRICT:
      if (is_cardinal)
	return MOVECOST_RIVER;
      break;
    case RMV_FAST_RELAXED:
      if (is_cardinal)
	return MOVECOST_RIVER;
      else
	return 2 * MOVECOST_RIVER;
    case RMV_FAST_ALWAYS:
      return MOVECOST_RIVER;
    default:
      break;
    }
  }

  return (civtile_get_terrain_type(t2)->movement_cost * SINGLE_MOVE);
}


/***********************************************************************
 tile_movecosts_ai is used to fill the movecost array of struct
 tile. The cached values of this array are used in server/gotohand.c
 and client/goto.c. tile_movecosts_ai returns the move cost as
 calculated by tile_movecosts (with no unit pointer to get
 unit-independent results) EXCEPT if either of the source or
 destination tile is an ocean tile. Then the result of the method
 shows if a ship can take the step from the source position to the
 destination position (return value is MOVECOST_SEA)
 or not (return value is maxcost).

 A ship can take the step if:
   - both tiles are ocean or
   - one of the tiles is ocean and the other is a city or is unknown
     (or is a river estuary).

 This does not adequately handle air units, right ???
***********************************************************************/
static int tile_movecosts_ai(
  CIVTILE *tile0, CIVTILE *tile1, bool is_cardinal, int maxcost)
{
  if (tile0->terrain == T_OCEAN) {
    return (tile1->city
      || tile1->terrain == T_OCEAN
      || tile1->terrain == T_UNKNOWN
      || (is_cardinal && civtile_has_specials(tile1, S_RIVER)))
      ? MOVECOST_SEA : maxcost;
  }
  
  if (tile1->terrain == T_OCEAN) {
    return (tile0->city
      || tile0->terrain == T_UNKNOWN 
      || (is_cardinal && civtile_has_specials(tile1, S_RIVER)))
      ? MOVECOST_SEA : maxcost;
  }

  /* Unfortunately, is_cardinal T_RIVER <-> city is possible by both
   * land and sea units with the river estuary patch. Thus one cannot
   * be represented with current dual use movecosts - ships lose.
   */
  return tile_movecosts(NULL, tile0, tile1, is_cardinal);
}

/***********************************************************************
 ...
***********************************************************************/
static void debug_log_movecosts(char *str, int x, int y, CIVTILE *tile0)
{
  /* the %x don't work so well for oceans, where
     movecost[]==-3 ,.. --dwp
  */
  freelog(LOG_DEBUG, "%s (%d, %d) [%d %d %d %d %d %d %d %d]", str, x, y,
	  tile0->movecost[0], tile0->movecost[1],
	  tile0->movecost[2], tile0->movecost[3],
	  tile0->movecost[4], tile0->movecost[5],
	  tile0->movecost[6], tile0->movecost[7]);
}

/***********************************************************************
 Recalculate tile->movecost[] for (x,y), and for adjacent
 tiles in direction back to (x,y).  That is, call this when
 something has changed on (x,y), eg roads, city, transform, etc.
 see warmap in gotohand for usage
***********************************************************************/
void civmap_movecosts_reset(const CIVMAP *pmap, int x, int y)
{
  CIVTILE *tile0, *tile1;

  /* Must run init first */
  if (0 >= pmap->maxcost)
    return;

  tile0 = civmap_get_tile(pmap, x, y);
  debug_log_movecosts("Resetting move costs for", x, y, tile0);

  /* trying to move off the screen is the default */
  adjc_dir_iterate(pmap, x, y, x1, y1, dir) {
    bool is_cardinal =
      (civmap_get_type(pmap, MAP_TYPE_HEX) || x1 == x || y1 == y);
    tile1 = civmap_get_tile(pmap, x1, y1);
    tile0->movecost[dir] =
      tile_movecosts_ai(tile0, tile1, is_cardinal, pmap->maxcost);
    tile1->movecost[DIR_REVERSE(dir)] =
      tile_movecosts_ai(tile1, tile0, is_cardinal, pmap->maxcost);
  } adjc_dir_iterate_end;
  debug_log_movecosts("Reset move costs for", x, y, tile0);
}

/***********************************************************************
 Initialize tile->movecost[] for all tiles, where movecost[i]
 is the unit-independent cost to move _from_ that tile, to
 adjacent tile in direction specified by i.
***********************************************************************/
void civmap_movecosts_init(const CIVMAP *pmap)
{
  /*
   * SINGLE_MOVE * 24 (also server to flag movecosts_init)
   * - exceeds max unit moverate, i.e. allowed but impossible
   */
  ((CIVMAP *)pmap)->maxcost = 72;

  whole_civmap_iterate(pmap, x, y) {
    CIVTILE *tile0, *tile1;
    tile0 = civmap_get_tile(pmap, x, y);
    adjc_dir_iterate(pmap, x, y, x1, y1, dir) {
      bool is_cardinal =
        (civmap_get_type(pmap, MAP_TYPE_HEX) || x1 == x || y1 == y);
      tile1 = civmap_get_tile(pmap, x1, y1);
      tile0->movecost[dir] =
        tile_movecosts_ai(tile0, tile1, is_cardinal, pmap->maxcost);
    }} else {{
      dir -= DIR_off[1];
      tile0->movecost[dir] = pmap->maxcost;
    } adjc_dir_iterate_end;
  } whole_civmap_iterate_end;
}

#ifdef	FC__UNIT_H
/***********************************************************************
 The cost to move punit from where it is to an adjacent tile at x,y.
 It is assumed the move is a valid one, e.g. the tiles are adjacent.
***********************************************************************/
int civmap_movecost(const CIVMAP *pmap, struct unit *punit, int x, int y)
{
  bool is_cardinal = 
    (civmap_get_type(pmap, MAP_TYPE_HEX) || punit->x == x || punit->y == y);
  return tile_movecosts(pmap, punit, 
    civmap_get_tile(pmap, punit->x, punit->y),
    civmap_get_tile(pmap, x, y), is_cardinal);
}
#endif	/* FC__UNIT_H */


/***********************************************************************
 Carry out the activity and do move cost reset if required.
***********************************************************************/
void civmap_irrigate(const CIVMAP *pmap, int x, int y)
{
  enum terrain_type tt = civtile_irrigate(civmap_get_tile(pmap, x, y));
  if (tt != T_LAST) {
    civmap_terrain_changed(pmap, x, y, tt);
  }
}

/***********************************************************************
 Carry out the activity and do move cost reset if required.
***********************************************************************/
void civmap_mine(const CIVMAP *pmap, int x, int y)
{
  enum terrain_type tt = civtile_mine(civmap_get_tile(pmap, x, y));
  if (tt != T_LAST) {
    civmap_terrain_changed(pmap, x, y, tt);
  }
}

/***********************************************************************
 Carry out the activity and do move cost reset if required.
***********************************************************************/
void civmap_transform(const CIVMAP *pmap, int x, int y)
{
  enum terrain_type tt = civtile_transform(civmap_get_tile(pmap, x, y));
  if (tt != T_LAST) {
    civmap_terrain_changed(pmap, x, y, tt);
  }
}

/***********************************************************************
 Terrain transformed, now update all the nooks and crannies.
***********************************************************************/
void civmap_terrain_changed(const CIVMAP *pmap, int x, int y, 
  enum terrain_type prev)
{
  /* TODO: see server/settlers.c: ai_calc_transform(), unittools.c()
   *       for a selection of land/ocean side effects.
   * Update continent/ocean numbers, ...
   */

  /* Finally, update the new tile values */
  civmap_movecosts_reset(pmap, x, y);
}


/***********************************************************************
 Count tiles close to x0,y0 having terrain type tt
***********************************************************************/
int count_terrain_adjc(const CIVMAP *pmap, int x0, int y0, enum terrain_type tt)
{
  int count = 0;

  adjc_iterate(pmap, x0, y0, x1, y1) {
    if (civmap_get_terrain(pmap, x1, y1) == tt)
      count++;
  } adjc_iterate_end;

  return count;
}

/***********************************************************************
 Check if any tile close to x0,y0 has terrain type tt
***********************************************************************/
bool is_terrain_adjc(const CIVMAP *pmap, int x0, int y0, enum terrain_type tt)
{
  adjc_iterate(pmap, x0, y0, x1, y1) {
    if (civmap_get_terrain(pmap, x1, y1) == tt)
      return TRUE;
  } adjc_iterate_end;

  return FALSE;
}

/***********************************************************************
 Count tiles close to x0,y0 having special type spe
***********************************************************************/
int count_special_adjc(const CIVMAP *pmap, int x0, int y0, 
  enum special_type spe)
{
  int count = 0;

  adjc_iterate(pmap, x0, y0, x1, y1) {
    if (civmap_has_specials(pmap, x1, y1, spe))
      count++;
  } adjc_iterate_end;

  return count;
}

/***********************************************************************
 Check if any tile close to x0,y0 has special type spe
***********************************************************************/
bool is_special_adjc(const CIVMAP *pmap, int x0, int y0, enum special_type spe)
{
  adjc_iterate(pmap, x0, y0, x1, y1) {
    if (civmap_has_specials(pmap, x1, y1, spe))
      return TRUE;
  } adjc_iterate_end;

  return FALSE;
}

/***********************************************************************
 ...
***********************************************************************/
bool is_special_close(const CIVMAP *pmap, int x0, int y0)
{
  square_iterate(pmap, x0, y0, 2, x1, y1) {
    if (civmap_has_specials(pmap, x1, y1, (S_SPECIAL_1 | S_SPECIAL_2)))
      return TRUE;
  } square_iterate_end;

  return FALSE;
}

/***********************************************************************
 ...
***********************************************************************/
bool is_hut_close(const CIVMAP *pmap, int x0, int y0)
{
  square_iterate(pmap, x0, y0, 3, x1, y1) {
    if (civmap_has_specials(pmap, x1, y1, S_HUT))
      return TRUE;
  } square_iterate_end;

  return FALSE;
}

/***********************************************************************
 Land square with river-type connection to the sea
***********************************************************************/
bool is_at_coast(const CIVMAP *pmap, int x0, int y0)
{
  enum terrain_type ter = civmap_get_terrain(pmap, x0, y0);
  if (ter != T_OCEAN && ter != T_UNKNOWN) {
    cartesian_adjc_iterate(pmap, x0, y0, x1, y1) {
      if (civmap_get_terrain(pmap, x1, y1) == T_OCEAN)
        return TRUE;
    } cartesian_adjc_iterate_end;
  }
  return FALSE;
}

/***********************************************************************
 Land square adjacent to the sea
***********************************************************************/
bool is_coast(const CIVMAP *pmap, int x0, int y0)
{
  enum terrain_type ter = civmap_get_terrain(pmap, x0, y0);
  if (ter != T_OCEAN && ter != T_UNKNOWN) {
    adjc_iterate(pmap, x0, y0, x1, y1) {
      if (civmap_get_terrain(pmap, x1, y1) == T_OCEAN)
        return TRUE;
    } adjc_iterate_end;
  }
  return FALSE;
}

/***********************************************************************
 Sea square adjacent to land. 
***********************************************************************/
bool is_coastal(const CIVMAP *pmap, int x0, int y0)
{
  enum terrain_type ter = civmap_get_terrain(pmap, x0, y0);
  if (ter == T_OCEAN) {
    adjc_iterate(pmap, x0, y0, x1, y1) {
      ter = civmap_get_terrain(pmap, x1, y1);
      if (ter != T_OCEAN && ter != T_UNKNOWN)
        return TRUE;
    } adjc_iterate_end;
  }
  return FALSE;
}

/***********************************************************************
 Check if there is a water (irrigation) source for this tile.
***********************************************************************/
bool is_water_adjc(const CIVMAP *pmap, int x0, int y0)
{
  CIVTILE *ptile;

  ptile = civmap_get_tile(pmap, x0, y0);
  if (ptile->terrain == T_OCEAN
      || ptile->terrain == T_RIVER
      || civtile_has_specials(ptile, (S_RIVER | S_IRRIGATION)))
    return TRUE;

  cartesian_adjc_iterate(pmap, x0, y0, x1, y1) {
    ptile = civmap_get_tile(pmap, x1, y1);
    if (ptile->terrain == T_OCEAN
	|| ptile->terrain == T_RIVER
	|| civtile_has_specials(ptile, (S_RIVER | S_IRRIGATION)))
      return TRUE;
  } cartesian_adjc_iterate_end;

  return FALSE;
}

/***********************************************************************
 TRUE if there is a land tile within a city radius.
***********************************************************************/
bool is_sea_usable(const CIVMAP *pmap, int x0, int y0)
{
  citymap_checked_iterate(pmap, x0, y0, cx, cy, x1, y1) {
    if (civmap_get_terrain(pmap, x1, y1) != T_OCEAN)
      return TRUE;
  } citymap_checked_iterate_end;

  return FALSE;
}

/***********************************************************************
 TRUE if the surrounding terrain is only grass or plains
***********************************************************************/
bool is_terrain_clean(const CIVMAP *pmap, int x0, int y0)
{
  square_iterate(pmap, x0, y0, 2, x1, y1) {
    enum terrain_type tt = civmap_get_terrain(pmap, x1,y1);
    if (tt != T_GRASSLAND && tt != T_PLAINS)
      return FALSE;
  } square_iterate_end;

  return TRUE;
}

/***********************************************************************
 Returns TRUE if (x0,y0) is _not_ a good position to start from;
 Bad places:
 - Non-suitable terrain;
 - On a hut;
 - On North/South pole continents
 - Too close to another starter on the same continent:
 'dist' is too close (real_map_distance)
 'nr' is the number of other start positions in
 map.start_positions to check for too closeness.
***********************************************************************/
bool is_starter_bad(const CIVMAP *pmap, int x0, int y0, int nr, int dist) 
{
  int i;
  enum terrain_type t = civmap_get_terrain(pmap, x0, y0);

  /* only start on clear terrain: */
  if (t != T_PLAINS && t != T_GRASSLAND && t != T_RIVER && t != T_HILLS)
    return TRUE;
  
  /* don't start on a hut: */
  if (civmap_has_specials(pmap, x0, y0, S_HUT))
    return TRUE;
  
  /* don't want them starting on the poles unless the poles are
     connected to more land: */
  if (pmap->generator > 0 
   && pmap->separatepoles 
   && civmap_get_continent(pmap, x0, y0) 
      <= (civmap_get_type(pmap, MAP_TYPE_WRAPY) ? 1 : 2))
    return TRUE;

  /* don't start too close to someone else: */
  for (i = 0; i < nr; i++) {
    int x1 = pmap->start_positions[i].x;
    int y1 = pmap->start_positions[i].y;
    if (civmap_get_continent(pmap, x0, y0) == civmap_get_continent(pmap, x1, y1)
	 && civmap_real_distance(pmap, x0, y0, x1, y1) < dist) {
      return TRUE;
    }
  }
  return FALSE;
}

/***********************************************************************
 This function returns true if the tile at the given location can be
 "reclaimed" from ocean into land.  This is the case only when there are
 a sufficient number of adjacent tiles that are not ocean.
***********************************************************************/
bool can_reclaim_ocean(const CIVMAP *pmap, int x, int y)
{
  int landtiles = pmap->terrain->ocean_reclaim_requirement;

  if (landtiles >= 9)
    return FALSE;
  if (landtiles <= 0)
    return TRUE;

  adjc_iterate(pmap, x, y, x1, y1) {
    if (civmap_get_tile(pmap, x1, y1)->terrain != T_OCEAN)
      if (--landtiles == 0)
	return TRUE;	
  } adjc_iterate_end;

  return FALSE;
}

/***********************************************************************
 This function returns true if the tile at the given location can be
 "channeled" from land into ocean.  This is the case only when there are
 a sufficient number of adjacent tiles that are ocean.
***********************************************************************/
bool can_channel_land(const CIVMAP *pmap, int x, int y)
{
  int oceantiles = pmap->terrain->land_channel_requirement;

  if (oceantiles >= 9)
    return FALSE;
  if (oceantiles <= 0)
    return TRUE;

  adjc_iterate(pmap, x, y, x1, y1) {
    if (civmap_get_tile(pmap, x1, y1)->terrain == T_OCEAN)
      if (--oceantiles == 0)
	return TRUE;
  } adjc_iterate_end;

  return FALSE;
}


/***********************************************************************
 Random point anywhere on the map.
 Only normalized positions (for which normalize_civmap_pos returns true)
 will be returned. The runtime is unbounded.
***********************************************************************/
void rand_civmap_pos(const CIVMAP *pmap, int *x, int *y)
{
  /* guarantee at least one real tile in the region. */
  assert(pmap->xsize > 0 && pmap->ysize > 0);

  do {
    *x = myrand(pmap->xsize);
    *y = myrand(pmap->ysize);
    native_to_civmap_pos(pmap, x, y, *x, *y);
  } while (!normalize_civmap_pos(pmap, x, y));
}

/***********************************************************************
 Random point in a fixed region.
 The region is a native rectangle bounded by the two civmap points.
 Only normalized positions (for which normalize_civmap_pos returns true)
 will be returned. The runtime is unbounded.
***********************************************************************/
void rand_region_pos(const CIVMAP *pmap, int *x, int *y,
  int x0, int y0, int x1, int y1)
{
  /* guarantee at least one real tile in the region. */
  assert(pmap->xsize > 0 && pmap->ysize > 0);
  assert(is_real_civmap_pos(pmap, x0, y0) || is_real_civmap_pos(pmap, x1, y1));

  /* switch to native coordinates */
  civmap_to_native_pos(pmap, &x0, &y0, x0, y0);
  civmap_to_native_pos(pmap, &x1, &y1, x1, y1);

  /* Convert two points to point and bounded offset. */
  if (x0 > x1)
    x1 = x0 - x1, x0 -= x1;
  else
    x1-= x0;
  if (++x1 >= pmap->xsize)
    x1 = pmap->xsize;

  if (y0 > y1)
    y1 = y0 - y1, y0 -= y1;
  else
    y1-= y0;
  if (++y1 >= pmap->ysize)
    y1 = pmap->ysize;

  do {
    *x = myrand(x1) + x0;
    *y = myrand(y1) + y0;
    native_to_civmap_pos(pmap, x, y, *x, *y);
  } while (!normalize_civmap_pos(pmap, x, y));
}

/***********************************************************************
 Random point in a fixed region of allowed terrain_types (non-zero bitmap),
 allowed special_types (non-zero bitmap) or allowed user condition, in
 other words the intersection of the activated (non-null) constraints.
 The region is a native rectangle bounded by the two civmap points.
 Only normalized positions (for which normalize_civmap_pos returns true)
 will be returned. The runtime is finite.

 Note: the user function is called in the inner loop with a random map
 position satisfying the region and terrain constraints. As long as it
 returns a FALSE result, the loop will continue to feed it new tiles.
 Thus the active code can be within the user function rather than it
 merely being a test condition to get a point for the caller to act on.
 Or the user function can store the points it receives for use by a 
 caller that wants a selection of tiles to operate on. This "trick" 
 saves a lot of the overhead of repeated calls when multiple selections 
 with identical conditions are required.
***********************************************************************/
void rand_terrain_pos(const CIVMAP *pmap, int *x, int *y, int x0, int y0,
  int x1, int y1, int terrains, int specials, bool (*okuserfn)())
{
  int nn, xy, *yx;

  /* switch to native coordinates */
  civmap_to_native_pos(pmap, &x0, &y0, x0, y0);
  civmap_to_native_pos(pmap, &x1, &y1, x1, y1);

  /* Convert two points to point and bounded offset. */
  if (x0 > x1)
    x1 = x0 - x1, x0 -= x1;
  else
    x1-= x0;
  if (++x1 > pmap->xsize)
    x1 = pmap->xsize;

  if (y0 > y1)
    y1 = y0 - y1, y0 -= y1;
  else
    y1-= y0;
  if (++y1 > pmap->ysize)
    y1 = pmap->ysize;

  /* map the full range - this can be expensive for large ranges if the
   * loop only needs to cycle a few times. But it is a tight loop of 
   * no more than 1 or 2 fast instructions, i.e. ~1 % of the mainloop.
   */
  xy = x1 * y1;
  yx = fc_malloc(sizeof(int) * xy);
  for (nn = xy; --nn >= 0; )
    yx[nn] = nn;

  /* Iterate once over the range in random order. */
  do {
    if (0 >= xy) {
      /* All done - there were no tiles matching the test conditions. */
      *x = -1, *y = -1;
      break;
    }
    /* Select one element from the remaining range */
    nn = myrand(xy--);
    *x = yx[nn], yx[nn] = yx[xy];
    *y = *x / y1 + y0;
    *x = *x % x1 + x0;
    native_to_civmap_pos(pmap, x, y, *x, *y);
  } while (!normalize_civmap_pos(pmap, x, y)
   || ( 0 != terrains
     && 0 == (terrains & (1 << civmap_get_terrain(pmap, *x, *y))))
   || ( 0 != specials
     && 0 == (specials & civmap_get_specials(pmap, *x, *y)))
   || ( okuserfn
     && !(*okuserfn)(pmap, x, y)));

  free(yx);
}

/***********************************************************************
 Random neighbouring square.
***********************************************************************/
void rand_adjc(const CIVMAP *pmap, int *x, int *y, int x0, int y0)
{
  bool is_border = civmap_is_border_pos(pmap, x0, y0);
  int choice;

  /* Fast check - any direction (0:7) is equally valid */
  choice = myrand(8);
  if (civmap_step(pmap, x, y, x0, y0, choice, is_border)) {
    /* this neighbour's OK */
    return;
  } else {
    /* 
     * Ok, we're on an edge or corner tile - now finish the hard way, 
     * First make a list of all 8 directions - note Dirs 0:7 is endemic
     * throughout the code so no need for enum complications.
     */
    int nn = 8, dirs[8] = { 0, 1, 2, 3, 4, 5, 6, 7 };

    /*
     * This clever loop by Trent Piepho will take no more than 8 tries 
     * to find a valid direction.
     */
    while (--nn > 0) {
      /*
       * Last choice was bad, so replace it with the last direction in
       * the list. On each iteration, one fewer choices will remain.
       */
      dirs[choice] = dirs[nn];
      choice = myrand(nn);
      if (civmap_step(pmap, x, y, x0, y0, dirs[choice], is_border)) {
        /* this neighbour's OK */
        return;
      }
    }
  }
  assert(FALSE);		/* Must be a 1x1 map with no wrapping??? */
}

