/********************************************************************** 
 Freeciv - Copyright (C) 1996 - A Kjeldberg, L Gregersen, P Unold
 Civmap - Copyright (C) 2002 - R. Wetmore

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
***********************************************************************/

#define FC__CIVMAP_MAIN_C

/**
 * civmap_main.c - program driver for standalone map generation
 *
 * Civmap is a standalone program used to create and/or update 
 * Freeciv save files. It includes the 5 core server generators
 * with extensions for wrapped maps in isometric and/or hexagonal
 * form.
 */
static char *USAGE = "
Usage: civmap [--verbose [0==off|1==normal|2==debug]]
              [--file {filename|$CIVMAPFILE|\"civmap.sav.gz\"}]
              [--outfile {filename|$CIVMAPFILE|\"civmap.sav.gz\"}]
              [--seed {1-ULONG_MAX}]
              [--generator {1-5}]
              [--maptype {0-7|EARTH_ISO==5}]
              [--options is_earth,tinyisles,separatepoles,
                         has_huts,has_specials,fixedstart]
              [--riches {0-landtiles/3}]
              [--huts {0-landtiles/5}]
              [--players {1-MAX_NUM_NATIONS}]
              // Simple
              [--size {small|normal|large}]
              [--landtype {[~]archipelago|normal|pangea}]
              [--age {old|normal|young}]
              [--wet {dry|normal|wet}]
              [--temperature {cold|cool|normal|warm|hot}]
              // Advanced parameters
              [--geometry 80x50]
              [--landpercent 33]
              [--terrain {\"default\"|index|name}.percent]
              [--river   {\"default\"|index|name}.percent]
              [--special {\"default\"|index|name}.percent]
";

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <ctype.h>
#include <getopt.h>

#include "log.h"
#include "capability.h"
#include "registry.h"
#include "property.h"

#include "civmap_main.h"

#include "civmap.h"
#include "civmapgen.h"

/* common/gane.h */
bool is_server = FALSE;

/* common/gane.c */
void dealloc_id(int id) {} /* kludge */

/* server/unittools.c */
struct player;
struct unit;
void send_unit_info(struct player *dest, struct unit *punit) {} /* kludge */


/* Getopt globals */
char *optarg;
int  optind, opterr, optopt;

static char *short_options = "vf:F:n::g:m:o:r:h:p:s:l:a:w:t:G:L:R:T:S:";

static struct option long_options[] = {
  // General parameters
  {"verbose",       2, NULL, 'v'},
  {"outfile",       2, NULL, 'F'},
  {"file",          2, NULL, 'f'},
  {"seed",          2, NULL, 'n'},
  {"generator",     1, NULL, 'g'},
  {"maptype",       1, NULL, 'm'},
  {"options",       1, NULL, 'o'},
  {"riches",        1, NULL, 'r'},
  {"huts",          1, NULL, 'h'},
  {"players",       1, NULL, 'p'},
  // Simple parameters
  {"size",          1, NULL, 's'},
  {"landtype",      1, NULL, 'l'},
  {"age",           1, NULL, 'a'},
  {"wet",           1, NULL, 'w'},
  {"temperature",   1, NULL, 't'},
  // Advanced parameters
  {"geometry",      1, NULL, 'G'},
  {"landpercent",   1, NULL, 'L'},
  {"river",         1, NULL, 'R'},
  {"terrain",       1, NULL, 'T'},
  {"special",       1, NULL, 'S'},
  {NULL,            0, NULL, 0}
};

static PROPERTY size_opts[] = { 
  { "small",  {40 + 25*65536} },
  { "normal", {80 + 50*65536} },
  { "large",  {160+100*65536} },
  { NULL, {0} }
};

static PROPERTY maptype_opts[] = { 
  { "FLATEARTH",      {0} },
  { "EARTH",          {1} },
  { "URANUS",         {2} },
  { "DONUTWORLD",     {3} },
  { "FLATEARTH_ISO",  {4} },
  { "EARTH_ISO",      {5} },
  { "URANUS_ISO",     {6} },
  { "DONUTWORLD_ISO", {7} },
  { NULL, {0} }
};

static PROPERTY landtype_opts[] = { 
  { "archipelago", {15} },
  { "normal",      {30} },
  { "pangea",      {45} },
  { NULL, {0} }
};

static PROPERTY age_opts[] = { 
  { "old",    {0} },
  { "normal", {1} },
  { "young",  {2} },
  { NULL, {0} }
};

static PROPERTY wet_opts[] = { 
  { "dry",    {0} },
  { "normal", {1} },
  { "wet",    {2} },
  { NULL, {0} }
};

static PROPERTY temperature_opts[]= { 
  { "cold",   {0} },
  { "cool",   {1} },
  { "normal", {2} },
  { "warm",   {3} },
  { "hot",    {4} },
  { NULL, {0} }
};

/* TODO: load from ruleset - issue with disambiguating 2 types of oil */
static PROPERTY terrain_opts[] = {
 /* basic terrain types */
 {"glacier", { 0}}, {"desert",   { 1}}, {"forest",   { 2}}, {"grassland",{ 3}},
 {"hills",   { 4}}, {"jungle",   { 5}}, {"mountains",{ 6}}, {"ocean",    { 7}},
 {"plains",  { 8}}, {"rivervale",{ 9}}, {"swamp",    {10}}, {"tundra",   {11}},
 {"unknown", {12}}, {"user1",    {13}}, {"user2",    {14}}, {"default",  {15}},

 /* river overlayed on basic terrain types */
 {"glacier", {16}}, {"desert",   {17}}, {"forest",   {18}}, {"grassland",{19}},
 {"hills",   {20}}, {"jungle",   {21}}, {"mountains",{22}}, {"ocean",    {23}},
 {"plains",  {24}}, {"rivervale",{25}}, {"swamp",    {26}}, {"tundra",   {27}},
 {"unknown", {28}}, {"user1",    {29}}, {"user2",    {30}}, {"default",  {31}},
 
 /* special_1 types */
 {"ivory",   { 0}}, {"oasis",    { 1}}, {"pheasant", { 2}}, {"grassland",{ 3}},
 {"coal",    { 4}}, {"gems",     { 5}}, {"gold",     { 6}}, {"fish",     { 7}},
 {"buffalo", { 8}}, {"rivervale",{ 9}}, {"peat",     {10}}, {"game",     {11}},
 {"unknown", {12}}, {"user1",    {13}}, {"user2",    {14}}, {"default",  {15}},

 /* special_2 types */
 {"polaroil",{16}}, {"oil",      {17}}, {"silk",     {18}}, {"savannah", {19}},
 {"wine",    {20}}, {"fruit",    {21}}, {"iron",     {22}}, {"whales",   {23}},
 {"wheat",   {24}}, {"rivervale",{25}}, {"spices",   {26}}, {"furs",     {27}},
 {"unknown", {28}}, {"user1",    {29}}, {"user2",    {30}}, {"default",  {31}},

 { NULL, {0} }
};


/* Globals */

struct civmap *theMap = NULL;

int  verbose = 1;
int  loglevel = LOG_NORMAL;
char *logfile = NULL;


/**
 * civmap program 
 */
int main (int argc, char **argv)
{
  const char *filename = CIVMAPFILE;
  struct section_file *fp = NULL;
  PROPERTY *props;
  char *px;
  int nn, cc, fflg = 0, errs = 0;

  /* system bootstrap */
  logdebug_suppress_warning;
  init_nls();
  dont_run_as_root(argv[0], "civmap");

  if (NULL != (px = getenv("CIVMAPLOGLEVEL"))
   && 0 != (nn = atoi(px))) {
    loglevel = nn;
  }
  if (NULL != (px = getenv("CIVMAPLOG"))) {
    logfile = px;
  }
  log_init(logfile, loglevel, NULL);

  /* initialize the map struct with our hardcoded defaults */
  theMap = civmap_init(NULL);

  /* Load the backing template savefile and merge defaults into map. */
  /* Note: in addition to providing a defaults customization hook,   */
  /*       this ensures compatibility with evolving savefile formats */
  /*       specifically, non-map related changes, and means we only  */
  /*       need to worry about the map section here.                 */
  if (NULL == (px = getenv("CIVMAPRC"))) {
    px = (char *)CIVMAPRC;
  }
  if (!(fp = section_file_load(fp, px))
   || 0 != civmap_load(theMap, fp)) {
    fprintf(stderr, "*** Error: loading CIVMAPRC %s\n", px);
    ++errs;
  }

  /* parse the options, updating map parameters in sequence */
  while (0 == errs) {
    int optindex = 0;
    cc = getopt_long(argc, argv, short_options, long_options, &optindex);
    if (-1 == cc)
      break;

    switch (cc) {
    default:
    case '?':
      fprintf(stderr, "*** Error: getopt return code 0%o at argument %s\n",
        cc, argv[optindex]);
      ++errs;
      break;

    case 'v':
      if (NULL != optarg)
        verbose = atoi(optarg);
      else
        verbose = 1;
      break;

    case 'F':
      if (NULL != optarg) {
        filename = optarg;
        fflg++;
        break;
      }
      ++errs;
      break;

    case 'f':
      if (NULL != optarg) {
        if (0 == fflg++)
          filename = optarg;
        /* Note: may overwrite any defaults or preceding input options */
        if (section_file_load(fp, optarg)
         && 0 == civmap_load(theMap, fp))
            break;
        fprintf(stderr, "*** Error: loading CIVMAPFILE %s\n", optarg);
      }
      ++errs;
      break;

    case 'n':
      if (NULL != optarg)
        theMap->seed = atoi(optarg);
      break;

    case 'm':
      if (NULL != optarg) {
        if (NULL != (props = getProperty_by_index(maptype_opts, optarg))) {
          theMap->type = props->value;
          break;
        }
        fprintf(stderr, "*** Error: maptype %s not found ... ", optarg);
        for (props = maptype_opts, nn = 0; nn < 8; nn++)
          fprintf(stderr," %ld-%s", props[nn].value, props[nn].name);
        fprintf(stderr, "\n");
      }
      ++errs;
      break;

    case 'o':
      if (NULL != optarg)
        theMap->type = atoi(optarg);
      else
        ++errs;
      break;

    case 'r':
      if (NULL != optarg)
        theMap->riches = atoi(optarg);
      else
        ++errs;
      break;

    case 'h':
      if (NULL != optarg)
        theMap->huts = atoi(optarg);
      else
        ++errs;
      break;

    case 'p':
      if (NULL == optarg
       || MAX_NUM_NATIONS <= (theMap->num_start_positions = atoi(optarg)))
        ++errs;
      break;

    case 'l':
      if (NULL != optarg) {
        if ('~' == optarg[0]) {
          if (NULL != (props = getProperty_by_index(landtype_opts, optarg+1))) {
            theMap->landpercent = 100 - props->value;
            break;
          }
        } else {
          if (NULL != (props = getProperty_by_index(landtype_opts, optarg))) {
            theMap->landpercent = props->value;
            break;
          }
        }
      }
      ++errs;
      break;

    case 'a':
      if (NULL != optarg
       && NULL != (props = getProperty_by_index(age_opts, optarg))) {
        /* TODO: massage the terrain types */
        break;
      }
      ++errs;
      break;

    case 'w':
      if (NULL != optarg
       && NULL != (props = getProperty_by_index(wet_opts, optarg))) {
        /* TODO: massage the terrain types */
        break;
      }
      ++errs;
      break;

    case 't':
      if (NULL != optarg
       && NULL != (props = getProperty_by_index(temperature_opts, optarg))) {
        /* TODO: massage the terrain types */
        break;
      }
      ++errs;
      break;

    case 's':
    case 'G':
      if (NULL != optarg
       && isdigit(optarg[0])
       && NULL != (px = strchr(optarg, 'x'))) {
        *(px++) = '\0';
        theMap->xsize = atoi(optarg), theMap->ysize = atoi(px);
        break;
      }

      if (NULL != optarg
       && NULL != (props = getProperty_by_index(size_opts, optarg))) {
        if (theMap->type & 4) 
          theMap->xsize = props->x, theMap->ysize = props->y;
        else
          theMap->xsize = props->y, theMap->ysize = props->x;
        break;
      }
      ++errs;
      break;

    case 'L':
      if (NULL != optarg)
        theMap->landpercent = atoi(optarg);
      else
        ++errs;
      break;

    case 'T':
      if (NULL != optarg
       && NULL != (px = strchr(optarg, '.'))) {
        *(px++) = '\0';
        if ( NULL != (props = getProperty_by_index(terrain_opts, optarg))) {
          theMap->terrainpercent[props->value] = atoi(px);
          break;
        }
        fprintf(stderr, "*** Error: terrain %s not found ... ", optarg);
        for (props = terrain_opts, nn = 0; nn < 16; nn++) {
          if (0 == (nn & 7))
            fprintf(stderr, "\n\t");
          fprintf(stderr," %ld-%s", props[nn].value, props[nn].name);
        }
        fprintf(stderr, "\n");
      }
      ++errs;
      break;

    case 'R':
      if (NULL != optarg
       && NULL != (px = strchr(optarg, '.'))) {
        *(px++) = '\0';
        if ( NULL != (props = getProperty_by_index(terrain_opts+16, optarg))) {
          theMap->terrainpercent[props->value+16] = atoi(px);
          break;
        }
        fprintf(stderr, "*** Error: river terrain %s not found ... ", optarg);
        for (props = terrain_opts, nn = 16; nn < 32; nn++) {
          if (0 == (nn & 7))
            fprintf(stderr, "\n\t");
          fprintf(stderr," %ld-%s", props[nn].value, props[nn].name);
        }
        fprintf(stderr, "\n");
      }
      ++errs;
      break;

    case 'S':
      if (NULL != optarg
       && NULL != (px = strchr(optarg, '.'))) {
        *(px++) = '\0';
        if ( NULL != (props = getProperty_by_index(terrain_opts+32, optarg))) {
          theMap->terrainpercent[props->value+32] = atoi(px);
          break;
        }
        fprintf(stderr, "*** Error: special %s not found ... ", optarg);
        for (props = terrain_opts+32, nn = 0; nn < 32; nn++) {
          if (0 == (nn & 7))
            fprintf(stderr, "\n\t");
          fprintf(stderr," %ld-%s", props[nn].value, props[nn].name);
        }
        fprintf(stderr, "\n");
      }
      ++errs;
      break;
    }
  }

  if (0 != errs) {
    fprintf (stderr, "*** Error: at argument %s\n", argv[optind]);
  } else
  if (optind < argc) {
    fprintf(stderr, "*** Error: surplus arguments ... ");
    while (optind < argc)
      fprintf (stderr, "%s ", argv[optind++]);
    fprintf(stderr, "\n");
    ++errs;
  }

  /* input errors - give syntax reminder and exit */
  if (0 != errs) {
    fprintf (stderr, USAGE);
    exit (1);
  }

  /* Final option massaging */
  if (0 != theMap->generator) {
    theMap->has_specials = 0;
    theMap->has_huts = 0;
  }
  if (0 == theMap->seed) {
    theMap->seed = time(NULL) + getpid();
  }
  nn = strlen(filename);
  if ((COMPRESSED > 0) == (nn < 3 || strcmp(filename+nn-3, ".gz") != 0)) {
    px = fc_malloc(nn+4);
    strcpy(px, filename);
    if (COMPRESSED > 0)
      strcpy(px+nn, ".gz");
    else
      px[nn-3] = '\0';
    filename = px;
  }

  /* Begin processing ... */
  if (!civmap_gen(theMap))
    exit (2);

  if (0 != civmap_save(theMap, fp))
    exit (3);

  if (!section_file_save_nodup(fp, filename, COMPRESSED))
    exit (4);

  /* ... success!!! */
  section_file_free(fp);
  civmap_free(&theMap);
  exit (0);
}

