/**********************************************************************
 Freeciv - Copyright (C) 1996 - A Kjeldberg, L Gregersen, P Unold
 civmapgen.c - Copyright (C) 2002 - R. Wetmore

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
***********************************************************************/

#define FC__CIVMAPGEN_C

/**
 * civmapgen.c - map generation algorithm
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "fcintl.h"
#include "log.h"
#include "mem.h"
#include "rand.h"
#include "shared.h"
#include "support.h"

#include "civmap.h"
#include "civmap_ex.h"
#include "civmapgen.h"


/* Wrapper for easy height_map access. It's a macro so it can be an lvalue. */
#define hmap(pmap, x, y)    height_map[civmap_pos_to_index((pmap), (x), (y))]

static int *height_map;


/* utility routines */
static int  hmapN(const CIVMAP *pmap, int x, int y, int n);
static int  hslope(const CIVMAP *pmap, int *n, int x, int y, int hi);
static int  count_terrainN(const CIVMAP *pmap, int x, int y, int nn, int tt);
static int  civmap_dist(const CIVMAP *pmap, int p0, int p1);
static bool civmap_is_cold(const CIVMAP *pmap, int x, int y);

/* core map generators */
static void mapgenerator1(CIVMAP *pmap);
static void mapgenerator2(CIVMAP *pmap);
static void mapgenerator3(CIVMAP *pmap);
static void mapgenerator4(CIVMAP *pmap);
static void mapgenerator5(CIVMAP *pmap);

static int  continent_flood(const CIVMAP *pmap, int x, int y, int nr);
static int  continent_flood1(const CIVMAP *pmap, int x, int y, int n0, int nr);
static int  continent_list1(const CIVMAP *pmap, int *list, int x, int y,int n0);
static void setup_isledata(CIVMAP *pmap);

static int  ocean_flood(const CIVMAP *pmap, int x, int y, int nr);
static int  ocean_flood1(const CIVMAP *pmap, int x, int y, int n0, int nr);
static int  ocean_list(const CIVMAP *pmap, int *list, int x, int y);
static int  ocean_list1(const CIVMAP *pmap, int *list, int x, int y, int n0);
static void setup_seadata(CIVMAP *pmap);

static void add_specials(CIVMAP *pmap, int prob);
static void make_huts(CIVMAP *pmap, int number);

static void map_check(const CIVMAP *pmap);

/* mapgenerator1&5 routines */
static void adjust_map2(const CIVMAP *pmap);
static void adjust_map(const CIVMAP *pmap);
static void smooth_map(const CIVMAP *pmap);

static void gen1_hmap(const CIVMAP *pmap);
static void gen5_hmap(const CIVMAP *pmap);
static void gen5rec(const CIVMAP *pmap, 
  int step, int x0, int y0, int x1, int y1);

static void make_mountains(const CIVMAP *pmap, int hill);
static void make_polar(const CIVMAP *pmap);
static int  make_desert(const CIVMAP *pmap, int x, int y, int height, int diff);
static void make_deserts(const CIVMAP *pmap);
static int  make_forest(const CIVMAP *pmap, int x, int y, int height, int diff);
static void make_forests(const CIVMAP *pmap);
static void make_swamps(const CIVMAP *pmap);
static int  make_river(const CIVMAP *pmap, int x, int y, int force);
static void make_rivers(const CIVMAP *pmap);
static void make_plains(const CIVMAP *pmap);
static void make_passable(const CIVMAP *pmap);
static void make_fair(const CIVMAP *pmap);
static void make_land(CIVMAP *pmap);
static bool is_tiny_island(const CIVMAP *pmap, int x, int y);
static void remove_tiny_islands(const CIVMAP *pmap);
static bool is_tiny_lake(const CIVMAP *pmap, int x, int y);
static void remove_tiny_lakes(const CIVMAP *pmap);

/* mapgenerator2-4 routines */
static void initworld(CIVMAP *pmap);
static void make_island(CIVMAP *pmap, int islemass, int starters);
static void create_island(const CIVMAP *pmap, int *height_map, int islemass);
static int  place_island(const CIVMAP *pmap, int *create_map, int forever);
static void fill_island(const CIVMAP *pmap, long int *bucket, int coast,
        int warm0_weight, int warm1_weight, int cold0_weight, int cold1_weight,
        enum terrain_type warm0, enum terrain_type warm1,
        enum terrain_type cold0, enum terrain_type cold1);

/* debug */
void map_printfmap(const CIVMAP *pmap);
void map_printhmap(const CIVMAP *pmap, int *height_map);

/* global data */
static int maxval=10000;
static int thills=6000;
static int tocean=3000;
static int trough=300;

struct isledata {
  int x,y;                          /* upper left corner of the islands */
  int count;
  int goodies;
  int starters;
};
static struct isledata *islands;

struct seadata {
  int x, y;                         /* upper left corner of the islands */
  int count;                        /* number of connected water tiles  */
  int rivers;                       /* number of connected river tiles  */
};
static struct seadata *oceans;

/* this is used for generators 2,3,4 */
#define MAP_NCONT 255


/**
 * Return the average height at map position x,y for radius N.
 */
static int hmapN(const CIVMAP *pmap, int x, int y, int n)
{
  int nn= 0, hh = 0;

  /* limit to sane distances */
  n = CLIP(0, n, 5);

  square_iterate(pmap, x, y, n, xx, yy) {
    nn++, hh+= hmap(pmap, xx, yy);
  } square_iterate_end

  /* paranoia */
  return (0 < nn ?  hh/nn : 0);
}

/**
 * Detect highest/lowest (3 point average) neighbouring direction
 * direction returned in dir if address not NULL.
 * hi != 0 returns the highest, otherwise lowest.
 */
static int hslope(const CIVMAP *pmap, int *dir, int x, int y, int hi)
{
  int n0, h0, ho, h1, h2, h3, hh, x1, y1;
  bool is_border = civmap_is_border_pos(pmap, x, y);

  /* 3-tile average of adjacent squares - init for direction 0 */
  n0 = 0;
  ho = hmap(pmap, x, y);
  h2 = civmap_step(pmap, &x1, &y1, x, y, 6, is_border) ? hmap(pmap, x1, y1):ho;
  h1 = civmap_step(pmap, &x1, &y1, x, y, 7, is_border) ? hmap(pmap, x1, y1):ho;
  h3 = civmap_step(pmap, &x1, &y1, x, y, 0, is_border) ? hmap(pmap, x1, y1):ho;
  h0 = hh = h1 + h2 + h3;

  /* wrap to end of direction vector and walk down */
  adjc_dir_iterate(pmap, x, y, x1, y1, nn) {
    hh-= h3, h3= h2, h2= h1;
    hh+= h1 = hmap(pmap, x1, y1);
    if( (hh < h0) == (hi == 0) )
      h0 = hh, n0 = nn;
  }} else {{
    hh-= h3, h3= h2, h2= h1;
    hh+= h1 = ho;
  } adjc_dir_iterate_end;
  if( NULL != dir ) *dir = n0;

  return ++h0/3;
}

/**
 * Counts tiles of a given type including pseudo rivers within radius nn
 * A negative range means "tiles not of the given type".
 */
static int count_terrainN(const CIVMAP *pmap, int x, int y, int nn, int tt)
{
  bool is_tile;
  int count = 0;

  /* set count type */
  if (!(is_tile = (nn >= 0)))
    nn = -nn;

  /* limit to sane distances */
  nn = CLIP(0, nn, 5);

  square_iterate(pmap, x, y, nn, xx, yy) {
    if (is_tile == civmap_has_terrain(pmap, xx, yy, tt))
      count++;
  } square_iterate_end

  return count;
}

/* fast approximate true distance without using square roots
 * i.e. diagonals counted as 3/2
 */
static int civmap_dist(const CIVMAP *pmap, int p0, int p1)
{
  int i0, i1, x0, y0, x1, y1;

  index_to_civmap_pos(pmap, &x0, &y0, &x0, &y0, p0);
  index_to_civmap_pos(pmap, &x1, &y1, &x1, &y1, p1);
  civmap_distance(pmap, &i0, &i1, x0, y0, x1, y1);
  i1= abs(i1), i0 = abs(i0);
  i0= (i0 < i1) ? (i1 + i0/2):(i0 + i1/2);
  return i0;
}

/* top or bottom 20% (fifth) of pmap->ysize are cold */
static bool civmap_is_cold(const CIVMAP *pmap, int x, int y)
{
  civmap_to_native_pos(pmap, &x, &y, x, y);
  return (y*5 < pmap->ysize || y*5 >= pmap->ysize*4);
}

/**************************************************************************
  make_mountains() will convert all squares that are higher than thills to
  mountains and hills. Notice that thills will be adjusted according to
  the pmap->mountains value in a 1:2 ratio, so increase pmap->mountains and
  you will get more hills and mountains (and vice versa).
  Add hills/mtns at low-level rough locations, and make 1/3 of highlands
  into plains/deserts/glaciers.
  Note try to balance height and rough, and adjust thills so that there is
  an excess of each for random selection. This spreads the mountains out.
**************************************************************************/
static void make_mountains(const CIVMAP *pmap, int hill)
{
  int *hs, *yx, txy, j, k, tmp = pmap->xsize*pmap->ysize;
  int total=(((tmp*pmap->landpercent)/100)*pmap->MOUNTAINS)/100;
  int hx[16], mount, tmount= (total * 20) / 100;  /* 20% pure height */
  int sx[16], rough, slope = 1000;

  /* Compute once and cache */
  yx = (int *)fc_malloc(sizeof(int) * tmp);
  hs = (int *)fc_malloc(sizeof(int) * tmp);
  txy = 0;
  whole_civmap_xy_iterate(pmap, xy, xn, yn, x, y) {
    CIVTILE *ptile = (civmap_get_tiles(pmap) + xy);
    if (civtile_get_terrain(ptile) == T_GRASSLAND
     || (civtile_get_terrain(ptile) == T_TUNDRA && myrand(3) == 0)) {
      /* bias against mountainous coastal/islands */
      tmp = count_terrainN(pmap, x, y, 1, T_OCEAN) + 1;
      if (myrand(tmp) + myrand(tmp) > 3)
        continue;
      hs[xy] = height_map[xy] - hslope(pmap, NULL, x, y, 0);
      k = myrand(++txy), yx[txy-1]= yx[k], yx[k]= xy;
    }
  } whole_civmap_xy_iterate_end;

  /* iterate the distribution counts to find a rough height cutoff */
  thills = hill;
  for (j = 10; --j > 0;) {
    memset(hx, 0, sizeof(hx));
    memset(sx, 0, sizeof(sx));
    rough = mount = 0;
    for (k = txy; k-- > 0;) {
      int xy = yx[k];

      /* find distribution of steepness around land tiles */
      if ((tmp = hs[xy] - slope) >= 0) {
        if (tmp > slope * 3)
          tmp = slope * 3;
        sx[(tmp*5) / slope]++;
        rough++;
      } else
      /* find height distribution of land tiles */
      if ((tmp = height_map[xy] - thills) >= 0) {
        if (tmp > 6000)
          tmp = 6000;
        hx[tmp / 400]++;
        mount++;
      }
    }
    slope= ((rough < (total - tmount) ? (20 - j) : (20 + j)) * slope) / 20;
    thills= ((mount < tmount ? (20 - j) : (20 + j)) * thills) / 20;
    thills= (thills >= maxval) ? (maxval-1) : thills;
    thills= (thills <= tocean) ? (tocean+1) : thills;
  }
  trough = rough;

  /* Crude estimation of distribution point for the top third */
  mount /= 3;
  for (j = 16; (j-- > 0) && ((mount * 2) > hx[j]);) mount -= hx[j];
  mount = ((j+1) * 400);
  rough /= 3;
  for (j = 16; (j-- > 0) && ((rough * 2) > sx[j]);) rough -= sx[j];
  rough = ((j+1) * slope / 5);

  for (k = txy; k-- > 0;) {
    int xy = yx[k];
    CIVTILE *ptile = (civmap_get_tiles(pmap) + xy);

    if ((tmp = hs[xy] - slope) >= 0) {
      /* rough terrain */
      if (tmp >= rough)
        civtile_set_terrain(ptile, T_MOUNTAINS);
      else
        civtile_set_terrain(ptile, T_HILLS);
    } else
    if ((tmp = height_map[xy] - thills) >= 0) {
      /* high terrain */
      if (myrand(total) < tmount ) {
        /* high - i.e. the usual mountain thing */
        if (tmp >= mount)
          civtile_set_terrain(ptile, T_MOUNTAINS);
        else
          civtile_set_terrain(ptile, T_HILLS);
      } else {
        /* high - i.e. substitute "flat" alpine (cold/dry) terrain types */
	int x, y;
        index_to_civmap_pos(pmap, &x, &y, &x, &y, xy);
        j= (tmp * 100) / thills / 2;
        j= (j < 40) ? 50 - j : 10;
        j=+ (y * 100) / pmap->ysize, j= (j > 70 ? j-70 : 30-j) / 2;
        if ((tmp = myrand(100)) < j || myrand(100) < j)
          civtile_set_terrain(ptile, T_ARCTIC);
        else if (tmp < j*2)
          civtile_set_terrain(ptile, T_DESERT);
        else if (tmp < j*2 + 50)
          civtile_set_terrain(ptile, T_PLAINS);
      }
    }
  }

  free(hs);
  free(yx);
}

/**************************************************************************
 add arctic and tundra squares in the arctic zone.
 (that is the top 10%, and low 10% of the map)
 Note: current server code "checks" for these things - should be optional
**************************************************************************/
static void make_polar(const CIVMAP *pmap)
{
  CIVTILE *ptile;
  enum terrain_type ttype;
  int n1, y1, y, x, xy, polar;

  /* Only arctic and tundra allowed at the poles (first and last two lines,
   * as defined in make_passable() ), to be consistent with generators 2,3,4.
   * turn every land tile on the 1st and 2nd lines into arctic or tundra,
   *
   * Also turn some high + polar latitude elevations as well
   */
  polar = (civmap_get_type(pmap, (MAP_TYPE_ISO|MAP_TYPE_HEX)) ? 3 : 1);
  for (y= 0; y < pmap->ysize/10; y++) {
    y1= pmap->ysize-1 - y;
    n1= (pmap->ysize/10-y)*1000;     /* pseudo-height <- arctic latitude */
    for (x= 0; x < pmap->xsize; x++) {
      xy = native_pos_to_index(pmap, x, y);
      ptile = (civmap_get_tiles(pmap) + xy);
      ttype = civtile_get_terrain(ptile);
      if (( pmap->separatepoles
         && (y <= polar || (ttype == T_OCEAN && myrand(6)+y < 3)) )
       || (ttype == T_GRASSLAND && myrand(maxval) < height_map[xy]+n1)) {
        if (myrand(y*50+1) + myrand(y*50+1) < 60)
          civtile_set_terrain(ptile, T_ARCTIC);
        else
          civtile_set_terrain(ptile, T_TUNDRA);
      }
      xy = native_pos_to_index(pmap, x, y1);
      ptile = (civmap_get_tiles(pmap) + xy);
      ttype = civtile_get_terrain(ptile);
      if (( pmap->separatepoles
         && (y <= polar || (ttype == T_OCEAN && myrand(6)+y < 3)) )
       || (ttype == T_GRASSLAND && myrand(maxval) < height_map[xy]+n1)) {
        if (myrand(y*50+1) + myrand(y*50+1) < 60)
          civtile_set_terrain(ptile, T_ARCTIC);
        else
          civtile_set_terrain(ptile, T_TUNDRA);
      }
    }
  }
}

/**************************************************************************
  a recursive function that adds desert to the current location and tries
  to spread out to the neighbours, it's called from make_deserts until
  enough desert has been planted. Use the map heights to make a desert
  irregularly shaped. diff is the recursion stopper, and is reduced more if
  desert wants to grow in the y direction, so we end up with "wide" deserts.
**************************************************************************/
static int make_desert(const CIVMAP *pmap, int x, int y, int height, int diff)
{
  CIVTILE *ptile = civmap_get_tile(pmap, x, y);
  int ret = 0;
  if( civtile_get_terrain(ptile) == T_GRASSLAND ) {
    ret++, civtile_set_terrain(ptile, T_DESERT);
    if (abs(hmap(pmap, x, y) - height) < diff) {
      cartesian_adjc_iterate(pmap, x, y, x1, y1) {
        ret+= make_desert(pmap, x1, y1, height, diff - (x != x1 ? 100 : 300));
      } cartesian_adjc_iterate_end;
    }
  }
  return ret;
}

/*************************************************************************
  make_deserts calls make_desert until we have enough deserts what happens
  is that we choose a random coordinate between 20 and 30 degrees north and
  south (deserts tend to be between 15 and 35, make_desert will expand them)
  and if it's a grassland square we call make_desert with this coordinate,
  we try this for each hemisphere until the count is exceeded, or we have
  tried roughly 4 times per count.
**************************************************************************/
static void make_deserts(const CIVMAP *pmap)
{
  CIVTILE *ptile;
  int x,y, xy, xm, ym, deserts, forever, nn= pmap->ysize * 7;

  /* temperate latitudes: 20-40%, 60-80% centered on 30,70% ysize */
  forever= (pmap->xsize * pmap->ysize * pmap->landpercent) / 25;
  deserts= (forever * pmap->DESERTS) / 400;
  while (--forever >= 0 && deserts > 0) {
    y= (pmap->ysize*60 + myrand(nn)+myrand(nn))/100;
    x= myrand(pmap->xsize);
    xy = native_pos_to_index(pmap, x, y);
    ptile = (civmap_get_tiles(pmap) + xy);
    if (civtile_get_terrain(ptile) == T_GRASSLAND) {
      native_to_civmap_pos(pmap, &xm, &ym, x, y);
      deserts-= make_desert(pmap, xm, ym, height_map[xy], 1000);
    }
    y= (pmap->ysize*40 - myrand(nn)-myrand(nn))/100;
    x= myrand(pmap->xsize);
    xy = native_pos_to_index(pmap, x, y);
    ptile = (civmap_get_tiles(pmap) + xy);
    if (civtile_get_terrain(ptile) == T_GRASSLAND) {
      native_to_civmap_pos(pmap, &xm, &ym, x, y);
      deserts-= make_desert(pmap, xm, ym, height_map[xy], 1000);
    }
  }
}

/**************************************************************************
  a recursive function that adds forest to the current location and tries
  to spread out to the neighbours, it's called from make_forests until
  enough forest has been planted. diff is again the block function.
  Near the poles, forests may be nothing more than low brush swamps,
  If we're close to equator it will with 50% chance generate jungle instead
  decrease slightly in the x direction to reflect east west rainfall loss
  TODO: adjust FOREST:JUNGLE ratio, based on wetness/temperature parameters
**************************************************************************/
static int make_forest(const CIVMAP *pmap, int xm, int ym, int height,int diff)
{
  CIVTILE *ptile;
  int x, y, xy, nn, forests = 0;

  civmap_to_native_pos(pmap, &x, &y, xm, ym);
  /* too cold for forests */
  if (y < 3 || y >= pmap->ysize-3)
    return forests;

  xy = native_pos_to_index(pmap, x, y);
  ptile = (civmap_get_tiles(pmap) + xy);
  if (civtile_get_terrain(ptile) == T_GRASSLAND) {
    nn= y*100 / pmap->ysize;
    nn= (nn > 50 ? 100-nn : nn) + 1;
    nn= myrand(nn) + myrand(nn);
    civtile_set_terrain(ptile,
      (nn > 40 ? T_JUNGLE : (nn > 10 ? T_FOREST : T_SWAMP)));
    if (abs(height_map[xy] - height) < diff) {
      cartesian_adjc_iterate(pmap, xm, ym, x1, y1) {
        if (myrand(10) < 4)
          forests += 
	    make_forest(pmap, x1, y1, height, diff - (x != x1 ? 600 : 400));
      } cartesian_adjc_iterate_end;
    }
    forests++;
  }
  return forests;
}

/**************************************************************************
  makeforest calls make_forest with random grassland locations until there
  has been made enough forests. (the pmap->forestsize value controls this)
**************************************************************************/
static void make_forests(const CIVMAP *pmap)
{
  int x,y, xm, ym, forever, forests;
  forever= (pmap->xsize*pmap->ysize*pmap->landpercent)/50;
  forests= (forever*pmap->FORESTS)/200;
  while( forests > 0 && --forever >= 0 ) {
    /* boreal and equatorial bands */
    y= (myrand(pmap->ysize*2)+(myrand(3)*3+1)*pmap->ysize)/10;
    x= myrand(pmap->xsize);
    native_to_civmap_pos(pmap, &xm, &ym, x, y);
    forests -= make_forest(pmap, xm, ym, hmap(pmap, xm, ym), 2500);
  }
}

/**************************************************************************
  swamps are placed on low lying locations, that will typically be close to
  the shoreline. They're put at random (where there is lowlying grassland)
  and with 50% chance each of it's neighbour squares will be converted to
  swamp as well.
  Note: some swamps are really lakes. Adjust the ratio from input some day.
**************************************************************************/
static void make_swamps(const CIVMAP *pmap)
{
  CIVTILE *ptile;
  int xy, x, y, xm, ym, xs, ys, t0, tt, swamps, forever;

  ys = (pmap->ysize - 4) * pmap->xsize;
  xs = 2 * pmap->xsize;
  forever= (pmap->xsize*pmap->ysize*pmap->landpercent)/50;
  swamps= (forever*pmap->SWAMPS)/200;
  while (--forever >= 0 && swamps > 0) {
    xy = myrand(ys) + xs;
    index_to_civmap_pos(pmap, &xm, &ym, &x, &y, xy);
    ptile = (civmap_get_tiles(pmap) + xy);
    tt= civtile_get_terrain(ptile);
    t0= T_SWAMP;
    if (tt == T_GRASSLAND
     && ( height_map[xy] < thills
       || count_terrainN(pmap, xm, ym, 1, T_OCEAN) > 2
       || (t0= T_OCEAN) == T_OCEAN)) {
      swamps--, civtile_set_terrain(ptile, t0);
      adjc_iterate(pmap, xm, ym, x1, y1) {
        ptile = civmap_get_tile(pmap, x1, y1);
        tt= civtile_get_terrain(ptile);
        if( (tt == T_GRASSLAND && myrand(10) < 5)
         || (tt == T_JUNGLE && myrand(10) < 3)
         || (tt == T_FOREST && myrand(10) < 2) 
         || (tt == T_TUNDRA && myrand(10) < 1) )
          swamps--, civtile_set_terrain(ptile, t0);
      } adjc_iterate_end;
    }
  }
}

/*************************************************************************
 this recursive function tries to make some decent rivers, that run towards
 the ocean, it does this by following a descending path on the map spiced
 with a bit of chance, if it fails to reach the ocean it rolls back.
**************************************************************************/
static int make_river(const CIVMAP *pmap, int x, int y, int force)
{
  static int ix[]= { 0, 250, 0, 300, 0, 250, 0, 200 };
  static int rlength= 0, olength= 0, ovalue= 0;
  CIVTILE *ptile, *ptile0;
  enum terrain_type tt, tx, ty, tz;
  int res, min0, min1, osave;
  int nr, no, m0, m1, xy, xx, yy, x2, y2;

  /* don't make rivers in bordertiles - it simplifies things */
  if (civmap_is_border_pos(pmap, x,y))
    return 0;

  /* oops checks */
  xy = civmap_pos_to_index(pmap, x, y);
  ptile0 = (civmap_get_tiles(pmap) + xy);
  tt= civtile_get_terrain(ptile0);
  if (tt >= 16)
    return 0;
  if (civtile_has_specials(ptile0, S_RIVER))
    return 0;

  /* (Re)set ovalue to current tile, unless we have passed a lake,
   *   then take the minimum positive value.
  */
  no= civtile_get_ocean(ptile0);
  if( olength == rlength || (no > 0 && (no < ovalue || ovalue == 0)) )
    ovalue= no;

  /* collect surrounding river/ocean stats
   * nr    - number of adjacent river squares
   * no    - number of adjacent ocean squares (i.e. THE ocean)
   * m0,xx - number of adjacent sea squares, and a direction
   * m1,yy - number of diagonal sea squares, and a direction
   * TODO: collect the highest and lowest ocean value+direction
  */
  res= -1, xx= -1, yy= -1, m0= 0, m1= 0, no= 0, nr= 0;
  cartesian_diagonal_dir_iterate(pmap, x, y, x1, y1, dir) {
    ptile = civmap_get_tile(pmap, x1, y1);
    no+= (civtile_get_ocean(ptile) == 1 ? 1 : 0);
    ty= (civtile_get_terrain(ptile) & 15);
    m1+= (ty == T_OCEAN ? (yy= dir, 1) : 0);
  } cartesian_diagonal_dir_iterate_end

  cartesian_adjc_dir_iterate(pmap, x, y, x1, y1, dir) {
    ptile = civmap_get_tile(pmap, x1, y1);
    no+= (civtile_get_ocean(ptile) == 1 ? 1 : 0);
    tx= (civtile_get_terrain(ptile) & 15);
    m0+= (tx == T_OCEAN ? (res= xx, xx= dir, 1) : 0);

    if( civtile_has_specials(ptile, S_RIVER) ) {
       /* do not end next to (i.e. make river fork in) these terrain */
       switch(tx) {
       case T_ARCTIC:
       case T_MOUNTAINS:
         nr++;
       default:
         break;
       }
       res= xx, xx= dir, nr++;
    }
  } cartesian_adjc_dir_iterate_end

  /* terminate on THE ocean, not a tiny lake
   * Note: if we continue, lakes have better connectivity (+diagonals)
   *       than rivers
   */
  if ( tt == T_OCEAN ) {
    if( ovalue == 1)
      return 1;    /* hit THE ocean */
    if( (m0+= m1) > 1 )
      return 1;   /* big lake, not a pseudo-river - too many choices */
    if( xx == -1 )
      xx= yy;
    else
    if( res == -1 )
      res= yy;
  }

  if ( !force ) {
    /* did we enter a tile bordered by river(s) and/or ocean? */
    switch(nr+(m0 > 0 ? 1 : 0)) {
    default:
      /* no double outflows */
      return 0;
    case 2:
      /* join up inland river/lake */
      if( res == -1 ) res= yy;
      if( res == -1 )
        return 0;
      if (civmap_step(pmap, &x2, &y2, x, y, res, 0)) {
        ty= civmap_get_ocean(pmap, x2, y2);
        if (civmap_step(pmap, &x2, &y2, x, y, xx, 0)) {
          tx= civmap_get_ocean(pmap, x2, y2);
          if( tx == ty
           || tx == ovalue
           || ty == ovalue
           || m0 == 0
           || (m0+nr) != 2
           || (m0+nr)+ty == no )
            return 0;
        }
      }
      m0= 0; /* and FALL THROUGH */
    case 1:
      if (civmap_step(pmap, &x2, &y2, x, y, xx, 0)) {
        tx= civmap_get_ocean(pmap, x2, y2);
        if(m0 > 0) {
          /* can follow a single sea, it is probably a lake */
          if( m0 == 1
           && no == 0
           && (rlength == olength || myrand(olength)<2) )
            break;

          /* join *exactly* two different source lakes
           * There is a hole here. 3-6 adjacent oceans can also be part of
           * 2 distinct seas, and there is no check for "different" in these
           * cases, allowing double outlet through a short river. Because it
           * is difficult to recognize which 2 are different, and the 3-6
           * cases have value on its own most of the time, we let this go.
           */
          if( m0 == 2 ) {
            if (civmap_step(pmap, &x2, &y2, x, y, res, 1)) {
              ty= civmap_get_ocean(pmap, x2,y2);
              if( tx == ty
               || tx == ovalue
               || ty == ovalue )
                return 0;
            }
          }
          else /* bias against short coastal creeks */
          if( myrand(3)+(tt == T_OCEAN ? 1 : 0) >= rlength )
            return 0;
        }
      }
      /* join up! */
      if (tt != T_OCEAN) {
        civtile_special_add(ptile0, S_RIVER);
      }
      civtile_set_ocean(ptile0, 0);
      ocean_flood1(pmap, x, y, -1, tx);
      ovalue= civtile_get_ocean(ptile0);
      return 1;
    case 0:
      /* keep on paddling */
      break;
    }
  }

  /* probe two directions for next possible river tile and recurse */
  /* temporarily increment the counts and block the current square */
  osave = olength, olength++, rlength++;
  civtile_set_terrain(ptile0, tt|16);

  if( m0 == 1 ) {
    /* no choice - step into the lake (or fjord) - try twice */
    m0= m1= xx;
    olength = 0;
  } else {
    /* extend the river - slight bias to the east (ix), 2 tries,
     * bias towards entering any costal square or towards ocean,
     * bias against entering arctic, mountains and hills
    */
    min0= hmapN(pmap, x, y, 1)+(tt == T_OCEAN ? 2 : 1)*ix[1];
    m0= m1= -1;
    cartesian_adjc_dir_iterate(pmap, x, y, x1, y1, dir) {
      if( (tx= civmap_get_terrain(pmap, x1, y1)) < 16
       && (tx != T_OCEAN)
       && (tx != T_ARCTIC)
       && (tx != T_MOUNTAINS || myrand(16) < 1)
       && (tx != T_HILLS || myrand(8) < 1)
       && (tx != T_DESERT || myrand(4) < 1)
       && (res= hmapN(pmap, x1,y1,1))-myrand
          (count_terrainN(pmap, x1,y1,1,T_OCEAN)?3:
          (count_terrainN(pmap, x1,y1,2,T_OCEAN)?2:1)*ix[dir]) < min0) {
        min1= min0, m1= m0;
        min0= res,  m0= dir;
      }
    } cartesian_adjc_dir_iterate_end
  }

  for(res=0; res == 0 && m0 != -1; m0= m1, m1= -1) {
    /* block orthogonal land tiles - cartesian (+/-2) or diagonal (+/-1) */
    /* TODO: diagonal sea walk does not block all looping tail biters */
    /*       +/-2 block would help, but probably need +/-3 as well    */
    nr= (m0 & 1)+1;
    if (civmap_step(pmap, &xx, &yy, x, y, DIR_CCW_N(m0,nr), 0)
     && (ptile = civmap_get_tile(pmap, xx, yy),
	 tx = civtile_get_terrain(ptile)) != T_OCEAN) {
      civtile_set_terrain(ptile, tx|16);
    }
    if (civmap_step(pmap, &xx, &yy, x, y, DIR_CW_N(m0,nr), 0)
     && (ptile = civmap_get_tile(pmap, xx, yy),
	 ty = civtile_get_terrain(ptile)) != T_OCEAN) {
      civtile_set_terrain(ptile, ty|16);
    }
    if (civmap_step(pmap, &xx, &yy, x, y, DIR_REVERSE(m0), 0)
     && (ptile = civmap_get_tile(pmap, xx, yy),
	 tz = civtile_get_terrain(ptile)) != T_OCEAN) {
      /* and looping tail biters */
      civtile_set_terrain(ptile, tz|16);
    }

    /* finally step ... */
    if( civmap_step(pmap, &xx, &yy, x, y, m0, 1))
      res= make_river(pmap, xx, yy, 0);

    /* unblock neighbours in reverse order */
    if (civmap_step(pmap, &xx, &yy, x, y, DIR_REVERSE(m0), 0))
      civtile_set_terrain(civmap_get_tile(pmap, xx, yy), tz);
    if (civmap_step(pmap, &xx, &yy, x, y, DIR_CW_N(m0,nr), 0))
      civtile_set_terrain(civmap_get_tile(pmap, xx, yy), ty);
    if (civmap_step(pmap, &xx, &yy, x, y, DIR_CCW_N(m0,nr), 0))
      civtile_set_terrain(civmap_get_tile(pmap, xx, yy), tx);
  }
  civtile_set_terrain(ptile0, tt);
  olength=osave, rlength--;

  /* back from recurse, mark this tile if successful and return result */
  if (res || force || tt == T_OCEAN) {
    res++;
    if (tt != T_OCEAN) {
      civtile_special_add(ptile0, S_RIVER);
    }
    civtile_set_ocean(ptile0, 0);
    ocean_flood1(pmap, x, y, -1, ovalue);
    ovalue= civtile_get_ocean(ptile0);
  }
  return res;
}

/**************************************************************************
  calls make_river until we have enough river tiles (pmap->riverlength),
  or forever patience runs out.
  Finally, try one last time to connect isolated inland watersheds.
**************************************************************************/
static void make_rivers(const CIVMAP *pmap)
{
  CIVTILE *ptile, *ptil2;
  enum terrain_type tx, tt;
  int m0, m1, no, nr;
  int x, y, xy, yxsize, *yx, k, land, rivers, forever;

  civmap_ocean_assign((CIVMAP *)pmap);

  /* collect a list of eligible land tiles */
  yxsize= pmap->ysize*pmap->xsize - pmap->xsize*2;
  yx= fc_malloc(sizeof(int)*yxsize);
  for( k = 0, xy = yxsize + pmap->xsize; --xy >= pmap->xsize; ) {
    ptile = (civmap_get_tiles(pmap) + xy);
    if (civtile_get_terrain(ptile) != T_OCEAN)
      yx[k++]= xy;
  }
  land = k;

  /* Assume pmap->RIVERS > 50 is an absolute number of squares         */
  /*   else treat values <=50 as percent landmass                      */
  /* Note: rivers don't pack well (circular flow) - 1 in 4 is optimal. */
  if( pmap->RIVERS < 50 )
    rivers= (pmap->xsize*pmap->ysize*pmap->landpercent*pmap->RIVERS)/10000;
  else
    rivers= pmap->RIVERS;

  /* River generation is quite picky, try roughly 4 times per tile */
  forever= 4;
  while ((k > 0 || (k= land, --forever) > 0) && rivers > 0) {
    y= myrand(k--), xy= yx[y], yx[y]= yx[k], yx[k]= xy;
    index_to_civmap_pos(pmap, &x, &y, &x, &y, xy);

    ptile = (civmap_get_tiles(pmap) + xy);
    tt= civtile_get_terrain(ptile);

    /* collect surrounding river/ocean stats
     * nr - number of adjacent river squares
     * m0 - number of adjacent ocean squares
     * m1 - number of diagonal ocean squares
    */
    m0= 0, m1= 0, nr= 0, no= 0;
    cartesian_diagonal_dir_iterate(pmap, x, y, x1, y1, dir) {
      ptil2 = (civmap_get_tiles(pmap) + civmap_pos_to_index(pmap, x1, y1));
      no+= (civtile_get_ocean(ptil2) == 1 ? 1 : 0);
      tx= civtile_get_terrain(ptil2);
      m1+= (tx == T_OCEAN ? 1 : 0);
    } cartesian_diagonal_dir_iterate_end

    cartesian_adjc_dir_iterate(pmap, x, y, x1, y1, dir) {
      ptil2 = (civmap_get_tiles(pmap) + civmap_pos_to_index(pmap, x1, y1));
      no+= (civtile_get_ocean(ptil2) == 1 ? 1 : 0);
      tx= civtile_get_terrain(ptil2);
      m0+= (tx == T_OCEAN ? 1 : 0);

      if( civtile_has_specials(ptil2, S_RIVER) ) {
        /* do not end next to (i.e. make river fork in) these terrain */
        switch(tx) {
        case T_MOUNTAINS:
        case T_ARCTIC:
          nr++;
        default:
          break;
        }
        nr++;
      }
    } cartesian_adjc_dir_iterate_end

    if( (m0 + nr) == 2
     && (m0 + nr) + m1 != no
     && tt != T_OCEAN
     && !(civtile_has_specials(ptile, S_RIVER)) )
      ;  /* join lakes */
    else /* when not to start here */
    if( height_map[xy] < myrand((maxval*60)/100) /* bias to high elevations */
     || tt == T_OCEAN
     || tt == T_DESERT
     || (tt == T_MOUNTAINS && myrand(16) > 0)
     || (tt == T_HILLS && myrand(8) > 0)
     || (tt == T_TUNDRA && myrand(4) > 0)
     || (civtile_has_specials(ptile, S_RIVER))
     || (((m0 += m1) > 0 ? 1 : 0) + nr > 1)      /* bad water connectivity */
     || (m0 > 0 && myrand(4) > (m0-no))          /*   quickie tests ...    */
     || ( count_terrainN(pmap, x,y,2,T_OCEAN)
        + count_terrainN(pmap, x,y,2,T_RIVER) > 3 && myrand(3) > 0)
    )
      continue;

    rivers-=make_river(pmap, x,y,0);
  }

  /* for each *inland* watershed
   *   list tiles
   *   for each tile
   *     walk adjacent2 (cartesian > knights > diagonal preference)
   *       if is_different_watershed then
   *         join up via intervening adjacent tile
  */
  setup_seadata((CIVMAP *)pmap);
  for( k= pmap->num_oceans+1; --k > 1; ) {
    int x2, y2;
    /* retest for the case where we just joined with the first ocean */
    if( (m0= civmap_get_ocean(pmap, oceans[k].x, oceans[k].y)) == 1 )
      continue;
    /* setup and collect the list of watershed tiles */
    yx[0]= yxsize-2, yx[1]= 0;
    rivers= ocean_list(pmap, yx+1, oceans[k].x, oceans[k].y);
    /* walk the watershed looking for adjacent tiles that touch 2 watersheds */
    for( rivers++; --rivers > 0; ) {
      x= yx[rivers+1], index_to_civmap_pos(pmap, &x, &y, &x, &y, x);
      /* Exactness not required - this is easier than fixing each case */
      if( civmap_is_border_pos(pmap, x,y) )
        continue;
      tx= T_OCEAN, tt= civmap_get_terrain(pmap, x, y);

      /* find nearby inland waterways */
      no= -1;
      /* first check for possible connecting rivers */
      cartesian_adjc_dir_iterate2(pmap, x, y, x1, y1, dir) {
        if( (m1= civmap_get_ocean(pmap, x1,y1)) > 0 && m1 != m0 ) {
          tx= civmap_get_terrain(pmap, x1,y1), no= dir/2;
          if( ( !civmap_step(pmap, &x2, &y2, x, y, DIR_CCW(no), 0)
              || civmap_get_ocean(pmap, x2, y2) == 0)
           && ( !civmap_step(pmap, &x2, &y2, x, y, DIR_CW(no), 0)
              || civmap_get_ocean(pmap, x2, y2) == 0) )
            break;
          no= -1;
        }
      } cartesian_adjc_dir_iterate2_end;
      if( no == -1 ) {
        /* next check for possible river (or ocean) to ocean connect */
        cartesian_knights_dir_iterate2(pmap, x, y, x1, y1, dir) {
          if( (tt == T_OCEAN || civmap_get_terrain(pmap, x1,y1) == T_OCEAN)
           && (m1= civmap_get_ocean(pmap, x1,y1)) > 0
           &&  m1 != m0 ) {
            tx= T_OCEAN, no= (tt == T_OCEAN ? ((dir+1)&-3) : ((dir&-3)+2))/2;
            if( ( !civmap_step(pmap, &x2, &y2, x, y, DIR_CCW(no), 0)
                || civmap_get_ocean(pmap, x2, y2) == 0)
             && ( !civmap_step(pmap, &x2, &y2, x, y, DIR_CW(no), 0)
                || civmap_get_ocean(pmap, x2, y2) == 0) )
              break;
            no= -1;
          }
        } cartesian_knights_dir_iterate2_end;
      }
      if( no == -1 ) {
        /* last check for possible ocean to ocean diagonal connect */
        cartesian_diagonal_dir_iterate2(pmap, x, y, x1, y1, dir) {
          if( (tt == T_OCEAN && civmap_get_terrain(pmap, x1,y1) == T_OCEAN)
           && (m1= civmap_get_ocean(pmap, x1,y1)) > 0
           &&  m1 != m0 ) {
            tx= T_OCEAN, no= dir/2;
            if( ( !civmap_step(pmap, &x2, &y2, x, y, DIR_CCW(no), 0)
                || civmap_get_ocean(pmap, x2, y2) == 0)
             && ( !civmap_step(pmap, &x2, &y2, x, y, DIR_CW(no), 0)
                || civmap_get_ocean(pmap, x2, y2) == 0) )
              break;
            no= -1;
          }
        } cartesian_diagonal_dir_iterate2_end;
      }
      if( no != -1 && civmap_step(pmap, &x2, &y2, x, y, no, 0) ) {
        /* connect watersheds via intervening adjacent tile */
        ptil2 = civmap_get_tile(pmap, x2, y2);
        if( tx == T_OCEAN
         || tx == T_ARCTIC
         || tx == T_MOUNTAINS ) {
          civtile_special_delete(ptil2, S_RIVER);
          civtile_set_terrain(ptil2, T_OCEAN);
          civtile_set_continent(ptil2, 0);
        }
        else
          civtile_special_add(ptil2, S_RIVER);

        civtile_set_ocean(ptil2, 0);
        ocean_flood1(pmap, x2, y2, -1, m0);
        m0= civtile_get_ocean(ptil2);
      }
    }
  }
  free(yx);
}

/**************************************************************************
  make_plains converts ~50% of the remaining grassland to plains based on
  height and latitude (plains are more likely at desert height/latitudes).
  TODO: change the ratio based on wetness or lushness parameter.
**************************************************************************/
static void make_plains(const CIVMAP *pmap)
{
  CIVTILE *ptile;
  int x, y, y0, y1, xy, nn, ng, np;

  do {
    ng= np= 0;
    for (y0 = pmap->ysize/2; --y0 > 0; ) {
      y1 = pmap->ysize-1 - y0;
      y  = (y0*10000)/pmap->ysize;   /* latitude to pseudo altitude */
      nn = (y < 2000 ? (4000-y) : y);
      for (x= 0; x < pmap->xsize; x++) {
	xy = native_pos_to_index(pmap, x, y0);
        ptile = (civmap_get_tiles(pmap) + xy);
        if( civtile_get_terrain(ptile) == T_GRASSLAND ) {
          if (myrand(height_map[xy] + 7000) > tocean + nn)
            civtile_set_terrain(ptile, (myrand(2500) > y ? T_TUNDRA:T_PLAINS));
          else
          if (myrand(2000) > y)
            civtile_set_terrain(ptile, T_TUNDRA);
        }
        ng+= civtile_get_terrain(ptile) == T_GRASSLAND ? 1 : 0;
        np+= civtile_get_terrain(ptile) == T_PLAINS ? 1 : 0;

	xy = native_pos_to_index(pmap, x, y1);
        ptile = (civmap_get_tiles(pmap) + xy);
        if( civtile_get_terrain(ptile) == T_GRASSLAND ) {
          if (myrand(height_map[xy] + 7000) > tocean + nn)
            civtile_set_terrain(ptile, (myrand(2500) > y ? T_TUNDRA:T_PLAINS));
          else
          if (myrand(2000) > y)
            civtile_set_terrain(ptile, T_TUNDRA);
        }
        ng+= civtile_get_terrain(ptile) == T_GRASSLAND ? 1 : 0;
        np+= civtile_get_terrain(ptile) == T_PLAINS ? 1 : 0;
      }
    }
  } while (ng*85 > np*100);
}

/**************************************************************************
  we want the map to be sailable east-west at least at north and south pole
  and make it a bit jagged at the edge as well.
  So this procedure converts 60% of the second line and the second last line
  to ocean, 40% of the 4th and 4th last line to ocean, and 50% of the 3rd
  and 3rd last line to ocean (100% if the 4th was not ocean).
**************************************************************************/
static void make_passable(const CIVMAP *pmap)
{
  CIVTILE *ptile, *ptil2;
  bool is_iso = (civmap_get_type(pmap, (MAP_TYPE_ISO|MAP_TYPE_HEX)) != 0);
  int xx, x0= 0, x= pmap->xsize, yy = pmap->ysize;
  int xp1 = 1, xm1 = -1;

  /* Handle edge cases specially if needed */
  if( !civmap_has_wrap(pmap, MAP_TYPE_WRAPX) ) {
    x--;
    ptile = (civmap_get_tiles(pmap) + native_pos_to_index(pmap, x, 2));
    civtile_set_terrain(ptile, T_OCEAN);
    ptile = (civmap_get_tiles(pmap) + native_pos_to_index(pmap, x, yy-3));
    civtile_set_terrain(ptile, T_OCEAN);
    ptile = (civmap_get_tiles(pmap) + native_pos_to_index(pmap, x0, 2));
    civtile_set_terrain(ptile, T_OCEAN);
    ptile = (civmap_get_tiles(pmap) + native_pos_to_index(pmap, x0, yy-3));
    civtile_set_terrain(ptile, T_OCEAN);
    if (is_iso) {
      ptile = (civmap_get_tiles(pmap) + native_pos_to_index(pmap, x, 3));
      civtile_set_terrain(ptile, T_OCEAN);
      ptile = (civmap_get_tiles(pmap) + native_pos_to_index(pmap, x, yy-4));
      civtile_set_terrain(ptile, T_OCEAN);
      ptile = (civmap_get_tiles(pmap) + native_pos_to_index(pmap, x0, 3));
      civtile_set_terrain(ptile, T_OCEAN);
      ptile = (civmap_get_tiles(pmap) + native_pos_to_index(pmap, x0, yy-4));
      civtile_set_terrain(ptile, T_OCEAN);
    }
    x0++;
  } else {
    /*
     * Only the first and last case can wrap on left/center/right checks.
     * Do this or use civmap_get_terrain_checked()
     */
    x--;
    if (is_iso)
      xp1 = (x & 1), xm1 = xp1 - 1;

    ptile = (civmap_get_tiles(pmap) + native_pos_to_index(pmap, x, 1));
    ptil2 = (civmap_get_tiles(pmap) + native_pos_to_index(pmap, x, 3));
    if (!is_iso) {
      if (myrand(100) < 40)
        civtile_set_terrain(ptile, T_OCEAN);
      if (myrand(100) < 60)
        civtile_set_terrain(ptil2, T_OCEAN);
    } else {
      if (civtile_get_terrain(ptile) != T_OCEAN || myrand(100) < 60)
        civtile_set_terrain(ptil2, T_OCEAN);
      ptil2 = (civmap_get_tiles(pmap) + native_pos_to_index(pmap, x, 4));
      if (myrand(100) < 80)
        civtile_set_terrain(ptil2, T_OCEAN);
    }

    xx = (xm1 < 0 ? 0 : x);
    if (civtile_get_terrain(ptil2) != T_OCEAN
     || ((ptil2 = civmap_get_tiles(pmap) + native_pos_to_index(pmap, xx, 3)),
          civtile_get_terrain(ptil2) != T_OCEAN)
     || ((ptil2 = civmap_get_tiles(pmap) + native_pos_to_index(pmap, x-xp1, 3)),
          civtile_get_terrain(ptil2) != T_OCEAN)
     || myrand(100) < 50)
      ptil2 = (civmap_get_tiles(pmap) + native_pos_to_index(pmap, x, 2)),
        civtile_set_terrain(ptil2, T_OCEAN);

    ptile = (civmap_get_tiles(pmap) + native_pos_to_index(pmap, x, yy-2));
    ptil2 = (civmap_get_tiles(pmap) + native_pos_to_index(pmap, x, yy-4));
    if (!is_iso) {
      if (myrand(100) < 40)
        civtile_set_terrain(ptile, T_OCEAN);
      if (myrand(100) < 60)
        civtile_set_terrain(ptil2, T_OCEAN);
    } else {
      if (civtile_get_terrain(ptile) != T_OCEAN || myrand(100) < 60)
        civtile_set_terrain(ptil2, T_OCEAN);
      ptil2 = (civmap_get_tiles(pmap) + native_pos_to_index(pmap, x, yy-5));
      if (myrand(100) < 80)
        civtile_set_terrain(ptil2, T_OCEAN);
    }

    xx = (xp1 > 0 ? 0 : x);
    if (civtile_get_terrain(ptil2) != T_OCEAN
     || ((ptil2= civmap_get_tiles(pmap) + native_pos_to_index(pmap,x+xm1,yy-4)),
          civtile_get_terrain(ptil2) != T_OCEAN)
     || ((ptil2= civmap_get_tiles(pmap) + native_pos_to_index(pmap, xx, yy-4)),
          civtile_get_terrain(ptil2) != T_OCEAN)
     || myrand(100) < 50)
      ptil2 = (civmap_get_tiles(pmap) + native_pos_to_index(pmap, x, yy-3)),
        civtile_set_terrain(ptil2, T_OCEAN);

    /* ... first */
    if (is_iso)
      xp1 = (x0 & 1), xm1 = xp1 - 1;

    ptile = (civmap_get_tiles(pmap) + native_pos_to_index(pmap, x0, 1));
    ptil2 = (civmap_get_tiles(pmap) + native_pos_to_index(pmap, x0, 3));
    if (!is_iso) {
      if (myrand(100) < 40)
        civtile_set_terrain(ptile, T_OCEAN);
      if (myrand(100) < 60)
        civtile_set_terrain(ptil2, T_OCEAN);
    } else {
      if (civtile_get_terrain(ptile) != T_OCEAN || myrand(100) < 60)
        civtile_set_terrain(ptil2, T_OCEAN);
      ptil2 = (civmap_get_tiles(pmap) + native_pos_to_index(pmap, x0, 4));
      if (myrand(100) < 80)
        civtile_set_terrain(ptil2, T_OCEAN);
    }

    xx = (xp1 > 0 ? pmap->xsize-1 : x0);
    if (civtile_get_terrain(ptil2) != T_OCEAN
     || ((ptil2= civmap_get_tiles(pmap) + native_pos_to_index(pmap, x0-xm1, 3)),
          civtile_get_terrain(ptil2) != T_OCEAN)
     || ((ptil2= civmap_get_tiles(pmap) + native_pos_to_index(pmap, xx, 3)),
          civtile_get_terrain(ptil2) != T_OCEAN)
     || myrand(100) < 50)
      ptil2 = (civmap_get_tiles(pmap) + native_pos_to_index(pmap, x0, 2)),
        civtile_set_terrain(ptil2, T_OCEAN);

    ptile = (civmap_get_tiles(pmap) + native_pos_to_index(pmap, x0, yy-2));
    ptil2 = (civmap_get_tiles(pmap) + native_pos_to_index(pmap, x0, yy-4));
    if (!is_iso) {
      if (myrand(100) < 40)
        civtile_set_terrain(ptile, T_OCEAN);
      if (myrand(100) < 60)
        civtile_set_terrain(ptil2, T_OCEAN);
    } else {
      if (civtile_get_terrain(ptile) != T_OCEAN || myrand(100) < 60)
        civtile_set_terrain(ptil2, T_OCEAN);
      ptil2 = (civmap_get_tiles(pmap) + native_pos_to_index(pmap, x0, yy-5));
      if (myrand(100) < 80)
        civtile_set_terrain(ptil2, T_OCEAN);
    }

    xx = (xm1 < 0 ? pmap->xsize-1 : x0);
    if ( civtile_get_terrain(ptil2) != T_OCEAN
     || ((ptil2= civmap_get_tiles(pmap) + native_pos_to_index(pmap, xx, yy-4)),
          civtile_get_terrain(ptil2) != T_OCEAN)
     || ((ptil2= civmap_get_tiles(pmap)+native_pos_to_index(pmap,x0+xp1,yy-4)),
          civtile_get_terrain(ptil2) != T_OCEAN)
     || myrand(100) < 50)
      ptil2 = (civmap_get_tiles(pmap) + native_pos_to_index(pmap, x0, yy-3)),
        civtile_set_terrain(ptil2, T_OCEAN);
      x0++;
  }

  while(--x >= x0) {
    if (is_iso)
      xp1 = (x & 1), xm1 = xp1 - 1;

    ptile = (civmap_get_tiles(pmap) + native_pos_to_index(pmap, x, 1));
    ptil2 = (civmap_get_tiles(pmap) + native_pos_to_index(pmap, x, 3));
    if (!is_iso) {
      if (myrand(100) < 40)
        civtile_set_terrain(ptile, T_OCEAN);
      if (myrand(100) < 60)
        civtile_set_terrain(ptil2, T_OCEAN);
    } else {
      if (civtile_get_terrain(ptile) != T_OCEAN || myrand(100) < 60)
        civtile_set_terrain(ptil2, T_OCEAN);
      ptil2 = (civmap_get_tiles(pmap) + native_pos_to_index(pmap, x, 4));
      if (myrand(100) < 80)
        civtile_set_terrain(ptil2, T_OCEAN);
    }

    if (civtile_get_terrain(ptil2) != T_OCEAN
     || ((ptil2 = civmap_get_tiles(pmap) + native_pos_to_index(pmap, x-xm1, 3)),
         civtile_get_terrain(ptil2) != T_OCEAN)
     || ((ptil2 = civmap_get_tiles(pmap) + native_pos_to_index(pmap, x-xp1, 3)),
         civtile_get_terrain(ptil2) != T_OCEAN)
     || myrand(100) < 50)
      ptil2 = (civmap_get_tiles(pmap) + native_pos_to_index(pmap, x, 2)),
        civtile_set_terrain(ptil2, T_OCEAN);

    ptile = (civmap_get_tiles(pmap) + native_pos_to_index(pmap, x, yy-2));
    ptil2 = (civmap_get_tiles(pmap) + native_pos_to_index(pmap, x, yy-4));
    if (!is_iso) {
      if (myrand(100) < 40)
        civtile_set_terrain(ptile, T_OCEAN);
      if (myrand(100) < 60)
        civtile_set_terrain(ptil2, T_OCEAN);
    } else {
      if (civtile_get_terrain(ptile) != T_OCEAN || myrand(100) < 60)
        civtile_set_terrain(ptil2, T_OCEAN);
      ptil2 = (civmap_get_tiles(pmap) + native_pos_to_index(pmap, x, yy-5));
      if (myrand(100) < 80)
        civtile_set_terrain(ptil2, T_OCEAN);
    }

    if (civtile_get_terrain(ptil2) != T_OCEAN
     || ((ptil2= civmap_get_tiles(pmap) + native_pos_to_index(pmap,x+xm1,yy-4)),
         civtile_get_terrain(ptil2) != T_OCEAN)
     || ((ptil2= civmap_get_tiles(pmap) + native_pos_to_index(pmap,x+xp1,yy-4)),
         civtile_get_terrain(ptil2) != T_OCEAN)
     || myrand(100) < 50)
      ptil2 = (civmap_get_tiles(pmap) + native_pos_to_index(pmap, x, yy-3)),
        civtile_set_terrain(ptil2, T_OCEAN);
  }
}

/**************************************************************************
  we don't want huge areas of monoterrain like grass/plains,
  so we put in a hill here and there, where it gets too 'clean'.
  Actually we can do a little better than that.
  Updated to clean out tile types the user had explicitly assigned as 0%.
**************************************************************************/
static void make_fair(const CIVMAP *pmap)
{
  CIVTILE *ptile, *ptil2;
  bool is_border;
  int ii, nn, t0, tt, x,y, x1,y1, xy,*yx;

  /* If the user turned off a given control, make sure there are none.
   * In this case, convert to the next non-zero type in some order.
   * NOTE: indexing depends on the order and value of enum_terrain_type.
   *       GRASSLAND is the constraint type, i.e. determined by the
   *       constraint that all percents must sum to 100, NOT pmap->grassize.
   */
  int ix[16], tx[] = {  /* Ordered list of conversion tile types */
    T_TUNDRA,           /* <- T_ARTIC     */
    T_PLAINS,           /* <- T_DESERT    */
    T_HILLS,            /* <- T_FOREST    */
    T_GRASSLAND,        /* <- T_GRASSLAND */
    T_PLAINS,           /* <- T_HILLS     */
    T_FOREST,           /* <- T_JUNGLE    */
    T_HILLS,            /* <- T_MOUNTAINS */
    T_SWAMP,            /* <- T_OCEAN     */
    T_GRASSLAND,        /* <- T_PLAINS    */
    T_PLAINS,           /* <- T_RIVERS    */
    T_JUNGLE,           /* <- T_SWAMPS    */
    T_DESERT,           /* <- T_TUNDRA    */
    0, 0, 0, 0,
                        /* Ordered list of active weights/percentage */
    1,                  /* <- T_ARTIC     */
    pmap->DESERTS,      /* <- T_DESERT    */
    pmap->FORESTS,      /* <- T_FOREST    */
    1,                  /* <- T_GRASSLAND, constraint type MUST BE non-zero! */
    1,                  /* <- T_HILLS     */
    pmap->FORESTS,      /* <- T_JUNGLE    */
    pmap->MOUNTAINS,    /* <- T_MOUNTAINS */
    1,                  /* <- T_OCEAN     */
    1,                  /* <- T_PLAINS    */
    pmap->RIVERS,       /* <- T_RIVERS    */
    pmap->SWAMPS,       /* <- T_SWAMPS    */
    1,                  /* <- T_TUNDRA    */
    0, 0, 0, 0 };

  for( nn= 12; nn-- > 0; ) {
    for( ii= nn; tx[ii+16] == 0; ii= tx[ii]);
    ix[nn]= ii;
  }

  /* randomize hits to avoid NW->SE waves - exclude double borders */
  xy = (pmap->ysize-4)*(pmap->xsize-4);
  yx = (int *)fc_malloc(xy*sizeof(int));
  x = pmap->ysize*2 + pmap->xsize*2;
  for( nn = 0; nn < xy; )
    y = myrand(++nn), yx[nn-1] = yx[y], yx[y] = nn + x;

  while (xy-- > 0) {
    ptile = (civmap_get_tiles(pmap) + yx[xy]);
    index_to_civmap_pos(pmap, &x, &y, &x, &y, yx[xy]);
    is_border = civmap_is_border_pos(pmap, x, y);

    t0= civtile_get_terrain(ptile);
    switch(t0) {
    case T_MOUNTAINS:
      if (is_at_coast(pmap, x, y)) {
        hslope(pmap, &nn, x, y, 0), nn= DIR_REVERSE(nn);
        if( myrand(5) < 4
         && civmap_step(pmap, &x1, &y1, x, y, nn, is_border)
	 && (ptil2= civmap_get_tile(pmap, x1, y1),
             tt= civtile_get_terrain(ptil2)) != T_OCEAN) {
          /* hack - increase coastal plain swap t0 one square inland */
          civtile_set_terrain(ptil2, t0);
          civtile_set_terrain(ptile, (t0= tt));
        }
        nn=2;
      } else
        nn=5;

      if (count_terrainN(pmap, x, y, 2, T_MOUNTAINS)
        + count_terrainN(pmap, x, y, 1, T_HILLS) > nn) {
        for(ii= 2; ii-- > 0; ) {
          if ((myrand(4) < 3)
           && (nn= myrand(8), civmap_step(pmap, &x1, &y1, x, y, nn, is_border))
	   && (ptil2 = civmap_get_tile(pmap, x1, y1),
              (tt= civtile_get_terrain(ptil2)) == T_MOUNTAINS)) {
            tt = (myrand(5)<1 ? T_ARCTIC : (myrand(3)<1 ? T_PLAINS : T_FOREST));
            civtile_set_terrain(ptil2, tt);
          }
        }
      }
      break;
    case T_HILLS:
      if (is_at_coast(pmap, x, y)) {
        hslope(pmap, &nn, x, y, 0), nn= DIR_REVERSE(nn);
        if( myrand(3) < 2
         && civmap_step(pmap, &x1, &y1, x, y, nn, is_border)
	 && (ptil2= civmap_get_tile(pmap, x1, y1),
             tt= civtile_get_terrain(ptil2)) != T_OCEAN) {
          /* hack - increase coastal plain swap t0 one square inland */
          civtile_set_terrain(ptil2, t0);
          civtile_set_terrain(ptile, (t0= tt));
        }
        nn=3;
      } else
        nn=6;

      if (count_terrainN(pmap, x, y, 2, T_HILLS)
        + count_terrainN(pmap, x, y, 1, T_MOUNTAINS) > nn) {
        for(ii= 3; ii-- > 0; ) {
          if ((myrand(4) < 2)
           && (nn= myrand(8), civmap_step(pmap, &x1, &y1, x, y, nn, is_border))
	   && (ptil2 = civmap_get_tile(pmap, x1, y1),
              (tt= civtile_get_terrain(ptil2)) == T_HILLS)) {
            tt = (myrand(3) < 1 ? T_PLAINS : T_FOREST);
            civtile_set_terrain(ptil2, tt);
          }
        }
      }
      break;
    case T_ARCTIC:
    case T_TUNDRA:
    case T_DESERT:
      if (count_terrainN(pmap, x, y, 2, t0) > 20) {
        for(ii= 2; ii-- > 0; ) {
          if ((myrand(3) < 2)
           && (nn= myrand(8), civmap_step(pmap, &x1, &y1, x, y, nn, is_border))
	   && (ptil2 = civmap_get_tile(pmap, x1, y1),
              (tt= civtile_get_terrain(ptil2)) == t0)) {
            tt = (myrand(3) < 1 ? T_MOUNTAINS : T_HILLS);
            civtile_set_terrain(ptil2, tt);
          }
        }
      }
      break;
    case T_JUNGLE:
    case T_FOREST:
      if (count_terrainN(pmap, x, y, 2, T_JUNGLE)
        + count_terrainN(pmap, x, y, 2, T_FOREST) > 20) {
        for(ii= 3; ii-- > 0; ) {
          if ((myrand(3) < 2)
           && (nn= myrand(8), civmap_step(pmap, &x1, &y1, x, y, nn, is_border))
	   && (ptil2 = civmap_get_tile(pmap, x1, y1),
              (tt= civtile_get_terrain(ptil2)) == t0)) {
            tt = (myrand(3)<1 ? T_SWAMP : (myrand(3)<2 ? T_PLAINS : T_HILLS));
            civtile_set_terrain(ptil2, tt);
          }
        }
      }
      break;
    case T_SWAMP:
      if (count_terrainN(pmap, x, y, 2, t0) > 20) {
        for(ii= 3; ii-- > 0; ) {
          if ((myrand(3) < 2)
           && (nn= myrand(8), civmap_step(pmap, &x1, &y1, x, y, nn, is_border))
	   && (ptil2 = civmap_get_tile(pmap, x1, y1),
              (tt= civtile_get_terrain(ptil2)) == t0)) {
            tt = (myrand(3)<1 ? T_FOREST:(myrand(3)<2 ? T_GRASSLAND:T_HILLS));
            civtile_set_terrain(ptil2, tt);
          }
        }
      }
      break;
    case T_PLAINS:
    case T_GRASSLAND:
      if (count_terrainN(pmap, x, y, 3, T_PLAINS)
        + count_terrainN(pmap, x, y, 3, T_GRASSLAND) > 40) {
        for(ii= 6; ii-- > 0; ) {
          if ((myrand(3) < 2)
           && (nn= myrand(8), civmap_step(pmap, &x1, &y1, x, y, nn, is_border))
	   && (ptil2 = civmap_get_tile(pmap, x1, y1),
              (tt= civtile_get_terrain(ptil2)) == t0)) {
            tt = (myrand(3)<1 ? T_FOREST : T_HILLS);
            civtile_set_terrain(ptil2, tt);
          }
        }
      }
      break;
    }
    t0= civtile_get_terrain(ptile), tt= ix[t0];
    if( tt != t0 ) {
      /* eradicate %0 tiles to keep the purists happy */
      civtile_set_terrain(ptile, tt);
    }
  }
  free(yx);
}

/**************************************************************************
  Returns 1 if this is a 1x1 island (weighted by diagonal accessibility)
  TODO:  detect and remove peninsular bridges between large continents.
**************************************************************************/
static bool is_tiny_island(const CIVMAP *pmap, int x, int y)
{
  static int ix[] = { 1, 16, 2, 32, 4, 64, 8, 128 };
  int nn = 0;

  if (civmap_get_terrain(pmap, x, y) == T_OCEAN)
    return FALSE;

  /* bias against small archipelagos outside coastal waters */
  if (count_terrainN(pmap, x, y, 2, T_OCEAN)+myrand(4) > 24)
    return TRUE;

  cartesian_adjc_iterate(pmap, x, y, x1, y1) {
    if (civmap_get_terrain(pmap, x1, y1) != T_OCEAN)
      return FALSE;
  } cartesian_adjc_iterate_end

  cartesian_diagonal_dir_iterate(pmap, x, y, x1, y1, dir) {
    if (civmap_get_terrain(pmap, x1, y1) != T_OCEAN)
      nn+= ix[dir];
  } cartesian_diagonal_dir_iterate_end

  /* some patterns are more aesthetically pleasing than others */
  switch(nn) {
  case 1:
  case 2:
  case 4:
  case 8:
    return (myrand(100) < 66);
  case 5:
  case 10:
    return (myrand(100) < 75);
  case 3:
  case 6:
  case 9:
  case 12:
    return (myrand(100) < 50);
  case 7:
  case 11:
  case 13:
  case 14:
    return (myrand(100) < 90);
  case 15:
  case 0:
    return TRUE;
  }
  return FALSE;
}

/**************************************************************************
  Removes all 1x1 islands (sets them to ocean).
**************************************************************************/
static void remove_tiny_islands(const CIVMAP *pmap)
{
  CIVTILE *ptile;
  int  nn, k, x, y, xy, *yx;

  /* Be careful with polar continents */
  nn = (pmap->separatepoles ? 3 : 0);

  /* generate a random list of land tiles, possibly excluding the poles */
  nn*= pmap->xsize;
  xy = pmap->ysize*pmap->xsize - nn;
  yx = (int *)fc_malloc(sizeof(int) * (xy - nn));
  for( k = 0; --xy >= nn; ) {
    ptile = (civmap_get_tiles(pmap) + xy);
    if (civtile_get_terrain(ptile) != T_OCEAN)
      y = myrand(++k), yx[k-1] = yx[y], yx[y] = xy;
  }

  /* process tiles from this list for small islands */
  for (xy = k; --xy >= 0;) {
    index_to_civmap_pos(pmap, &x, &y, &x, &y, yx[xy]);
    if (is_tiny_island(pmap, x, y)) {
      ptile = (civmap_get_tiles(pmap) + yx[xy]);
      civtile_set_terrain(ptile, T_OCEAN);
    }
  }

  /* 1x1 final cleanup */
  for (xy = k; --xy >= 0;) {
    index_to_civmap_pos(pmap, &x, &y, &x, &y, yx[xy]);
    if (count_terrainN(pmap, x, y, 1, T_OCEAN) >= 8) {
      ptile = (civmap_get_tiles(pmap) + yx[xy]);
      civtile_set_terrain(ptile, T_OCEAN);
    }
  }
  free(yx);
}

/**************************************************************************
  Returns 1 if this is a 1x1 lake (weighted by cartesian accessibility)
**************************************************************************/
static bool is_tiny_lake(const CIVMAP *pmap, int x, int y)
{
  static int ix[] = { 16, 1, 32, 2, 64, 4, 128, 8 };
  int nn = 0;

  if (civmap_get_terrain(pmap, x,y) != T_OCEAN)
    return FALSE;

  /* Do not remove river outlets */
  cartesian_adjc_dir_iterate(pmap, x, y, x1, y1, dir) {
    if (civmap_has_specials(pmap, x1, y1, S_RIVER))
      return FALSE;
    nn+= (civmap_get_terrain(pmap, x1, y1) == T_OCEAN ? ix[dir] : 0);
  } cartesian_adjc_dir_iterate_end

  /* bias against small (potentially) landlocked seas */
  if (count_terrainN(pmap, x,y,2, T_OCEAN)+myrand(4) < 4)
     return TRUE;

  cartesian_diagonal_dir_iterate(pmap, x, y, x1, y1, dir) {
    nn+= (civmap_get_terrain(pmap, x1, y1) == T_OCEAN ? ix[dir] : 0);
  } cartesian_diagonal_dir_iterate_end

  /* some patterns are more aesthetically pleasing than others */
  switch(nn) {
  case 1:
  case 2:
  case 4:
  case 8:
    return (myrand(100) < 66 ? 1:0);
  case 3:
  case 5:
  case 6:
  case 9:
  case 10:
  case 12:
    return (myrand(100) < 50 ? 1:0);
  case 7:
  case 11:
  case 13:
  case 14:
    return (myrand(100) < 33 ? 1:0);
  case 15:
    return (myrand(100) < 90 ? 1:0);
  case 0:
  case 240:
    return TRUE;
  }
  return FALSE;
}

/**************************************************************************
  Removes all 1x1 lakes (sets them to grass).
**************************************************************************/
static void remove_tiny_lakes(const CIVMAP *pmap)
{
  CIVTILE *ptile;
  int  nn, k, x,y, xy, *yx;

  /* Be careful not to break a polar passage */
  if (!pmap->separatepoles)
    nn = 0;
  else
  if (civmap_get_type(pmap, (MAP_TYPE_ISO|MAP_TYPE_HEX)) != 0)
    nn = 5;
  else
    nn = 4;

  /* generate a random list of sea tiles, possibly excluding the poles */
  nn*= pmap->xsize;
  xy = pmap->ysize*pmap->xsize - nn;
  yx = (int *) fc_malloc(sizeof(int) * (xy - nn));
  for( k = 0; --xy >= nn; ) {
    ptile = (civmap_get_tiles(pmap) + xy);
    if (civtile_get_terrain(ptile) == T_OCEAN)
      y = myrand(++k), yx[k-1] = yx[y], yx[y] = xy;
  }

  /* process tiles from this list */
  while (--k >= 0) {
    index_to_civmap_pos(pmap, &x, &y, &x, &y, yx[k]);
    if (is_tiny_lake(pmap, x, y)) {
      ptile = (civmap_get_tiles(pmap) + yx[k]);
      civtile_set_terrain(ptile, T_GRASSLAND);
    }
  }
  free(yx);
}

/**************************************************************************
  make land simply does it all based on a generated heightmap
  1) with pmap->landpercent it generates a ocean/grassland map
  2) it then calls the above functions to generate the different terrains
**************************************************************************/
static void make_land(CIVMAP *pmap)
{
  int count, forever=50;
  int total= (pmap->xsize*pmap->ysize*pmap->landpercent)/100;

  /* Adjust sealevel to approximate landpercent */
  tocean= (maxval*pmap->landpercent)/100;
  do {
    count= 0;
    whole_civmap_xy_iterate(pmap, xy, xn, yn, x, y) {
      CIVTILE *ptile = (civmap_get_tiles(pmap) + xy);
      if (height_map[xy] < tocean) {
        civtile_set_terrain(ptile, T_OCEAN);
      } else {
        civtile_set_terrain(ptile, T_GRASSLAND);
        count++;
      }
    } whole_civmap_xy_iterate_end
    tocean= (((count > total) ? 11 : 9)*tocean)/10;
  } while (--forever > 0 && abs(total-count)*20 > total);

  /* these can change land/ocean counts */
  make_polar(pmap);
  if( pmap->separatepoles )
    make_passable(pmap);
  if( !pmap->tinyisles )
    remove_tiny_islands(pmap);
  make_swamps(pmap);
  make_rivers(pmap);
  remove_tiny_lakes(pmap);

  make_mountains(pmap, (maxval*66)/100);
  make_deserts(pmap);
  make_forests(pmap);
  make_plains(pmap);
  make_fair(pmap);

  civmap_continent_assign(pmap);
  setup_seadata(pmap);
  map_check(pmap);
}

/**************************************************************************
  this function spreads out huts on the map, a position can be used for a
  hut if there isn't another hut close and if it's not on the ocean or an
  ARCTIC border. ARCTIC tiles are in addition only 30% likely.
**************************************************************************/
static void make_huts(CIVMAP *pmap, int number)
{
  CIVTILE *ptile;
  enum terrain_type ttype;
  int x,y, xy, *yx;

  /* generate a list of tile indicies excluding the poles */
  xy = (pmap->ysize-2)*pmap->xsize;
  yx = (int *)fc_malloc(xy*sizeof(int));
  for( y = xy; --y >= 0; )
    yx[y]= y;

  /* randomly pick "number" tiles from this list */
  while (number > 0 && xy > 0) {
    /* Randomly select one of the remaining tiles */
    y = myrand(xy--), x = yx[y] + pmap->xsize, yx[y] = yx[xy];
    ptile = (civmap_get_tiles(pmap) + x);
    ttype = civtile_get_terrain(ptile);
    if (ttype != T_OCEAN
    && (ttype != T_ARCTIC || myrand(10) < 3)) {
      index_to_civmap_pos(pmap, &x, &y, &x, &y, x);
      if (!is_hut_close(pmap, x, y)) {
        number--;
        civtile_special_add(ptile, S_HUT);
        /* Don't add to islands[].goodies because islands[] not
         * setup at this point, except for generators 2,3,4, but they
         * have pre-set starters anyway.
         */
      }
    }
  }
  free(yx);
  pmap->has_huts = 1;
}

static void add_specials(CIVMAP *pmap, int prob)
{
  CIVTILE *ptile;
  enum terrain_type ttype;
  int x,y, xy, *yx;

  /* generate a list of tile indicies excluding the poles */
  xy = (pmap->ysize-2)*pmap->xsize;
  yx = (int *)fc_malloc(xy*sizeof(int));
  for( y = xy; --y >= 0; )
    yx[y]= y;

  /* randomly pick tiles from this list */
  while (xy > 0) {
    y= myrand(xy--), x= yx[y]+pmap->ysize, yx[y]= yx[xy];
    ptile = (civmap_get_tiles(pmap) + x);
    ttype = civtile_get_terrain(ptile);
    index_to_civmap_pos(pmap, &x, &y, &x, &y, x);
    if ( ttype != T_OCEAN || is_coastal(pmap, x,y) ) {
      if (myrand(1000) < (ttype == T_GRASSLAND ? prob*4/3 : prob) ) {
        if (ttype == T_GRASSLAND || !is_special_close(pmap, x,y)) {
          if (pmap->terrain->types[ttype].special[1].name[0]
           && (!(pmap->terrain->types[ttype].special[2].name[0])
                || myrand(100) < 50)) {
            civtile_special_add(ptile, S_SPECIAL_1);
          }
          else if (pmap->terrain->types[ttype].special[2].name[0]) {
            civtile_special_add(ptile, S_SPECIAL_2);
          }
        }
      }
    }
  }
  free(yx);
  pmap->has_specials = 1;
}

/**************************************************************************
 List this tile and recursively adjacent continent tiles with number "n0"
**************************************************************************/
static int continent_list1(const CIVMAP *pmap, int *list, int x, int y, int n0)
{
  CIVTILE *ptile;

  if( (ptile = civmap_get_tile(pmap, x, y))
   && (civtile_get_continent(ptile) == n0) ) {
    /* block this square */
    civtile_set_continent(ptile, -n0);
    adjc_iterate(pmap, x, y, x1, y1) {
      continent_list1(pmap, list, x1, y1, n0);
    } adjc_iterate_end;
    list[++list[0]]= civmap_pos_to_index(pmap, x, y);
    civtile_set_continent(ptile, n0);
  }
  return list[0];
}

/**************************************************************************
 Number this tile and recursive adjacent tiles with given number to
 specified continent number, by flood-fill algorithm, i.e. join.
 Returns the number of tiles successfully assigned this number.
**************************************************************************/
static int continent_flood1(const CIVMAP *pmap, int x, int y, int n0, int nr)
{
  CIVTILE *ptile;
  int  ret= 0;

  if( is_real_civmap_pos(pmap, x, y)
   && (ptile = civmap_get_tile(pmap, x, y))
   && (civtile_get_continent(ptile) == n0)
   && (civtile_get_terrain(ptile) != T_OCEAN) ) {
    ret++, civtile_set_continent(ptile, nr);
    adjc_iterate(pmap, x, y, x1, y1) {
      ret+= continent_flood1(pmap, x1, y1, n0, nr);
    } adjc_iterate_end;
  }
  return ret;
}

/**************************************************************************
 Number this tile and recursive adjacent tiles with unassigned number
 to specified continent number, by flood-fill algorithm.
 Returns the number of tiles successfully assigned this number.
**************************************************************************/
static int continent_flood(const CIVMAP *pmap, int x, int y, int nr)
{
  return continent_flood1(pmap, x, y, 0, nr);
}

/**************************************************************************
 Assign continent numbers to all tiles.
 Numbers 1 and 2 are reserved for polar continents if
 pmap->generator != 0; otherwise are not special.
 Also sets pmap->num_continents (note 0 is ocean, and continents
 have numbers 1 to pmap->num_continents _inclusive_).
 Note this is not used by generators 2,3 or 4 at map creation
 time, as these assign their own continent numbers.
**************************************************************************/
void civmap_continent_assign(CIVMAP *pmap)
{
  CIVTILE *ptile;
  int isle = 0;

  if( pmap->generator != 1 ) {
    freelog(LOG_ERROR, "Map continent numbers called for generator %d. "
      "Islands will be renumbered invalidating start positions.",
      pmap->generator);
  }

  whole_civmap_xy_iterate(pmap, xy, xn, yn, x, y) {
    ptile = (civmap_get_tiles(pmap) + xy);
    civtile_set_continent(ptile, 0);
  } whole_civmap_xy_iterate_end;

  if (pmap->generator > 0 && pmap->separatepoles) {
    int x, y;
    native_to_civmap_pos(pmap, &x, &y, 0, 0);
    continent_flood(pmap, x, y, ++isle);
    if( !civmap_get_type(pmap, MAP_TYPE_WRAPY) ) {
      native_to_civmap_pos(pmap, &x, &y, pmap->xsize-1, pmap->ysize-1);
      continent_flood(pmap, x, y, ++isle);
    }
  }

  whole_civmap_xy_iterate(pmap, xy, xn, yn, x, y) {
    ptile = (civmap_get_tiles(pmap) + xy);
    if (civtile_get_continent(ptile) == 0
     && civtile_get_terrain(ptile) != T_OCEAN) {
      continent_flood(pmap, x, y, ++isle);
    }
  } whole_civmap_xy_iterate_end;

  pmap->num_continents = isle;
  freelog(LOG_VERBOSE, "Map has %d continents", pmap->num_continents);
}

/**************************************************************************
 List this tile and recursively adjacent water tiles with number "n0".
 This function flips the sign of the ocean number as each tile is counted.
 The "list" array has its current count at list[0] and size at list[-1].
**************************************************************************/
static int ocean_list1(const CIVMAP *pmap, int *list, int x, int y, int n0)
{
  CIVTILE *ptile;
  int xy;

  assert(list[0] < list[-1] );

  xy = civmap_pos_to_index(pmap, x, y);
  ptile = (civmap_get_tiles(pmap) + xy);
  if( n0 != 0 && civtile_get_ocean(ptile) == n0 && list[0] < list[-1] ) {
    /* block this square */
    civtile_set_ocean(ptile, -n0);
    list[++list[0]]= xy;
    adjc_iterate(pmap, x, y, x1, y1) {
      ocean_list1(pmap, list, x1, y1, n0);
    } adjc_iterate_end;
  }
  return list[0];
}

/**************************************************************************
 List the tile and recursively adjacent water tiles with its ocean number.
 It returns the number of tiles added to the list.
 The "list" array has its current count at list[0] and size at list[-1].
**************************************************************************/
static int ocean_list(const CIVMAP *pmap, int *list, int x, int y)
{
  CIVTILE *ptile;
  int  xy, m0, n0, nn;

  assert(list[0] < list[-1] );

  xy = civmap_pos_to_index(pmap, x, y);
  ptile = (civmap_get_tiles(pmap) + xy);
  n0= list[0];
  m0= civtile_get_ocean(ptile);
  ocean_list1(pmap, list, x, y, m0);

  /* Undo blocking marks ... */
  for(nn= list[0]; nn > n0; nn--) {
    ptile = (civmap_get_tiles(pmap) + list[nn]);
    civtile_set_ocean(ptile, m0);
  }
  return list[0]-n0;
}

/**************************************************************************
 Number this tile and recursively adjacent water tiles with number "n0"
   to ocean number "nr", by flood-fill algorithm (i.e. join watersheds).
 Returns the number of tiles successfully assigned number "nr".
 If n0 is negative, sets all positive tiles to the lowest number found.
 In this case the return totals are meaningless due to double counting.
**************************************************************************/
static int ocean_flood1(const CIVMAP *pmap, int x, int y, int n0, int nr)
{
  CIVTILE *ptile;
  int  tx, mm, ret= 0;

  assert(n0 != nr);

  /* sanity check for valid map, water tile */
  if( is_real_civmap_pos(pmap, x, y)
   && (ptile = civmap_get_tile(pmap, x, y))
   && ( ((tx= (civtile_get_terrain(ptile) & 15)) == T_OCEAN)
     || civtile_has_specials(ptile, S_RIVER) )) {
    /* 
     * loop: "if" in the normal case, repeats if changed for n0 == -1.
     *       For n0 == -1, also reset nr if current ocean number is lower.
     * TODO: should we clear the current tile to always force one cycle?
     */
    while( (mm= civtile_get_ocean(ptile), n0) >= 0
        ? (mm == n0) : (mm != nr && (mm > nr || mm <= 0 || (nr= mm))) ) {
      /* update this square */
      ret++, civtile_set_ocean(ptile, nr);
      /* update river accessible adjacent */
      cartesian_adjc_iterate(pmap, x, y, x1, y1) {
        if( 0 < (mm = ocean_flood1(pmap, x1, y1, n0, nr)) ) {
          ret+= mm;
          nr = civmap_get_ocean(pmap, x1, y1);
        }
      } cartesian_adjc_iterate_end;
      if(tx == T_OCEAN) {
        /* plus update ocean accessible diagonals */
        cartesian_diagonal_iterate(pmap, x, y, x1, y1) {
          if( civmap_get_terrain(pmap, x1,y1) == T_OCEAN
           && 0 < (mm = ocean_flood1(pmap, x1, y1, n0, nr)) ) {
            ret+= mm;
            nr = civmap_get_ocean(pmap, x1, y1);
          }
        } cartesian_diagonal_iterate_end;
      }
    }
  }
  return ret;
}

/**************************************************************************
 Number this tile and recursive adjacent water tiles with unasssigned
   number to specified ocean number, by flood-fill algorithm.
 Returns the number of tiles successfully assigned this number.
**************************************************************************/
static int ocean_flood(const CIVMAP *pmap, int x, int y, int nr)
{
  return ocean_flood1(pmap, x, y, 0, nr);
}

/**************************************************************************
 Assign ocean numbers to all tiles.
 Also sets civmap.num_oceans (oceans have numbers 1 to
 civmap.num_oceans _inclusive_).
**************************************************************************/
void civmap_ocean_assign(CIVMAP *pmap)
{
  int seas = 0;

  whole_civmap_xy_iterate(pmap, xy, xn, yn, x, y) {
    CIVTILE *ptile = (civmap_get_tiles(pmap) + xy);
    civtile_set_ocean(ptile, 0);
  } whole_civmap_xy_iterate_end;

  whole_civmap_xy_iterate(pmap, xy, xn, yn, x, y) {
    CIVTILE *ptile = (civmap_get_tiles(pmap) + xy);
    if (!civtile_get_ocean(ptile)
     && civtile_get_terrain(ptile) == T_OCEAN) {
      ocean_flood(pmap, x, y, ++seas);
    }
  } whole_civmap_xy_iterate_end;

  pmap->num_oceans = seas;
  freelog(LOG_VERBOSE, "Map has %d oceans", pmap->num_oceans);
}

/**************************************************************************
 Allocate oceans array and fill in values.
**************************************************************************/
static void setup_seadata(CIVMAP *pmap)
{
  int ii;

  civmap_ocean_assign(pmap);
  assert(pmap->num_oceans > 0);

  /* allocate + 1 so can use ocean number as index */
  if( oceans != NULL )
    free(oceans);
  oceans= fc_malloc((pmap->num_oceans+1)*sizeof(struct seadata));

  /* initialize: */
  for(ii= 0; ii <= pmap->num_oceans; ii++) {
    oceans[ii].x= oceans[ii].y= -1;             /* flag */
    oceans[ii].count= oceans[ii].rivers= 0;
  }

  /* get x and y positions: (top left) */
  whole_civmap_xy_iterate(pmap, xy, xn, yn, x, y) {
    CIVTILE *ptile = (civmap_get_tiles(pmap) + xy);
    int tx= civtile_get_terrain(ptile);
    ii= civtile_get_ocean(ptile);
    oceans[ii].count++;
    if (civtile_has_specials(ptile, S_RIVER))
      oceans[ii].rivers++;
    else
    if (oceans[ii].x == -1 && tx == T_OCEAN)
      oceans[ii].x= x, oceans[ii].y= y;
  } whole_civmap_xy_iterate_end;

  freelog(LOG_VERBOSE, _("Map has %i seas of size: "), pmap->num_oceans);
  for(ii= 1; ii <= pmap->num_oceans; ii++) {
    freelog(LOG_VERBOSE, _("%2d %2d,%-2d = %d/%d"),
      ii, oceans[ii].x, oceans[ii].y, oceans[ii].count, oceans[ii].rivers);
  }
}

/**************************************************************************
 Allocate islands array and fill in values.
 Note this is only use for pmap->generator<=1, since others
 setups islands and starters explicitly.
**************************************************************************/
static void setup_isledata(CIVMAP *pmap)
{
  int oldisles, goodisles, firstcont;
  int count, mincount, starters;
  int good, mingood, maxgood, riches;
  int nn, guard1 = 0;

  assert(pmap->num_continents>0);

  /* allocate + 1 so can use continent number as index */
  if( islands )
    free(islands);
  islands = fc_malloc((pmap->num_continents+1)*sizeof(struct isledata));

  /* initialize: */
  for(nn = 0; nn <= pmap->num_continents; nn++) {
    islands[nn].x = islands[nn].y = -1;             /* flag */
    islands[nn].count = islands[nn].goodies = islands[nn].starters = 0;
  }

  /* Add up the goodies: for useable ocean, add value to continent
     for _every_ adjacent land tile.  This is IMO not very good,
     because it adds potentially many times, and usable sea of
     distance 2 from land is ignored, but this re-produces the
     results of the previous method which used flood_fill().  --dwp
     Changed to *not* do multiple adds - bugwards compatibility be damned.
     This is also the correct place to add S_HUT bonus.
  */
  whole_civmap_xy_iterate(pmap, xy, xn, yn, x, y) {
    CIVTILE *ptile = (civmap_get_tiles(pmap) + xy);
    int cont = civtile_get_continent(ptile);
    int goodval = civtile_goodness(ptile);
    /* get x and y positions: (top left) */
    if (islands[cont].x == -1) {
      islands[cont].x = x;
      islands[cont].y = y;
    }
    /* add up goodies */
    if (cont) {
      islands[cont].count++;
      islands[cont].goodies += goodval;
      if (civtile_has_specials(ptile, S_HUT)) {
        islands[cont].goodies += 0; /* 3; */   /* regression testing */
      }
    } else {
      assert(civtile_get_terrain(ptile) == T_OCEAN);
      /* no need to check is_sea_usable(x,y), because will
         only use for adjacent land (cont1>0) below */
      {
        int dirs[8] = { 0 };
        adjc_dir_iterate(pmap, x, y, x1, y1, dir) {
          int cont1 = civmap_get_continent(pmap, x1, y1);
          int gval = goodval;
          /* Only count once per continent - trick first test == ocean */
          for( nn = dir+1; --nn >= 0; ) {
            if( dirs[nn] == cont1 ) {
              gval = 0;
              break;
            }
          }
          if (gval > 0) {
            islands[cont1].goodies += goodval;
            dirs[dir] = cont1;
          }
        } adjc_iterate_end;
      }
    }
  } whole_civmap_xy_iterate_end;

  /* the arctic and the antarctic are continents 1 and 2 for generator>0*/
  firstcont = 1;
  if (pmap->generator > 0 && pmap->separatepoles) {
    firstcont += (!civmap_get_type(pmap, MAP_TYPE_WRAPY) ? 2 : 1);
  }

  count=0;
  riches=0;
  for (nn = firstcont; nn <= pmap->num_continents; nn++) {
      riches += islands[nn].goodies;
      count  += islands[nn].count;
  }

  /* scaling factor based on number of players for "big" continents */

  /* estimate the minmum size continent for pmap->num_start_positions
     then adjust to make sure there is at least this number */
  count /= (pmap->num_start_positions + pmap->num_continents - firstcont + 1);
  if( count < 30 )
    count = 30;
  do {
    goodisles = 0;
    for (nn = firstcont; nn <= pmap->num_continents; nn++) {
      goodisles += islands[nn].count/count;
    }
  } while( goodisles < pmap->num_start_positions && (count= (count*8)/10) > 30);

  /* Now compute average number of players on each continent for real */
  mincount = pmap->num_start_positions;
  goodisles= 0;
  for (nn = firstcont; nn <= pmap->num_continents; nn++) {
      /* To remove the size cutoff add 1 to islands[nn].count.
      islands[nn].count = 1 + islands[nn].count/count;
      */
      if( pmap->generator == 2 || pmap->generator == 3 )
        islands[nn].count = 1;
      else
        islands[nn].count = islands[nn].count/count;
      goodisles += islands[nn].count;
      if( 0 < mincount && mincount < islands[nn].count )
        mincount = islands[nn].count;
  }
  if( goodisles < pmap->num_start_positions ) {
    map_printtmap(pmap);
    map_printcmap(pmap);
    map_printstats(pmap, 0);
    freelog(LOG_ERROR, "goodisles=%i players=%d",
      goodisles, pmap->num_start_positions);
    assert( goodisles >= pmap->num_start_positions );
  }

  starters= pmap->num_start_positions+mincount+1, oldisles = goodisles+1;
  while (starters > pmap->num_start_positions+mincount && oldisles > goodisles){
    freelog(LOG_VERBOSE, "goodisles=%i", goodisles);
    oldisles= goodisles;
    goodisles= 0;
    mincount= pmap->num_start_positions;
    maxgood= 1;
    mingood= riches;
    good= 0;
    starters= 0;

    /* compute goody stats */
    for (nn = firstcont; nn <= pmap->num_continents; nn++) {
      islands[nn].starters= 0;
      count = (islands[nn].count > 0)
        ? islands[nn].goodies/islands[nn].count : 0;
      if( (count > (riches+oldisles-1)/oldisles)
       || (oldisles > pmap->num_start_positions
          && (count*4 > 3*(riches+oldisles-1)/oldisles)) ) {
        goodisles += islands[nn].count;
        if( RANGE_CHECK_0(mincount, islands[nn].count) )
          mincount = islands[nn].count;
        good += islands[nn].goodies;
        if(mingood > count)
          mingood= count;
        if(maxgood < count)
          maxgood= count;
      }
    }

    if(mingood < maxgood/pmap->num_start_positions-1)
      mingood= maxgood/pmap->num_start_positions-1;

    /* collect starters */
    for (nn = firstcont; nn <= pmap->num_continents; nn++) {
      count = (islands[nn].count > 0)
        ? islands[nn].goodies/islands[nn].count : 0;
      if( (count > (riches+oldisles-1)/oldisles)
       || (oldisles > pmap->num_start_positions
          && (count*4 > 3*(riches+oldisles-1)/oldisles)) ) {
        freelog(LOG_VERBOSE, "islands[nn].goodies=%i", islands[nn].goodies);
        if (pmap->generator == 2
         || pmap->generator == 3
         || (islands[nn].starters= (islands[nn].goodies+guard1)/mingood) <= 0) {
          islands[nn].starters= 1;
        }
        starters+= islands[nn].starters;
      }
    }

    if(starters < pmap->num_start_positions) {
      starters= pmap->num_start_positions+mincount+1;
      goodisles= oldisles;
      oldisles= goodisles+1;
      riches= good+1;
      if(mingood/pmap->num_start_positions > 5) {
        guard1+= mingood/pmap->num_start_positions;
      } else {
        guard1+=5;
      }
      freelog(LOG_DEBUG,
              _("Map generator: not enough start positions, fixing."));
    }
  }
  freelog(LOG_DEBUG, "The map has %i starting positions on %i isles.",
          starters, goodisles);
}

/**************************************************************************
  where do the different races start on the map? well this function tries
  to spread them out on the different islands.
  TODO:  change asserts to logged error return and cleaner caller recovery
**************************************************************************/
void create_start_positions(CIVMAP *pmap)
{
  CIVTILE *ptile;
  int tries, nr= 0;
  int k, sum, dist;
  int tx, x,y, xy,*yx;

  if (islands == NULL)              /* already setup for generators 2,3,4 */
    setup_isledata(pmap);

  sum=0;
  for (k= 0; k <= pmap->num_continents; k++) {
    islands[k].count = islands[k].starters;
    sum += islands[k].starters;
    if (islands[k].starters != 0) {
      freelog(LOG_VERBOSE, "%d starters on isle %i", islands[k].starters, k);
    }
  }
  freelog(LOG_NORMAL, "Map summary: %d starters on %d isles for %d players",
    sum, pmap->num_continents, pmap->num_start_positions);
  assert(pmap->num_start_positions <= nr+sum);

  /* Collect a randomized list of starting tiles */
  xy= pmap->ysize*pmap->xsize;
  yx= (int *)fc_malloc(xy*sizeof(int));
  k= 0;
  whole_civmap_xy_iterate(pmap, xy, xn, yn, x, y) {
    ptile = (civmap_get_tiles(pmap) + xy);
    if (islands[(int)civtile_get_continent(ptile)].starters) {
      tx= civtile_get_terrain(ptile);
      if( tx == T_GRASSLAND
       || tx == T_PLAINS
       || tx == T_RIVER
       || tx == T_HILLS ) {
        int nf, food, nofood= 0;
        food = (civtile_get_food_base(ptile) <= 1);
        square_iterate(pmap, x, y, 2, x1, y1) {
          nf= civmap_get_food_base(pmap, x1, y1);
          if( nf > 3 ) food+= 2;
          else
          if( nf > 1 ) food++;
          else
          if( nf < 1 ) nofood++;
        } square_iterate_end;
        /* at least a couple food rich for early development,
         * and not too many duds for long term potential. */
        if( food > 2 && food+2 >= nofood ) {
          yx[k++]= xy, islands[(int)civtile_get_continent(ptile)].count--;
        } else {
          freelog(LOG_VERBOSE,
            "dud starter %d:%2d.%-2d food=%d nofood=%d",
            civtile_get_continent(ptile), x, y, food, nofood);
        }
      }
    }
  } whole_civmap_xy_iterate_end;
  assert(pmap->num_start_positions <= nr+k);
  for (tx= 0; tx <= pmap->num_continents; tx++) {
    if (islands[tx].count > 0) {
      map_printtmap(pmap);
      map_printcmap(pmap);
      map_printstats(pmap, tx);
      freelog(LOG_ERROR,
        "%d starters unallocated on isle %i", islands[tx].count, tx);
    }
  }

  tries = 5;
  do {
    dist= 40;    /* sanity */
    if(dist >= pmap->xsize/3)
      dist= pmap->xsize/3;
    if(dist >= pmap->ysize/3)
      dist= pmap->ysize/3;
    if(dist < 5)
      dist= 5;   /* sanity */

    /* Pick off tiles separated by at least dist */
    for (nr= 0; nr < pmap->num_start_positions && dist >= 4; dist=(dist*8)/10) {
      for (xy= k; --xy >= 0;) {
        y= myrand(xy+1), tx= yx[y], yx[y]= yx[xy], yx[xy]= tx;
        ptile = (civmap_get_tiles(pmap) + tx);
        index_to_civmap_pos(pmap, &x, &y, &x, &y, tx);
        tx= civtile_get_continent(ptile);
        if (islands[tx].starters > 0) {
          if (!is_starter_bad(pmap, x, y, nr, dist)) {
          freelog(LOG_VERBOSE, "starter for isle %i, %2d.%-2d", tx, x, y);
            islands[tx].starters--;
            pmap->start_positions[nr].x= x;
            pmap->start_positions[nr].y= y;
            if( ++nr >= pmap->num_start_positions )
              break;
          } else {
            freelog(LOG_VERBOSE,
              "starter not chosen at %2d.%-2d %d chances on isle %i",
              x, y, islands[tx].starters, tx);
          }
        }
      }
    }
    freelog(LOG_NORMAL,
      "Map summary: %d starters chosen, %d/%d (allowed/found) for %d players",
      nr, sum, k, pmap->num_start_positions);

    /* Reset starter counts */
    for(xy = nr; --xy >= 0;) {
      x= pmap->start_positions[xy].x;
      y= pmap->start_positions[xy].y;
      islands[(int)civmap_get_continent(pmap, x, y)].starters++;
    }
  } while( pmap->num_start_positions > nr && tries-- > 0 );
  assert(pmap->num_start_positions <= nr);
  pmap->num_start_positions = nr;

  free(yx);
  free(islands);
  islands = NULL;
}

/**************************************************************************
 readjust terrain counts so that it makes sense for mapgen 1, 2, 3 and 4
 idea: input is the target percent of landmass for each terrain.
       mapgen 1, mapgen 2 and 3 and 4 all use percents, currently.
 GRASSLAND must always be non-zero, as it is the current constraint type.
 TODO: add map types or controls to vary ratios of
       (age)         mountains/hills/desert, height smoothing algorithm
       (wetness)     desert/plains/grass/swamps, mts/hills/forest/jungle
       (temperature) desert/arctic, plains/tundra or jungles/forest,

 This function needs to be called from the server everytime a
 a parameter changes.
 It will be called again at game start, too.
**************************************************************************/
void adjust_terrain_param(CIVMAP *pmap)
{
  int total;

  /* TODO: convert to independent percentages */
  pmap->MOUNTAINS += pmap->HILLS,   pmap->HILLS   = 0;
  pmap->ARCTIC    += pmap->TUNDRA,  pmap->TUNDRA  = 0;
  pmap->FORESTS   += pmap->JUNGLES, pmap->JUNGLES = 0;
  pmap->SWAMPS    += pmap->OCEANS,  pmap->OCEANS  = 0;

  /*
   * ignore riverlength
   * make_rivers replaces/overrides an amount of existing terrain
   */
  total = /* pmap->RIVERS + */
      pmap->MOUNTAINS + pmap->HILLS
    + pmap->ARCTIC    + pmap->TUNDRA 
    + pmap->FORESTS   + pmap->JUNGLES 
    + pmap->DESERTS
    + pmap->SWAMPS    + pmap->OCEANS
    + pmap->PLAINS    + pmap->GRASSLAND;

  /* recalibrate if things don't add up */
  if(total < 90 || total > 100) {
    /* pmap->RIVERS= pmap->RIVERS*100/total; */
    pmap->MOUNTAINS= pmap->MOUNTAINS*100/total;
    pmap->HILLS=     pmap->HILLS*100/total;
    pmap->ARCTIC=    pmap->ARCTIC*100/total;
    pmap->TUNDRA=    pmap->TUNDRA*100/total;
    pmap->FORESTS=   pmap->FORESTS*100/total;
    pmap->JUNGLES=   pmap->JUNGLES*100/total;
    pmap->DESERTS=   pmap->DESERTS*100/total;
    pmap->SWAMPS=    pmap->SWAMPS*100/total;
    pmap->OCEANS=    pmap->OCEANS*100/total;
    pmap->PLAINS=    pmap->PLAINS*100/total;
  }

  /* residual goes implicitly to grass, so always compute it */
  total = /* pmap->RIVERS + */ 
      pmap->MOUNTAINS + pmap->HILLS
    + pmap->ARCTIC    + pmap->TUNDRA 
    + pmap->FORESTS   + pmap->JUNGLES 
    + pmap->DESERTS
    + pmap->SWAMPS    + pmap->OCEANS
    + pmap->PLAINS    + 0;
  pmap->GRASSLAND= 100 - total;
}

/**************************************************************************
  This scales heights from mapgenerators > 1 so we can call make_<feature>.
**************************************************************************/
static void adjust_map2(const CIVMAP *pmap)
{
  int nn, tt;

  /* Fixup create_island heights */
  maxval= 0;
  whole_civmap_xy_iterate(pmap, xy, xn, yn, x, y) {
    nn= height_map[xy];
    if( nn < 0 ) nn= myrand(100)+3000;    /* polar/blocking paranoia */
    else
    if( nn < 100 ) nn= myrand(2000)+1000; /* oceans */
    else {                                /* island landmass */
      nn+= 3000;
      /* terrain assignment does not respect heighs - fake/fix this */
      if((tt= civmap_get_terrain(pmap, x, y)) == T_HILLS)
        nn+= myrand(3000)+1000;
      else
      if(tt == T_MOUNTAINS)
        nn+= myrand(6000)+2000;
      if (nn > maxval) maxval=nn;
    }
    height_map[xy]= nn;

    /* cursory smoothing to stabilize height diffs a little */
    if((nn= xy+pmap->xsize+1) < pmap->ysize*pmap->xsize
     && height_map[nn] >= 3000) {
      index_to_civmap_pos(pmap, &x, &y, &xn, &yn, nn);
      height_map[nn]= (hmapN(pmap, x,y,1)+height_map[nn])/2;
    }
  } whole_civmap_xy_iterate_end

  /* scale 0:tocean(== 3000) and tocean:maxval(== 7000) */
  /* Note this tends to smooth ocean floors and accentuate land features */
  whole_civmap_xy_iterate(pmap, xy, xn, yn, x,y) {
    if( (nn= height_map[xy]) >= tocean)
      height_map[xy]= ((nn-tocean)*7000)/(maxval-tocean)+3000;
  } whole_civmap_xy_iterate_end

  tocean= 3000;
  maxval= 10000;
}

/**************************************************************************
  since the generated map always has a positive number as minimum height
  reduce the height so the lowest height is zero and the maximum is 10000,
  also, set ~sealevel to 3000, this makes calculations easier
**************************************************************************/
static void adjust_map(const CIVMAP *pmap)
{
  int nn, minval;

  tocean= maxval= 0;
  minval= 32767;
  whole_civmap_xy_iterate(pmap, xy, nx, ny, x, y) {
    tocean+= nn= height_map[xy];
    if (nn > maxval) maxval= nn;
    else
    if (nn < minval) minval= nn;
  } whole_civmap_xy_iterate_end;

  /* maxval, average shifted to minval == 0 */
  tocean/= pmap->ysize*pmap->xsize;
  tocean= tocean-minval+1;
  maxval= maxval-minval+1;
  if( maxval <= tocean) maxval= tocean+1; /* paranoia */

  /* scale 0:tocean(== 3000) and tocean:maxval(== 7000) */
  /* Note this tends to smooth ocean floors and accentuate land features */
  whole_civmap_xy_iterate(pmap, xy, nx, ny, x, y) {
    nn= height_map[xy]-minval;
    if (nn > tocean)
      height_map[xy]= ((nn-tocean)*7000)/(maxval-tocean)+3000;
    else
      height_map[xy]= (nn*3000)/tocean;
  } whole_civmap_xy_iterate_end;

  tocean= 3000;
  maxval= 10000;
}

/**************************************************************************
  smooth_map should be viewed as a corrosion function on the map, it levels
  out the differences in the heightpmap->
**************************************************************************/
static void smooth_map(const CIVMAP *pmap)
{
  int a0, a1, a2, n0, n1, n2, nn;
  int x1, y1, yxsize, *yx;

  /* save results in a scratch array to avoid polluting the source */
  yxsize = pmap->ysize*pmap->xsize;
  yx = (int *)fc_malloc(yxsize*sizeof(int));
  memset(yx, 0, yxsize*sizeof(int));

  nn= 0;
  whole_civmap_xy_iterate(pmap, xy, xn, yn, x, y) {
    /* city sized and adjacent averages */
    n0= hmapN(pmap, x, y, 2);
    n0+= a1= hmapN(pmap, x, y, 1);
    n0+= (a1 > 1500 ? 600 : 0);       /* continental shelf drop off */

    /* detect key directions of elevation changes
     * Note: the height multipliers provide a key control over continent
     *       fragmentation and border smoothness.
    */
    a2= hslope(pmap, &n2, x, y, 0) - a1;
    a1= hslope(pmap, &n1, x, y, 1) - a1;
    a0= hmap(pmap, x, y);
    if( a0 < 4000 )
      a1= (a0 < 2000 ? 4 : 2) * a1;
    else
      a2= (a0 > 6000 ? 4 : 2) * a2;

    /* now propagate ridgelines and trenches. */
    if( a1 + a2 > 0 )
      n0 += (a1 > 1200 ? myrand(300) + myrand(a1) : 0);
    else
      n0 -= (a2 < -600 ? myrand(300) + myrand(-a2) : 0);

    /* build depressions == land lakes and valleys */
    a1= a0;
    if( !civmap_is_border_pos(pmap, x, y)
     && a0 > 3000
     && a0 < (civmap_step(pmap, &x1,&y1, x,y, DIR_CCW(n2),0), hmap(pmap, x1,y1))
     && a0 < (civmap_step(pmap, &x1,&y1, x,y, DIR_CW(n2), 0), hmap(pmap, x1,y1))
     && a0 < (civmap_step(pmap, &x1,&y1, x,y, n2, 0), hmap(pmap, x1,y1)) )
      a1 -= 1200, yx[civmap_pos_to_index(pmap, x1, y1)]-= 300;
    n0+= a1;
    n0/= 3;

    /* Sanity checks - 3x overkill */
    if ((unsigned)(n0+2000) > 32000) {
      freelog(LOG_VERBOSE, "mapgen.c: bad adjustment at %d,%d: %d/%08x",
        x, y, n0, n0);
	n0 = (n0 > -2000 ? (n0 > 30000 ? 30000 : n0) : 0);
    }
    if ((unsigned)a0 > 30000) {
      freelog(LOG_VERBOSE, "mapgen.c: bad height_map at %d,%d: %d/%08x",
        x, y, n0, n0);
	a0 = (a0 > 0 ? (a0 > 30000 ? 30000 : a0) : 0);
    }

    /* track the global adjustments */
    nn-= a0;
    nn+= yx[xy]+= n0;
  } whole_civmap_xy_iterate_end;
  nn/= yxsize;

  /* update source height_map, adjust for constant mass (i.e. settling) */
  whole_civmap_xy_iterate(pmap, xy, nx, ny, x, y) {
    height_map[xy]= (yx[xy] > nn) ? (yx[xy]-nn) : 0;
  } whole_civmap_xy_iterate_end;

  free(yx);
}

/**************************************************************************
  gen1_hmap - build a heightmap using the generator1 algorithm
**************************************************************************/
static void gen1_hmap(const CIVMAP *pmap)
{
  int x, y,  x0, y0,  x1, y1,  xy,  *yx,  s0, s1, s2,  hh, nn;
  bool is_border;

  /* start with base level fill plus polar basins */
  yx= fc_malloc(pmap->xsize*pmap->ysize*sizeof(int));
  for (y= 0;y < (pmap->ysize+1)/2; y++) {
    y1 = pmap->ysize-1 - y;
    nn = y * 5000 / pmap->ysize;
    nn = nn < 500 ? nn+1 : 500;
    for (x= 0;x < pmap->xsize; x++) {
      height_map[native_pos_to_index(pmap, x, y)] = myrand(nn)+nn;
      height_map[native_pos_to_index(pmap, x,y1)] = myrand(nn)+nn;
    }
  }

  /* sort map coordinates by decreasing height */
  /* first scramble the map coordinates to avoid subtle NS bias */
  xy= pmap->ysize*pmap->xsize;
  for( yx[0]= 0, y=1; y < xy; y++)
    x= myrand(y+1), yx[y]= yx[x], yx[x]= y;
  for(y= xy; --y >= 0;) {
    y1= yx[y], hh= height_map[y1];
    for(x= y; ++x < xy && height_map[yx[x]] > hh;)
      yx[x-1]= yx[x];
    yx[x-1]= y1;
  }

  /* pick three "seas" to avoid
   * actually pick land as new Atlantis to recycle super continents
   */
  s1= s2= 0, s0= yx[(xy*8)/10 - myrand(xy)/3];
  while( s2 == 0 ) {
    for(x= 0, y1= (pmap->ysize+pmap->xsize)/2; (y1*=2, y1/=3) > 9;) {
      x= yx[(myrand(xy)*pmap->landpercent)/100];
      if( civmap_dist(pmap, yx[x],s0) < y1 ) {
        continue;
      } else if( s1 == 0 ) {
        s1 = x;
      } else if( civmap_dist(pmap, yx[x],s1) < y1 ) {
        continue;
      } else if( s2 == 0 ) {
        s2 = x;
        break;
      }
    }
  }

  /* asteroid rain on landpercent of the map */
  for (nn= (xy*pmap->landpercent)/100; --nn >= 0;) {

    /*  biasing strikes to higher elevations */
    /*  and away from designated major seas  */
    for(x= 0, y1= (pmap->ysize+pmap->xsize)/2; (y1/=2) > 0;) {
      x= myrand(xy), y= myrand(xy), x= (x < y) ? x : y;
      if( civmap_dist(pmap, yx[x],s0) >= y1
       && civmap_dist(pmap, yx[x],s1) >= y1
       && civmap_dist(pmap, yx[x],s2) >= y1 )
      break;
    }
    index_to_civmap_pos(pmap, &x0, &y0, &x0, &y0, yx[x]);
    is_border = civmap_is_border_pos(pmap, x0, y0);

    /* point hit with directional splash (i.e. ridge) - 20% ejectile */
    hh= myrand(3000)+2000;
    if( myrand(100) < 20 )
      hh= -(hh > height_map[yx[x]] ? height_map[yx[x]] : hh)/2;
    y= myrand(8);  /* direction */

    /* hit in direction y */
    if( civmap_step(pmap, &x1, &y1, x0, y0, y, is_border)
     && (y1 = civmap_pos_to_index(pmap, x1, y1)) >= pmap->xsize
     && (xy-1 - y1) >= pmap->xsize)
      height_map[y1]+= (-hh > height_map[y1] ? -height_map[y1]/2 : hh);

    /* half hit in neighbouring directions */
    hh/= 2;
    if( civmap_step(pmap, &x1, &y1, x0, y0, DIR_CCW(y), is_border)
     && (y1 = civmap_pos_to_index(pmap, x1, y1)) >= pmap->xsize
     && (xy-1 - y1) >= pmap->xsize)
      height_map[y1]+= (-hh > height_map[y1] ? -height_map[y1]/2 : hh);

    if( civmap_step(pmap, &x1, &y1, x0, y0, DIR_CW(y), is_border)
     && (y1 = civmap_pos_to_index(pmap, x1, y1)) >= pmap->xsize
     && (xy-1 - y1) >= pmap->xsize)
      height_map[y1]+= (-hh > height_map[y1] ? -height_map[y1]/2 : hh);

    /* quarter hit at impact */
    hh/= 2;
    height_map[yx[x]]+= hh;

    /* periodically smooth, recalibrate and resort */
    if ((nn%pmap->xsize) == 0) {
      /* flip the highest and lowest 1% heights to stir things up */
      for(y= xy/100; y-- > 0;) {
        hh= height_map[yx[y]],
          height_map[yx[y]]= height_map[yx[xy-1 - y]],
          height_map[yx[xy-1 - y]]= hh;
      }
      smooth_map(pmap);
      adjust_map(pmap);
      for(y= xy; y-- > 0;) {
        y1= yx[y], hh= height_map[y1];
        for(x= y; ++x < xy && height_map[yx[x]] > hh;)
          yx[x-1]= yx[x];
        yx[x-1]= y1;
      }
      /* drop one sea and choose a new altantis */
      s0= s1, s1= s2, s2= 0;
      while( s2 == 0 ) {
        for(x= 0, y1= (pmap->ysize+pmap->xsize)/2; (y1*=2, y1/=3) > 9;) {
          x= yx[(myrand(xy)*pmap->landpercent)/100];
          if( civmap_dist(pmap, yx[x],s0) >= y1
           && civmap_dist(pmap, yx[x],s1) >= y1 ) {
            s2 = x;
            break;
          }
        }
      }
    }
  }
  smooth_map(pmap);
  adjust_map(pmap);
  free(yx);
}

/**************************************************************************
  mapgenerator1, highlevel function, that calls all the previous functions
**************************************************************************/
static void mapgenerator1(CIVMAP *pmap)
{
  /* allocate working storage */
  height_map= fc_malloc(pmap->xsize*pmap->ysize*sizeof(int));

  /* input (and if necessary convert) map generation parameters */
  adjust_terrain_param(pmap);

  /* generate height_map data */
  gen1_hmap(pmap);

  /* populate the generated height_map */
  make_land(pmap);

  free(height_map);
  height_map= NULL;
}

/**************************************************************************
  globals for generator 2, 3 & 4
**************************************************************************/
static int  *create_map;
static int  isleindex, n, e, s, w;
static long int totalmass; /* better a global than a duplicate formula */
static long int checkmass;
static long int totalweight;

/**************************************************************************
  fill an island with up four types of terrains, rivers done separately
**************************************************************************/
static void fill_island(const CIVMAP *pmap, long int *bucket, int coast,
        int warm0_weight, int warm1_weight, int cold0_weight, int cold1_weight,
        enum terrain_type warm0, enum terrain_type warm1,
        enum terrain_type cold0, enum terrain_type cold1)
{
  CIVTILE *ptile;
  int x, y, k, xx, xy, *yx;

  if( *bucket <= 0
   || (warm0_weight+warm1_weight+cold0_weight+cold1_weight <= 0))
    return;

  /* collect a list of eligible locations */
  xx= pmap->ysize*pmap->xsize;
  yx= fc_malloc(sizeof(int)*xx);
  for( k= 0, xy= xx; --xy >= 0; ) {
    ptile = (civmap_get_tiles(pmap) + xy);
    if (civtile_get_continent(ptile) == isleindex &&
        civtile_get_terrain(ptile) == T_GRASSLAND)
      yx[k++]= xy;
  }

  /* loop until quota filled or random locations run out */
  xx= *bucket/10 + 2; /* number of (guaranteed) seed locations */
  while (*bucket > 0 && k > 0) {
    y= myrand(k--), x= yx[y], yx[y]= yx[k];
    ptile = (civmap_get_tiles(pmap) + x);
    index_to_civmap_pos(pmap, &x, &y, &x, &y, x);

    /* the first condition helps make terrain more contiguous,
       the next four help control feature clumping
       the last lets it avoid the coast. */
    if( ( (is_terrain_adjc(pmap,x,y,warm0) || is_terrain_adjc(pmap,x,y,cold0))
        || myrand(totalmass) < *bucket
        || xx-- > 0 )
     && (!warm0_weight || count_terrain_adjc(pmap,x,y,warm0) < warm0_weight)
     && (!warm1_weight || count_terrain_adjc(pmap,x,y,warm1) < warm1_weight)
     && (!cold0_weight || count_terrain_adjc(pmap,x,y,cold0) < cold0_weight)
     && (!cold1_weight || count_terrain_adjc(pmap,x,y,cold1) < cold1_weight)
     && (!is_at_coast(pmap, x, y) || myrand(100) < coast )) {
      if ( civmap_is_cold(pmap, x,y) )
        (*bucket)--, civtile_set_terrain(ptile,
          (myrand(cold0_weight+cold1_weight+2) >= cold1_weight)?cold0:cold1);
      else
        (*bucket)--, civtile_set_terrain(ptile,
          (myrand(warm0_weight+warm1_weight+2) >= warm1_weight)?warm0:warm1);
    }
  }
  free(yx);
}

/**************************************************************************
  finds a place and drops the island(s) found in create_map
  forever can be used to limit the number of tries
  Returns the number of land squares placed.
**************************************************************************/
static int place_island(const CIVMAP *pmap, int *create_map, int forever)
{
  int x, y, xx, yy, xy, *yx, k, xo= 0, yo= 0, i=0;

  /* collect a list of eligible starting squares */
  xx= (pmap->ysize-(s-n))*pmap->xsize;
  yx= fc_malloc(sizeof(int)*xx);
  for( k= 0, xy= xx; --xy >= 0; )
    yx[k++]= xy;

  /* loop over the list or requested count, whichever runs out first */
  if(forever == 0) forever= k;  /* convenience */
  while(--forever >= 0 && k > 0) {
    yy= myrand(k--), xx= yx[yy], yx[yy]= yx[k];
    index_to_civmap_pos(pmap, &xx, &yy, &xo, &yo, xx);

    /* a diagonal pre-check helps a lot for maps with high landmass */
    for (y= n, x= w; y <= s && x <= e; y++, x++) {
      native_to_civmap_pos(pmap, &xx, &yy, x, y);
      if( normalize_civmap_pos(pmap, &xx, &yy)
       && create_map[civmap_pos_to_index(pmap, xx,yy)]) {
        xx+= xo - w, yy+= yo - n;
        native_to_civmap_pos(pmap, &xx, &yy, xx, yy);
        if (!normalize_civmap_pos(pmap, &xx, &yy) || is_coastal(pmap, xx, yy))
          goto end_while;
      }
    }

    /* full rectangular overlap check */
    block_xy_iterate2(pmap, xy0, x, y, w, n, e, s) {
      if (create_map[xy0]) {
        xx= x + xo - w, yy= y + yo - n;
        native_to_civmap_pos(pmap, &xx, &yy, xx, yy);
        if (!normalize_civmap_pos(pmap, &xx, &yy) || is_coastal(pmap, xx, yy))
          goto end_while;
      }
    } block_xy_iterate2_end;

    /* finally, update the real map(s) */
    block_xy_iterate2(pmap, xy0, x, y, w, n, e, s) {
      if ((xy= create_map[xy0])) {
        if(checkmass-- <= 0) {
          freelog(LOG_ERROR, "mapgen.c: mass doesn't sum up.");
          forever = 0;
          goto exit_while;
        }
        xx= x + xo - w, yy= y + yo - n;
        native_to_civmap_pos(pmap, &xx, &yy, xx, yy);
        if (normalize_civmap_pos(pmap, &xx, &yy)) {
          CIVTILE *ptile =  civmap_get_tile(pmap, xx, yy);
          civtile_set_terrain(ptile, T_GRASSLAND);
          civtile_set_continent(ptile, isleindex);
          hmap(pmap, xx,yy)= xy;
          i++;
    }
      }
    } block_xy_iterate2_end;
    break;
    end_while:
  }
  exit_while:
  free(yx);

  /* reset to real map bounding rectangle */
  s += yo - n;
  e += xo - w;
  n = yo;
  w = xo;

  return i;
}

/**************************************************************************
  create an island centred in a scratch map when called with islemass != 0
**************************************************************************/
static void create_island(const CIVMAP *pmap, int *height_map, int islemass)
{
  int dd[8] = {1000, 500, 300, 200, 500, 200, 300, 500};
  int hh, xx, yy, xy, *yx, j, k, l, m, d0, d1= 1, na= 0, nc=0;

  /* setup list buffer and initialize the scratch height_map */
  xy= pmap->ysize*pmap->xsize;
  yx= fc_malloc(sizeof(int) * xy);
  memset(height_map,              -1, sizeof(int)*pmap->xsize);
  memset(height_map+pmap->xsize,     0, sizeof(int)*(xy-pmap->xsize*2));
  memset(height_map+xy-pmap->xsize, -1, sizeof(int)*pmap->xsize);

  /* Add a hill at the map centre - bias growth in the d0'th direction */
  xy= pmap->ysize/2, xy= xy*pmap->xsize+pmap->xsize/2;
  yx[na++] = xy;
  height_map[xy]= 7000;
  d0= myrand(8);

  /* Seeding done - now loop */
  while(nc < islemass && na > 0) {
    /* Choose a tile to add - bias choice to higher adjacent */
    j= myrand(na), k= myrand(na), k= (j < k ? j : k) + nc;

    /* move it down to the selected set and reset counts */
    xy= yx[k];
    while( --k >= nc )
      yx[k+1]= yx[k];
    yx[(na--,nc++)]= xy;
    index_to_civmap_pos(pmap, &xx, &yy, &xx, &yy, xy);

    /* Update the directional bias, i.e. non-spheriodal preference  */
    /* in combination with a rapid height dropoff, makes peninsulas */
    /* Note: this uses an octal heartbeat with 1-2-1 change probability */
    if( (nc&7) == 0 )
      d1++, d0= ((myrand(4)-1)/2 + d0) & 7;

    /* more anti-spherism - start a new coastal growth bump */
    if( d1 > 1 ) {
      m= 0;
      adjc_iterate(pmap, xx, yy, x1, y1) {
        m+= (height_map[civmap_pos_to_index(pmap, x1, y1)] == 0 ? 1 : 0);
      } adjc_iterate_end
      /* bump iff really coastal */
      if( m > 4 ) {
        d1= 0, height_map[xy]= 7000;

        /* ... and deflate current/adjacent fill areas */
        for(k= nc+na; --k >= nc;) {
          m= 0, index_to_civmap_pos(pmap, &j, &l, &j, &l, yx[k]);
          adjc_dir_iterate(pmap, j, l, x1, y1, dir) {
            m+= (height_map[civmap_pos_to_index(pmap, x1,y1)] == 0
              ? (DIR_IS_CARDINAL(dir) ? 2 : 1) : 0);
          } adjc_dir_iterate_end
          /* don't deflate singly connected or less lakes */
          if( m > 2 )
            height_map[yx[k]]= (height_map[yx[k]]+1)/2;
        }
      }
    }

    /* Add all uncounted adjacent tiles to the sorted selection list */
    hh= height_map[xy], hh= (hh > 1000 ? hh-1000 : 1);
    adjc_dir_iterate(pmap, xx, yy, x1, y1, dir) {
      j= civmap_pos_to_index(pmap, x1, y1);
      if( height_map[j] == 0 && (l= myrand(dd[DIR_CCW_N(dir,d0)])+hh) > 200 ) {
        height_map[j]= l;
        for(k= nc+(na++); --k >= nc && height_map[yx[k]] < l;)
          yx[k+1]= yx[k];
        yx[k+1]= j;
      }
    } adjc_dir_iterate_end
  }

  /* Remove leftovers - first the polar blocks */
  xy= pmap->ysize*pmap->xsize;
  memset(height_map+xy-pmap->xsize, 0, sizeof(int)*pmap->xsize);
  memset(height_map,              0, sizeof(int)*pmap->xsize);

  /* ... and adjacent but uncounted */
  for(k= nc+na; --k >= nc;)
    height_map[yx[k]]= 0;

  /* Set nwse extents, i.e. native bounding rectangle */
  /* TODO: get rid of n, w, s, e single char globals, use a struct arg */
  n= pmap->ysize, w= pmap->xsize,  s= e= 0;
  for(l= 0, k++; --k >= 0;) {
    l+= (height_map[yx[k]] ? 1 : 0);

    xy= yx[k]/pmap->xsize;
    if( xy < n ) n= xy;
    else
    if( xy > s ) s= xy;

    xx= yx[k]%pmap->xsize;
    if( xx < w ) w= xx;
    else
    if( xx > e ) e= xx;
  }
  free(yx);

  if(nc<islemass) {
    /*  (-: NEVER REACHED :-)  */
    freelog(LOG_ERROR, "create_island ended early with %d/%d.",
            nc, islemass);
  }
}

/**************************************************************************
  make an island, fill every tile type except plains
  note: you have to create big islands first.
**************************************************************************/
static void make_island(CIVMAP *pmap, int islemass, int starters)
{
  static long int tilefactor, balance, lastplaced; /* int may be only 2 byte! */
  static long int riverbuck, mountbuck, desertbuck, forestbuck, swampbuck;
  int i= 0;

  if (!islemass) {

    /* first time initialization */
    balance= 0;     /* setup initial static parameters */
    isleindex=      /* 0= none, 1= arctic, 2= antarctic */
      (pmap->separatepoles ? (!civmap_get_type(pmap, MAP_TYPE_WRAPY) ? 3:2):1);
    checkmass= totalmass;

    /* caveat: this should really be sent to all players */
    if(totalmass > 3000)
      freelog(LOG_NORMAL, _("High landmass - this may take a few seconds."));

    i = (pmap->terrain->river_style == R_AS_TERRAIN) ? pmap->RIVERS : 0;
    i+= pmap->MOUNTAINS + pmap->DESERTS + pmap->FORESTS + pmap->SWAMPS
      + pmap->GRASSLAND;

    tilefactor = i-5;  /* overbudget 5% - smaller islands get dregs */
    riverbuck  = 0;
    mountbuck  = 0;
    desertbuck = 0;
    forestbuck = 0;
    swampbuck  = 0;
    lastplaced = totalmass;
  } else {

    /* makes the islands here */
    if (islemass <= balance) return;

    /* don't create continents without a number */
    if(isleindex >= MAP_NCONT)
      return;
    islands[isleindex].starters = starters;
    islands[isleindex].count = starters;
    freelog(LOG_VERBOSE, "island %i", isleindex);

    /*don't create big isles we can't place*/
    islemass = islemass - balance;
    if(islemass > lastplaced+1+lastplaced/50)
      islemass= lastplaced+1+lastplaced/50;

    i = islemass;
    do {
      create_island(pmap, create_map, i--);
      lastplaced= place_island(pmap, create_map, 0);
    } while (lastplaced > 0 && i*10 > islemass );

    if(lastplaced > 0)
      balance = lastplaced - islemass;
    else
      balance = 0, lastplaced = i;

    freelog(LOG_VERBOSE, "ini=%d, plc=%d, bal=%ld, tot=%ld",
            islemass, i, balance, checkmass);

    mountbuck += (pmap->MOUNTAINS*i)/tilefactor;
    fill_island(pmap, &mountbuck, 30, 5,3, 5,3,
          T_HILLS, T_MOUNTAINS, T_HILLS, T_MOUNTAINS);

    desertbuck += (pmap->DESERTS*lastplaced)/tilefactor;
    fill_island(pmap, &desertbuck, 20, 6,0, 7,3,
          T_DESERT, T_DESERT, T_TUNDRA, T_ARCTIC);

    forestbuck += (pmap->FORESTS*lastplaced)/tilefactor;
    fill_island(pmap, &forestbuck, 60, 7,4, 4,2,
          T_FOREST, T_JUNGLE, T_FOREST, T_SWAMP);

    swampbuck += (pmap->SWAMPS*lastplaced)/tilefactor;
    fill_island(pmap, &swampbuck, 80, 6,2, 6,2,
          T_SWAMP, T_JUNGLE, T_SWAMP, T_HILLS);

    /* map_printstats(pmap, isleindex); */
    isleindex++;
    pmap->num_continents++;
  }
}

/**************************************************************************
  allocate working arrays (caller must free), fill ocean and make polar
**************************************************************************/
static void initworld(CIVMAP *pmap)
{
  CIVTILE *ptile;
  int xy, xn, yn;

  height_map = fc_malloc(sizeof(int) * pmap->ysize * pmap->xsize);
  create_map = fc_malloc(sizeof(int) * pmap->ysize * pmap->xsize);
  islands = fc_malloc((MAP_NCONT+1)*sizeof(struct isledata));
  memset(islands, 0, (MAP_NCONT+1)*sizeof(struct isledata));

  /* ocean */
  whole_civmap_xy_iterate(pmap, xy, xn, yn, x, y) {
    ptile = (civmap_get_tiles(pmap) + xy);
    civtile_set_terrain(ptile, T_OCEAN);
    civtile_set_continent(ptile, 0);
    height_map[xy]= 0;
  } whole_civmap_xy_iterate_end;

  /* polar */
  pmap->num_continents = 0;
  if( pmap->separatepoles ) {
    pmap->num_continents = (!civmap_get_type(pmap, MAP_TYPE_WRAPY) ? 2 : 1);
    for (xn= 0 ; xn < pmap->xsize; xn++) {
      yn = 0, xy = native_pos_to_index(pmap, xn, yn);
      ptile = (civmap_get_tiles(pmap) + xy);
      civtile_set_terrain(ptile, myrand(9) ? T_ARCTIC : T_TUNDRA);
      civtile_set_continent(ptile, 1);
      height_map[xy]= tocean+100;

      if( civmap_get_type(pmap, (MAP_TYPE_ISO|MAP_TYPE_HEX)) ) {
        yn++, xy = native_pos_to_index(pmap, xn, yn);
        ptile = (civmap_get_tiles(pmap) + xy);
        civtile_set_terrain(ptile, myrand(3) ? T_ARCTIC : T_TUNDRA);
        civtile_set_continent(ptile, 1);
        height_map[xy]= tocean+100;
      }

      yn++, xy = native_pos_to_index(pmap, xn, yn);
      if (!myrand(9)) {
        ptile = (civmap_get_tiles(pmap) + xy);
        civtile_set_terrain(ptile, myrand(9) ? T_ARCTIC : T_TUNDRA);
        civtile_set_continent(ptile, 1);
      }
      height_map[xy]= tocean+100;

      yn = pmap->ysize-1, xy = native_pos_to_index(pmap, xn, yn);
      ptile = (civmap_get_tiles(pmap) + xy);
      civtile_set_terrain(ptile, myrand(9) ? T_ARCTIC : T_TUNDRA);
      civtile_set_continent(ptile, pmap->num_continents);
      height_map[xy]= tocean+100;

      if( civmap_get_type(pmap, (MAP_TYPE_ISO|MAP_TYPE_HEX)) ) {
        yn--, xy = native_pos_to_index(pmap, xn, yn);
        ptile = (civmap_get_tiles(pmap) + xy);
        civtile_set_terrain(ptile, myrand(3) ? T_ARCTIC : T_TUNDRA);
        civtile_set_continent(ptile, pmap->num_continents);
        height_map[xy]= tocean+100;
      }

      yn--, xy = native_pos_to_index(pmap, xn, yn);
      if (!myrand(9)) {
        ptile = (civmap_get_tiles(pmap) + xy);
        civtile_set_terrain(ptile, myrand(9) ? T_ARCTIC : T_TUNDRA);
        civtile_set_continent(ptile, pmap->num_continents);
      }
      height_map[xy]= tocean+100;
    }
  }

  /* basic housekeeping */
  make_island(pmap, 0, 0);
  islands[0].starters = 0;
}

/**************************************************************************
  island base map generators
**************************************************************************/
static void mapgenerator2(CIVMAP *pmap)
{
  int i;
  int spares= 1;
  /* constant to make up for islands actually needing additional space
     i.e. ocean boundaries and internal seas */

  if (pmap->landpercent > 85) {
    pmap->generator = 1;
    return;
  }

  adjust_terrain_param(pmap);
  totalmass= ((pmap->ysize-6-spares)*(pmap->xsize-spares)
           * pmap->landpercent)/100;

  /*!PS: The weights NEED to sum up to totalweight (dammit) */
  /* copying the flow of the make_island loops is the safest way */
  totalweight= 0;
  for (i = pmap->num_start_positions ; i ; i--)
    totalweight+= 70;
  for (i = pmap->num_start_positions ; i ; i--)
    totalweight+= 20;
  for (i = pmap->num_start_positions ; i ; i--)
    totalweight+= 10;

  initworld(pmap);

  for (i = pmap->num_start_positions ; i ; i--)
    make_island(pmap, (70*totalmass)/totalweight, 1);
  for (i = pmap->num_start_positions ; i ; i--)
    make_island(pmap, (20*totalmass)/totalweight, 0);
  for (i = pmap->num_start_positions ; i ; i--)
    make_island(pmap, (10*totalmass)/totalweight, 0);

  /* final cleanups using mapgenerator1 routines */
  adjust_map2(pmap);
  make_fair(pmap);
  make_plains(pmap);
  make_rivers(pmap);
  remove_tiny_lakes(pmap);
  setup_seadata(pmap);
  map_check(pmap);

  /* release initworld allocations */
  free(create_map);
  free(height_map);
  create_map= height_map= NULL;

  if(checkmass > pmap->xsize+pmap->ysize+totalweight) {
    freelog(LOG_VERBOSE, "%ld mass left unplaced", checkmass);
  }
}

/**************************************************************************
On popular demand, this tries to mimic the generator 3 as best as possible.
**************************************************************************/
static void mapgenerator3(CIVMAP *pmap)
{
  int poles = (pmap->separatepoles
            ? (!civmap_get_type(pmap, MAP_TYPE_WRAPY) ? 2 : 1) : 0);
  int spares = 1;
  int j = 0;

  long int islandmass,landmass, size;
  long int maxmassdiv6= 20;
  int bigislands;

  if( pmap->xsize < 40 || pmap->ysize < 40 || pmap->landpercent > 80 ) {
    freelog(LOG_NORMAL,_("Falling back to generator 2."));
    pmap->generator = 2;
    return;
  }

  adjust_terrain_param(pmap);
  totalmass= ((pmap->ysize-6-spares)*(pmap->xsize-spares)
           * pmap->landpercent)/100;

  /* subtracting the arctics */
  landmass= ((pmap->ysize-6)*pmap->xsize*pmap->landpercent)/100;
  if( landmass > 3*pmap->ysize+pmap->num_start_positions*3 ) {
    landmass-= 3*pmap->ysize;
  }
  bigislands= pmap->num_start_positions;

  islandmass= (landmass)/(3*bigislands);
  if(islandmass<4*maxmassdiv6 )
    islandmass= (landmass)/(2*bigislands);
  if(islandmass < 3*maxmassdiv6 && pmap->num_start_positions*2 < landmass )
    islandmass= (landmass)/(bigislands);

  if(islandmass < 2)
    islandmass= 2;
  else
  if(islandmass > maxmassdiv6*6)
    islandmass= maxmassdiv6*6;  /* !PS: let's try this */

  initworld(pmap);

  while(isleindex-poles <= bigislands && checkmass > islandmass && ++j < 500) {
    make_island(pmap, islandmass, 1);
  }

  if(j == 500){
    freelog(LOG_NORMAL, _("Generator 3 didn't place all big islands."));
  }

  islandmass= (islandmass*11)/8;
  /*!PS: I'd like to mult by 3/2, but starters might make trouble then*/
  if(islandmass < 2)
    islandmass= 2;

  while(isleindex <= MAP_NCONT-20 && checkmass > islandmass && ++j < 1500) {
      if(j < 1000)
        size = myrand((islandmass+1)/2+1)+islandmass/2;
      else
        size = myrand((islandmass+1)/2+1);
      if(size < 2)
        size= 2;

      make_island(pmap, size, (isleindex-poles <= pmap->num_start_positions) ? 1 : 0 );
  }

  /* final cleanups using mapgenerator1 routines */
  adjust_map2(pmap);
  make_fair(pmap);
  make_plains(pmap);
  make_rivers(pmap);
  remove_tiny_lakes(pmap);
  setup_seadata(pmap);
  map_check(pmap);

  /* release initworld allocations */
  free(create_map);
  free(height_map);
  create_map= height_map= NULL;

  if(j == 1500) {
    freelog(LOG_NORMAL,_("Generator 3 left %li landmass unplaced."),checkmass);
  } else if (checkmass > pmap->xsize+pmap->ysize+totalweight) {
    freelog(LOG_VERBOSE, "%ld mass left unplaced", checkmass);
  }
}

/**************************************************************************
...
**************************************************************************/
static void mapgenerator4(CIVMAP *pmap)
{
  int bigweight=70;
  int spares= 1;
  int i;

  /* no islands with mass >> sqr(min(xsize,ysize)) */

  i = pmap->num_start_positions / 2;
  if ( pmap->num_start_positions < 2 || pmap->landpercent > 80) {
    pmap->generator = 3;
    return;
  }

  if (pmap->landpercent > 60)
    bigweight=30;
  else if (pmap->landpercent > 40)
    bigweight=50;
  else
    bigweight=70;

  spares= (pmap->landpercent-5)/30;

  adjust_terrain_param(pmap);
  totalmass = ((pmap->ysize-6-spares)*(pmap->xsize-spares)
            * pmap->landpercent)/100;

  /*!PS: The weights NEED to sum up to totalweight (dammit) */
  totalweight= 0;
  if (pmap->num_start_positions % 2)
    totalweight+= bigweight*3;
  else
    i++;
  while (--i > 0)
    totalweight+= bigweight*2;
  for (i = pmap->num_start_positions ; i > 0 ; i--)
    totalweight+= 20;
  for (i = pmap->num_start_positions ; i > 0 ; i--)
    totalweight+= 10;

  initworld(pmap);

  i = pmap->num_start_positions / 2;
  if (pmap->num_start_positions % 2)
    make_island(pmap, (bigweight*3*totalmass)/totalweight, 3);
  else
    i++;
  while (--i > 0)
    make_island(pmap, (bigweight*2*totalmass)/totalweight, 2);
  for (i = pmap->num_start_positions ; i > 0 ; i--)
    make_island(pmap, (20*totalmass)/totalweight, 0);
  for (i = pmap->num_start_positions ; i > 0 ; i--)
    make_island(pmap, (10*totalmass)/totalweight, 0);

  /* final cleanups using mapgenerator1 routines */
  adjust_map2(pmap);
  make_fair(pmap);
  make_plains(pmap);
  make_rivers(pmap);
  remove_tiny_lakes(pmap);
  setup_seadata(pmap);
  map_check(pmap);

  /* release initworld allocations */
  free(create_map);
  free(height_map);
  create_map= height_map= NULL;

  if(checkmass > pmap->xsize+pmap->ysize+totalweight) {
    freelog(LOG_VERBOSE, "%ld mass left unplaced", checkmass);
  }
}

/**************************************************************************
Recursive function which does the work for generator 5
**************************************************************************/
static void gen5rec(const CIVMAP *pmap, 
  int step, int x0, int y0, int x1, int y1) {
  int ix[3], iy[3];
  int xy00, xy10, xy20, xy01, xy11, xy21, xy02, xy12, xy22;

  /* Need a midpoint in at least one direction to fill in. */
  if ((y1 - y0 <= 1) && (x1 - x0 <= 1)) {
    return;
  }
  /* Sanity check - paranoia */
  if ((y1 - y0 < 0) || (x1 - x0 < 0)) {
    return;
  }

  /* indices of the current 3x3 grid (Note: 3x2, 1x3, etc. possible) */
  ix[0] = x0;
  ix[2] = (x1 >= pmap->xsize) ? 0 : x1;
  iy[0] = y0;
  iy[2] = (y1 >= pmap->ysize) ? 0 : y1;

  /* Compute the "midpoint" of the range
   * Note: asymmetric "rectangles" introduce cartesian linear features
   *       This is not particularly pleasing in many cases.
  */
  /* weighted at 20-80% of the range
   * ix[1] = x0;
   * if ((x1 - x0) > 1) {
   *   ix[1] = myrand(61) + 20; -- or --
   *   ix[1] = myrand(31) - myrand(31) + 50;
   *     ix[1] = ((100 - ix[1]) * x0 + ix[1] * x1) / 100;
   *   if (ix[1] == x0)
   *     ix[1]++;
   *   else
   *   if (ix[1] == x1)
   *     ix[1]--;
   * }
   */
  ix[1] = (x0 + x1)/2;

  /* weighted at 20-80% of the range
   * iy[1] = y0;
   * if ((y1 - y0) > 1) {
   *   iy[1] = myrand(61) + 20; -- or --
   *   iy[1] = myrand(31) - myrand(31) + 50;
   *     iy[1] = ((100 - iy[1]) * y0 + iy[1] * y1) / 100;
   *   if (iy[1] == y0)
   *     iy[1]++;
   *   else
   *   if (iy[1] == y1)
   *     iy[1]--;
   * }
   */
  iy[1] = (y0 + y1)/2;

  /* Set midpoints of sides to avg of side vertices plus a random factor */
  /* Unset points are zero, don't reset if set. This can happen if less  */
  /* than a full 3x3 or there are preset "features".                     */

  xy00 = native_pos_to_index(pmap, ix[0], iy[0]);
  xy10 = native_pos_to_index(pmap, ix[1], iy[0]);
  xy20 = native_pos_to_index(pmap, ix[2], iy[0]);
  xy01 = native_pos_to_index(pmap, ix[0], iy[1]);
  xy11 = native_pos_to_index(pmap, ix[1], iy[1]);
  xy21 = native_pos_to_index(pmap, ix[2], iy[1]);
  xy02 = native_pos_to_index(pmap, ix[0], iy[2]);
  xy12 = native_pos_to_index(pmap, ix[1], iy[2]);
  xy22 = native_pos_to_index(pmap, ix[2], iy[2]);

  if (height_map[xy10] == 0)
    height_map[xy10] = (height_map[xy00] + height_map[xy20])/2
      + myrand(step) - step/2;
  if (height_map[xy12] == 0)
    height_map[xy12] = (height_map[xy02] + height_map[xy22])/2
      + myrand(step)- step/2;
  if (height_map[xy01] == 0)
    height_map[xy01] = (height_map[xy00] + height_map[xy02])/2
      + myrand(step) - step/2;
  if (height_map[xy21] == 0)
    height_map[xy21] = (height_map[xy20] + height_map[xy22])/2
      + myrand(step) - step/2;

  /* Set middle to average of midpoints plus a random factor
   * TODO:
   * 1)  better handling of saddle point conditions for ridgeline and
   *     valley/trench effects
   * 2)  low-probability "super"-step perhaps gated by "too smooth"
   */
  if (height_map[xy11] == 0)
    height_map[xy11] = ( height_map[xy01] + height_map[xy21]
      + height_map[xy10] + height_map[xy12] )/4 + myrand(step) - step/2;

  /* Recurse for each quadrant sub-rectangle.  Block update conditions are
   * tested at the start of each iteration, so no need to check here.
   */
  gen5rec(pmap, 2*step/3, x0, y0, ix[1], iy[1]);
  gen5rec(pmap, 2*step/3, x0, iy[1], ix[1], y1);
  gen5rec(pmap, 2*step/3, ix[1], y0, x1, iy[1]);
  gen5rec(pmap, 2*step/3, ix[1], iy[1], x1, y1);
}

/**************************************************************************
Generator 5 makes earthlike worlds with one or more large continents and
a scattering of smaller islands. It does so by dividing the world into
blocks and on each block raising or lowering the corners, then the
midpoints and middle and so on recursively.
Fiddling with 'xdiv' and 'ydiv' will vary the uniformity of initial blocks.
If the map does not wrap in a given direction, 'avoidedge' will reduce the
chance of raising land near the corresponding edges.
**************************************************************************/
static void gen5_hmap(const CIVMAP *pmap)
{
  int x, y, tmp;
  int xwrap = 1; /* could come from topology */
  int ywrap = 0; /* could come from topology */
  int xmax = pmap->xsize-1 + xwrap;
  int ymax = pmap->ysize-1 + ywrap;
  int xdiv = 6; /* how many blocks should the x and y directions be */
  int ydiv = 4; /* divided into initially */
  int avoidedge = 100; /* edges are avoided more strongly as this increases */
  /* just need something > log(max(xsize, ysize)) for the recursion */
  int step = (pmap->xsize + pmap->ysize) * 100;

  /* initialize height_map data */
  whole_civmap_xy_iterate(pmap, xy, xn, yn, x, y) {
    height_map[xy] = 0;
  } whole_civmap_xy_iterate_end;

  /* set some initial points - blockcount plus end point if not wrapped  */
  /* Note: the effect here is equal random maximum on a regular grid.    */
  /*       once the iteration starts, step is decreased steadily, and/or */
  /*       midpoints are less regular.                                   */
  /*       However, first and last gives almost the same result.         */
  for (y = ydiv+1 - ywrap; y-- > 0;)
  for (x = xdiv+1 - xwrap; x-- > 0;) {
    int xy = native_pos_to_index(pmap, x*xmax/xdiv, y*ymax/ydiv);
    height_map[xy] =  myrand(2*step) - (2*step)/2;
  }

  /* if we aren't wrapping stay away from edges to some extent */
  /* tmp is a rough approximation to avoidedge below sea level */
  tmp = (pmap->landpercent - 50) * step / 100 - avoidedge;
  if (xwrap == 0)
    for (y = ydiv+1 - ywrap; y-- > 0;) {
      int yy = y*ymax/ydiv;
      int xy0 = native_pos_to_index(pmap, 0, yy);
      int xy1 = native_pos_to_index(pmap, 2, yy);
      int xy2 = native_pos_to_index(pmap, xmax-0, yy);
      int xy3 = native_pos_to_index(pmap, xmax-2, yy);
      height_map[xy0] = height_map[xy0]/4 + tmp;
      height_map[xy1] = height_map[xy0] - myrand(4*avoidedge) + avoidedge;
      height_map[xy2] = height_map[xy2]/4 + tmp;
      height_map[xy3] = height_map[xy2] - myrand(xmax-4*avoidedge) + avoidedge;
    }
  if (ywrap == 0)
    for (x = xdiv+1 - xwrap; x-- > 0;) {
      int xx = x*xmax/ydiv;
      int xy0 = native_pos_to_index(pmap, xx, 0);
      int xy1 = native_pos_to_index(pmap, xx, 2);
      int xy2 = native_pos_to_index(pmap, xx, ymax-0);
      int xy3 = native_pos_to_index(pmap, xx, ymax-2);
      height_map[xy0] = height_map[xy0]/4 + tmp;
      height_map[xy1] = height_map[xy0] - myrand(4*avoidedge) + avoidedge;
      height_map[xy2] = height_map[xy2]/4 + tmp;
      height_map[xy3] = height_map[xy2] - myrand(xmax-4*avoidedge) + avoidedge;
    }

  /* Ok, now calculate recursively on each block */
  for (y = ydiv; y-- > 0;)
  for (x = xdiv; x-- > 0;)
    gen5rec(pmap, step,
      x*xmax/xdiv, y*ymax/ydiv, (x+1)*xmax/xdiv,(y+1)*ymax/ydiv);

  /* put in some final random fuzz */
  whole_civmap_xy_iterate(pmap, xy, xn, yn, x, y) {
    height_map[xy] += myrand(4) - 2;
  } whole_civmap_xy_iterate_end;

  /* Rescale to the standard format (0-3000 ocean, 3000-10000 land) */
  adjust_map(pmap);
}

/**************************************************************************
This generator makes earthlike worlds with one or more large continents and
a scattering of smaller islands. It does so by dividing the world into
blocks and on each block raising or lowering the corners, then the
midpoints and middle and so on recursively.  Fiddling with 'xdiv' and
'ydiv', and, if the map does not wrap in at least one direction, 'avoidedge'
will vary the result.
**************************************************************************/
static void mapgenerator5(CIVMAP *pmap)
{
  /* allocate working storage */
  height_map= fc_malloc(pmap->xsize*pmap->ysize*sizeof(int));

  /* input (and if necessary convert) map generation parameters */
  adjust_terrain_param(pmap);

  /* generate height_map data */
  gen5_hmap(pmap);

  /* populate the generated height_map */
  make_land(pmap);

  free(height_map);
  height_map = NULL;
}

/**************************************************************************
  See stdinhand.c for information on map generation methods.

FIXME: Some continent numbers are unused at the end of this function, fx
       removed completely by remove_tiny_islands.
       When this function is finished various data is written to "islands",
       indexed by continent numbers, so a simple renumbering would not
       work...
**************************************************************************/
CIVMAP *civmap_gen(CIVMAP *aMap)
{
  /* save the current random state: */
  RANDOM_STATE rstate = get_myrand_state();

  /* Sanity */
  if (!aMap)
    aMap= civmap_init(NULL);

  if (!aMap->terrain)
      aMap->terrain = civterrain_init(NULL);
  else
  if (!aMap->terrain->types) {
    /* override - once set it becomes const */
    CIVTERRAIN *pter = (CIVTERRAIN *)aMap->terrain;
    pter->types = civterrain_types_init(NULL, 0, pter);
  }

  if (aMap->seed == 0)
    aMap->seed = (myrand(MAX_UINT32) ^ time(NULL)) & (MAX_UINT32 >> 1);
  mysrand(aMap->seed);

  /* don't generate tiles with mapgen == 0 as we've loaded them from file */
  /* also, don't delete (the handcrafted!) tiny islands in a scenario */
  if (aMap->generator != 0) {
    civmap_tiles_init(aMap);

    /* if one mapgenerator fails, it will choose another mapgenerator */
    /* with a lower number to try again */
    if (aMap->generator == 5 )
      mapgenerator5(aMap);
    if (aMap->generator == 4 )
      mapgenerator4(aMap);
    if (aMap->generator == 3 )
      mapgenerator3(aMap);
    if( aMap->generator == 2 )
      mapgenerator2(aMap);
    if( aMap->generator == 1 )
      mapgenerator1(aMap);
  }

  if (!aMap->has_specials) /* some scenarios already provide specials */
    add_specials(aMap, aMap->riches);

  if (!aMap->has_huts)
    make_huts(aMap, aMap->huts);

  /* restore previous random state: */
  set_myrand_state(rstate);
  return aMap;
}

static void map_check(const CIVMAP *pmap)
{
  int tx, err = 0;
  whole_civmap_iterate(pmap, x, y) {
    int cont = civmap_get_continent(pmap, x, y);
    int sea  = civmap_get_ocean(pmap, x, y);
    if ((tx= civmap_get_terrain(pmap, x, y)) == T_OCEAN) {
      if(cont != 0) {
        err++;
        freelog(LOG_ERROR, "Error: x,y,tx,cont=(%d,%d,%d,%d)\n", x,y,tx,cont);
      }
      if( sea == 0) {
        err++;
        freelog(LOG_ERROR, "Error: x,y,tx,sea=(%d,%d,%d,%d)\n", x,y,tx,sea);
      }
      adjc_iterate(pmap, x, y, x1, y1) {
        if ((cont= civmap_get_ocean(pmap, x1, y1)) != 0 && cont != sea
          && !civmap_has_specials(pmap, x1, y1, S_RIVER)) {
          err++;
          freelog(LOG_ERROR,"Error: x,y,sea=(%d,%d,%d) x1,y1,sea1=(%d,%d,%d)\n",
            x,y,sea, x1,y1,cont);
        }
      } adjc_iterate_end;
    } else {
      if(cont == 0) {
        err++;
         freelog(LOG_ERROR, "Error: x,y,tx,cont=(%d,%d,%d,%d)\n", x,y,tx,cont);
      }
      if( !civmap_has_specials(pmap, x,y, S_RIVER) && sea != 0) {
        err++;
        freelog(LOG_ERROR, "Error: x,y,tx,sea=(%d,%d,%d,%d)\n", x,y,tx,sea);
      }
      adjc_iterate(pmap, x, y, x1, y1) {
        if (civmap_get_terrain(pmap, x1, y1) != T_OCEAN
         && (sea= civmap_get_continent(pmap, x1, y1)) != cont) {
          err++;
          freelog(LOG_ERROR,
            "Error: x,y,cont=(%d,%d,%d) x1,y1,cont1=(%d,%d,%d)\n",
            x,y,cont, x1,y1,sea);
        }
      } adjc_iterate_end;
    }
  } whole_civmap_iterate_end;

  assert( err == 0 );
}

void map_printstats(const CIVMAP *pmap, int ncont)
{
  static int  ix[] = {0,11,1,10,5,2,6,4,8,3,9,7,12,13,14,15};
  static char tx[] = "adfghjm~prstXDEF";
  char buf[256];
  int  xx[16];
  int  nn, tt;

  freelog(LOG_NORMAL,
    "\n  Stats: maxval=%d thills=%d tocean=%d trough=%d isles=%d seas=%d",
    maxval, thills, tocean, trough, pmap->num_continents, pmap->num_oceans);

  for (tt= sizeof(xx); tt-- > 0;) xx[tt]= 0;
  whole_civmap_xy_iterate(pmap, xy, xn, yn, x, y) {
    CIVTILE *ptile = (civmap_get_tiles(pmap) + xy);
    if( ncont > 0 && civtile_get_continent(ptile) != ncont )
      continue;
    xx[15]+= (civtile_get_continent(ptile) != 0);
    xx[9]+= (civtile_has_specials(ptile, S_RIVER) ? 1 : 0);
    xx[(tt= civtile_get_terrain(ptile)) & 15]++;
    if( tt > 15 )
      freelog(LOG_ERROR, "*** Bad terrain type (%d,%d)=%4.4x", x, y, tt);
  } whole_civmap_xy_iterate_end

  buf[0]= buf[sizeof(buf)-1]= 0;
  my_snprintf(buf, sizeof(buf), "%2d=%4d ", ncont, xx[15]);
  /* Note there are only 12 currently used terrain types */
  for (tt= 12; tt-- > 0;) {
    nn= strlen(buf);
    my_snprintf(buf+nn, sizeof(buf)-nn, "%4d%c ", xx[ix[tt]], tx[ix[tt]]);
  }
  nn= strlen(buf), buf[nn-1]=0;
  freelog(LOG_NORMAL, buf);
  freelog(LOG_NORMAL, " ");
}

/* Debugging routines */

void map_printtmap(const CIVMAP *pmap)
{
  char buf[256];
  char tx[] = "adfghjm.prstX123?";
  int  nn, x,y, xy, tt, errs = 0;

  freelog(LOG_NORMAL,
    "\n  Tmap: maxval=%d thills=%d tocean=%d trough=%d isles=%d seas=%d",
    maxval, thills, tocean, trough, pmap->num_continents, pmap->num_oceans);
  for (y= 0; y < pmap->ysize; y++) {
    for (nn= 0, x= 0; x < pmap->xsize; x++) {
      xy = native_pos_to_index(pmap, x, y);
      tt= civmap_get_tiles(pmap)[xy].terrain;
      tt= !RANGE_CHECK_0(T_COUNT, tt) ? (errs++, 16) : tt;
      buf[nn++]= tx[tt];
    }
    buf[nn++]= '\0';
    freelog(LOG_NORMAL, buf);
  }
  if( errs )
    freelog(LOG_NORMAL, "*** Terrain Errors: %d\n", errs);
}

void map_printcmap(const CIVMAP *pmap)
{
  char buf[256];
  char tx[]= ".abcdefghijklmnopqrstuvwxyz$ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";
  int  nn, x,y, xm, ym;

  freelog(LOG_NORMAL,
    "\n  Cmap: maxval=%d thills=%d tocean=%d trough=%d isles=%d seas=%d",
    maxval, thills, tocean, trough, pmap->num_continents, pmap->num_oceans);
  for (y= 0; y < pmap->ysize; y++) {
    for (nn= 0, x= 0; x < pmap->xsize; x++) {
      native_to_civmap_pos(pmap, &xm, &ym, x, y);
      buf[nn++]= tx[civmap_get_continent(pmap, xm, ym)&63];
    }
    buf[nn++]= '\0';
    freelog(LOG_NORMAL, buf);
  }
}

void map_printOmap(const CIVMAP *pmap)
{
  char tx[] = "-123456789abcdefghijklmnopqrstuvwxyz";
  int  x,y, xm, ym;

  printf("\nOmap: maxval=%d thills=%d tocean=%d trough=%d isles=%d seas=%d\n",
    maxval, thills, tocean, trough, pmap->num_continents, pmap->num_oceans);
  for (y= 0; y < pmap->ysize; y++) {
    for (x= 0; x < pmap->xsize; x++) {
      native_to_civmap_pos(pmap, &xm, &ym, x, y);
      putchar(tx[civmap_get_ocean(pmap, xm, ym)]);
    }
    putchar('\n');
  }
}

#ifndef NDEBUG

/* the following only work when the internal arrays are allocated  */
/* they are primarily for use when debugging, e.g. called from gdb */

void map_printhmap(const CIVMAP *pmap, int *height_map)
{
  char tx[] = ".abcdefghijklmnopqrstuvwxyz";
  int  x,y, nn;

  if (!height_map)
    return;

  printf("\nHmap: maxval=%d thills=%d tocean=%d trough=%d isles=%d seas=%d\n",
    maxval, thills, tocean, trough, pmap->num_continents, pmap->num_oceans);
  for (y= 0; y < pmap->ysize; y++) {
    for (x= 0; x < pmap->xsize; x++)
      if((nn= height_map[pmap->xsize*y+x]) < 0)
        putchar(tx[25]);
      else
      if(nn < tocean)
        putchar(tx[((tocean-nn)*25)/(tocean+1)]);
      else
      if(nn < maxval)
        putchar(tx[((nn-tocean)*25)/(maxval-tocean)+1]-32);
      else
        putchar(tx[25]-32);
    putchar('\n');
  }
}
#endif  /* NDEBUG */

