/**********************************************************************
 Freeciv - Copyright (C) 1996 - A Kjeldberg, L Gregersen, P Unold
 civterrain.c - Copyright (C) 2002 - R. Wetmore

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
***********************************************************************/

#define FC__CIVTERRAIN_C

/**
 * Functions for accessing and updating the CIVTERRAIN structure.
 *
 * TODO: ruleset load/save
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "mem.h"
#include "support.h"

#include "civterrain.h"


/***********************************************************************
  Free memory associated with a single terrain type.
***********************************************************************/
void civterrain_type_free(CIVTERRAIN_TYPE *ptype)
{
  if (ptype && ptype->helptext) {
    free(ptype->helptext);
    ptype->helptext = NULL;
  }
}

/***********************************************************************
  Free memory associated with a vector of terrain types.
  For now, special types are hardcoded defaults, never freed.
***********************************************************************/
void civterrain_types_free(CIVTERRAIN_TYPE **ptype, int count)
{
  if (!ptype || !*ptype)
    return;

  /* Convenience */
  if (count <= 0)
    count = T_COUNT;

  while (--count >= 0 ) {
    civterrain_type_free(&(*ptype)[count]);
  }
  free(*ptype);
  *ptype = NULL;
}

/***********************************************************************
  (Allocate and) init memory associated with a vector of terrain types.
  For now, special types are hardcoded defaults, never allocated.
***********************************************************************/
CIVTERRAIN_TYPE *civterrain_types_init(CIVTERRAIN_TYPE *ptype, int count,
  CIVTERRAIN *pglobal)
{
  /* Convenience */
  if ((unsigned)(count-1) >= T_COUNT)
    count = T_COUNT;

  if (!ptype) {
    ptype = fc_malloc(sizeof(CIVTERRAIN_TYPE) * count);   
    memset(ptype, 0, sizeof(CIVTERRAIN_TYPE) * count);   
  }

  while (--count >= 0 ) {
    civterrain_type_free(&ptype[count]);
    ptype[count] = CIVTERRAIN_TYPE_DEFAULT[count];
    ptype[count].global = pglobal;
  }

  return ptype;
}

/***********************************************************************
  Free memory associated with terrain type control.
***********************************************************************/
void civterrain_free(CIVTERRAIN **pter)
{
  if (!pter || !*pter)
    return;

  if ((*pter)->river_helptext) {
    free((*pter)->river_helptext);
    (*pter)->river_helptext = NULL;
  }
  civterrain_types_free((CIVTERRAIN_TYPE **)&(*pter)->types, T_COUNT);

  free(*pter);
  *pter = NULL;
}

/***********************************************************************
  (Allocate and) init memory associated with terrain type control.
***********************************************************************/
CIVTERRAIN *civterrain_init(CIVTERRAIN *pter)
{
  if (!pter) {
    pter = fc_malloc(sizeof(CIVTERRAIN));
    *pter = CIVTERRAIN_DEFAULT;
    pter->types = NULL;
  } else 
  if (pter->river_helptext) {
    free(pter->river_helptext);
    pter->river_helptext = NULL;
  }
  pter->types = civterrain_types_init(
    (CIVTERRAIN_TYPE *)pter->types, T_COUNT, pter);
  return pter;
}


/***********************************************************************
  Return the terrain_type definitions for terrain type "type".
  This is the main access point for all terrain_type lookups.
***********************************************************************/
const CIVTERRAIN_TYPE *civterrain_get_type(const CIVTERRAIN *pter, 
  enum terrain_type type)
{
  if (!((unsigned)type < T_COUNT)) {
    assert((unsigned)type < T_COUNT);
    type = T_UNKNOWN;
  }
  return pter->types + type;
}

/***********************************************************************
  Return the terrain name given its enum index "type".
***********************************************************************/
const char *civterrain_get_name(const CIVTERRAIN *pter, 
  enum terrain_type type)
{
  return civterrain_get_type(pter, type)->name;
}


/***********************************************************************
  Return the enum index of the terrain type given its "name".
***********************************************************************/
enum terrain_type civterrain_get_type_by_name(const CIVTERRAIN *pter, 
  const char *name)
{
  enum terrain_type tt = T_COUNT;
  while (--tt >= 0) {
    if (0 == strcmp(pter->types[tt].name, name)
     || 0 == strcmp(pter->types[tt].special[1].name, name)
     || 0 == strcmp(pter->types[tt].special[2].name, name)) {
      return tt;
    }
  }
  return T_UNKNOWN;
}


/***********************************************************************
  Return the enum (bit) of the special type given its "name".
***********************************************************************/
enum special_type civterrain_get_special_by_name(const CIVTERRAIN *pter,
  const char *name)
{
  int tt = S_COUNT;

  while (--tt >= 0) {
    if (0 == strcmp(name, pter->stypes[tt].name))
      return pter->stypes[tt].spe;
  }
  return S_NO_SPECIAL;
}

/***********************************************************************
  Return the special name given its enum index "type".
  If type is a set, the highest member is returned.
***********************************************************************/
const char *civterrain_get_special_name(const CIVTERRAIN *pter, 
  enum special_type spe)
{
  /* TODO: this should really match vs the new generalized stypes->spe. */
  unsigned tt = -1;
  while (spe > 0)
    tt++, spe >>= 1;

  if (tt < S_COUNT)
    return pter->stypes[tt].name;

  /* Npte: there is no name for S_NO_SPECIAL. */
  return NULL;
}

/***********************************************************************
  Return any prerequsite relationship between specials
***********************************************************************/
int civterrain_get_special_prerequisite(const CIVTERRAIN *pter, int spe)
{
  /* TODO: this should really match vs the new generalized stypes->spe. */
  unsigned tt = -1;
  while (spe > 0)
    tt++, spe >>= 1;

  if (tt < S_COUNT)
    return pter->stypes[tt].prereq;

  return S_NO_SPECIAL;
}


/***********************************************************************
  Return the basic food value for the fully qualified terrain type.
***********************************************************************/
int civterrain_type_get_food(const CIVTERRAIN_TYPE *ptype, 
  enum special_type spe)
{
  int food = 0;

  if (spe & S_SPECIAL_2)
    food += ptype->special[2].food;
  else
  if (spe & S_SPECIAL_1)
    food += ptype->special[1].food;
  else
    food += ptype->food;

  /* farmland requires a (city with) supermarket - only base terrain here */
  if (spe & S_IRRIGATION)
    food += ptype->irrigation_food_incr;

  if (ptype->global) {
    if (spe & S_RAILROAD)
      food += (food * ptype->global->rail_food_bonus)/100;
    if (spe & S_POLLUTION)
      food -= (food * ptype->global->pollution_food_penalty)/100;
    if (spe & S_FALLOUT)
      food -= (food * ptype->global->fallout_food_penalty)/100;
  }

  return food;
}

/***********************************************************************
  Return the basic food value for the fully qualified terrain type.
***********************************************************************/
int civterrain_type_get_food_base(const CIVTERRAIN_TYPE *ptype,
  enum special_type spe)
{
  return civterrain_type_get_food(ptype, spe & (S_SPECIAL_1 | S_SPECIAL_2));
}

/***********************************************************************
  Return the basic shield value for the fully qualified terrain type.
***********************************************************************/
int civterrain_type_get_shield(const CIVTERRAIN_TYPE *ptype, 
  enum special_type spe)
{
  int shield = 0;

  if (spe & S_SPECIAL_2)
    shield += ptype->special[2].shield;
  else
  if (spe & S_SPECIAL_1)
    shield += ptype->special[1].shield;
  else
    shield += ptype->shield;

  if (spe & S_MINE)
    shield += ptype->mining_shield_incr;

  if (ptype->global) {
    if (spe & S_RAILROAD)
      shield += (shield * ptype->global->rail_shield_bonus)/100;
    if (spe & S_POLLUTION)
      shield -= (shield * ptype->global->pollution_shield_penalty)/100;
    if (spe & S_FALLOUT)
      shield -= (shield * ptype->global->fallout_shield_penalty)/100;
  }

  return shield;
}

/***********************************************************************
  Return the basic shield value for the fully qualified terrain type.
***********************************************************************/
int civterrain_type_get_shield_base(const CIVTERRAIN_TYPE *ptype,
  enum special_type spe)
{
  return civterrain_type_get_shield(ptype, spe & (S_SPECIAL_1 | S_SPECIAL_2));
}

/***********************************************************************
  Return the basic trade value for the fully qualified terrain type.
***********************************************************************/
int civterrain_type_get_trade(const CIVTERRAIN_TYPE *ptype, 
  enum special_type spe)
{
  int trade = 0;

  if (spe & S_SPECIAL_2)
    trade += ptype->special[2].trade;
  else
  if (spe & S_SPECIAL_1)
    trade += ptype->special[1].trade;
  else
    trade += ptype->trade;

  if (spe & S_ROAD)
    trade += ptype->road_trade_incr;

  if (ptype->global) {
    if (spe & S_RIVER || ptype->base == T_RIVER)
      trade += ptype->global->river_trade_incr;
    if (spe & S_RAILROAD)
      trade += (trade * ptype->global->rail_trade_bonus)/100;
    if (spe & S_POLLUTION)
      trade -= (trade * ptype->global->pollution_trade_penalty)/100;
    if (spe & S_FALLOUT)
      trade -= (trade * ptype->global->fallout_trade_penalty)/100;
  }

  return trade;
}

/***********************************************************************
  Return the basic trade value for the fully qualified terrain type.
***********************************************************************/
int civterrain_type_get_trade_base(const CIVTERRAIN_TYPE *ptype,
  enum special_type spe)
{
  return civterrain_type_get_trade(ptype, spe & (S_SPECIAL_1 | S_SPECIAL_2));
}

/***********************************************************************
  Return a (static) string with special(s) name(s);
  eg: "Mine"
  eg: "Road/Farmland"
***********************************************************************/
const char *civterrain_type_get_infrastructure_text(const CIVTERRAIN_TYPE *pter,
  enum special_type spe)
{
  static char s[64];
  int  set;
  
  /* Nullstring including space for leading slash */
  s[0] = s[1] = '\0';

  /* Since railroad requires road, Road/Railroad is redundant */
  if ((set = spe & (S_RAILROAD | S_ROAD)))
    cat_snprintf(s, sizeof(s), "/%s", 
      civterrain_get_special_name(pter->global, set));

  /* Likewise for farmland on irrigation or mine */
  if ((set = spe & (S_FARMLAND | S_IRRIGATION | S_MINE)))
    cat_snprintf(s, sizeof(s), "/%s", 
      civterrain_get_special_name(pter->global, set));

  if ((set = spe & (S_FORTRESS | S_AIRBASE)))
    cat_snprintf(s, sizeof(s), "/%s", 
      civterrain_get_special_name(pter->global, set));

  /* return string without the leading slash */
  return s + 1;
}

/***********************************************************************
  Return a (static) string with terrain info;
  eg: "Hills"
  eg: "Hills (Coal)"
  eg: "Hills (Coal) [Pollution]"
***********************************************************************/
const char *civterrain_type_get_info_text(const CIVTERRAIN_TYPE *pter,
  enum special_type spe)
{
  static char s[64];
  char *sep;

  sz_strlcpy(s, pter->name);

  if ((spe & S_RIVER)) {
    sz_strlcat(s, "/");
    sz_strlcat(s, pter->global->stypes[S_RIVER_BIT].name);
  }

  if ((spe & S_SPECIAL_2)) {
    sz_strlcat(s, " (");
    sz_strlcat(s, pter->special[2].name);
    sz_strlcat(s, ")");
  } else
  if ((spe & S_SPECIAL_1)) {
    sz_strlcat(s, " (");
    sz_strlcat(s, pter->special[1].name);
    sz_strlcat(s, ")");
  }

  sep = " [";
  if ((spe & S_POLLUTION)) {
    sz_strlcat(s, sep);
    sz_strlcat(s, civterrain_get_special_name(pter->global, S_POLLUTION));
    sep = "/";
  }
  if ((spe & S_FALLOUT)) {
    sz_strlcat(s, sep);
    sz_strlcat(s, civterrain_get_special_name(pter->global, S_FALLOUT));
    sep = "/";
  }
  if (sep[0] == '/') {
    sz_strlcat(s, "]");
  }

  return s;
}

/***********************************************************************
  Convenience access method.
***********************************************************************/
int civterrain_type_road_time(const CIVTERRAIN_TYPE *pter)
{
  return pter->road_time;
}

/***********************************************************************
  Convenience access method.
***********************************************************************/
int civterrain_type_railroad_time(const CIVTERRAIN_TYPE *pter)
{
  return 3;
}

/***********************************************************************
  Convenience access method.
***********************************************************************/
int civterrain_type_irrigation_time(const CIVTERRAIN_TYPE *pter)
{
  return pter->irrigation_time;
}

/***********************************************************************
  Convenience access method.
***********************************************************************/
int civterrain_type_mine_time(const CIVTERRAIN_TYPE *pter)
{
  return pter->mining_time;
}

/***********************************************************************
  Convenience access method.
***********************************************************************/
int civterrain_type_transform_time(const CIVTERRAIN_TYPE *pter)
{
  return pter->transform_time;
}

/***********************************************************************
  Convenience access method.
***********************************************************************/
int civterrain_type_fortress_time(const CIVTERRAIN_TYPE *pter)
{
  return 3;
}

/***********************************************************************
  Convenience access method.
***********************************************************************/
int civterrain_type_airbase_time(const CIVTERRAIN_TYPE *pter)
{
  return 3;
}

/***********************************************************************
  Convenience access method.
***********************************************************************/
int civterrain_type_pollution_time(const CIVTERRAIN_TYPE *pter)
{
  return 3;
}

/***********************************************************************
  Convenience access method.
***********************************************************************/
int civterrain_type_fallout_time(const CIVTERRAIN_TYPE *pter)
{
  return 3;
}

#ifdef FC__UNIT_H
/***********************************************************************
  Time to complete the given unit activity on the given terrain.
***********************************************************************/
int civterrain_type_activity_time(const CIVTERRAIN_TYPE *pter,
  enum unit_activity activity)
{
  switch  {
  case ACTIVITY_ROAD:
    return pter->road_time;
  case ACTIVITY_RAILROAD:
    return 3;  /* return pter->railroad_time;      */
  case ACTIVITY_MINE:
    return pter->mine_time;
  case ACTIVITY_IRRIGATE:
    return pter->irrigation_time;
  case ACTIVITY_FORTRESS:
    return 3;  /* return pter->fortress_time;  */
  case ACTIVITY_TRANSFORM:
    return pter->transform_time;
  case ACTIVITY_AIRBASE:
    return 3;  /* return pter->airbase_time;   */
  case ACTIVITY_POLLUTION:
    return 3;  /* return pter->pollution_time; */
  case ACTIVITY_FALLOUT:
    return 3;  /* return pter->fallout_time;   */
  default:
    return 0;
  }
}
#endif

/***********************************************************************
  Hardcoded preferences for pillage selection.
***********************************************************************/
int get_preferred_pillage(int pset)
{
  /* pillage highest numbered special first, i.e. S_FARMLAND */
  static int pillage_order[] = {
    S_ROAD,
    S_RAILROAD,
    S_AIRBASE,
    S_FORTRESS,
    S_MINE,
    S_IRRIGATION,
    S_FARMLAND,
  };
  int nn = sizeof(pillage_order)/sizeof(pillage_order[0]);
  while (--nn >= 0) {
    if (pset & pillage_order[nn])
      return pillage_order[nn];
  }
  return S_NO_SPECIAL;
}

