/**********************************************************************
 Freeciv - Copyright (C) 1996 - A Kjeldberg, L Gregersen, P Unold
 civterrain.h - Copyright (C) 2002 - R. Wetmore

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
***********************************************************************/

#ifndef FC__CIVTERRAIN_H
#define FC__CIVTERRAIN_H

#ifndef N_
#define N_(String) String
#endif


/**
 *  civterrain.h - terrain and terrain_types definitions for CIVTILEs.
 *
 *  TODO:
 *    It would be *really* nice to merge terrain and specials into
 *  a single terrain type whose single index would be a complete
 *  description of the tile. But noting that basic terrain types
 *  are ordered by index value 0-15 or 4-bits of terrain, and
 *  specials are ordered by bit value, one recognizes that there
 *  are two types of values. One type like terrain is exclusive
 *  meaning there may only be one value per tile, while the second 
 *  types are non-exclusive meaning that any combination of values
 *  is possible.
 *    Thus one may have a T_DESERT or a T_HILLS tile, but not a
 *  hilly-desert tile. One may have a S_SPECIAL_1 or S_SPECIAL_2
 *  but not both specials in the same tile. However, one can have
 *  any mix of T_HILLS or T_DESERT plus S_SPECIAL_1 or S_SPECIAL_2
 *  plus S_IRRIGATION or S_FARMLAND or S_MINE, etc.
 *    This means there are sets of exclusive elements where only
 *  one of each set, but any combination of sets is a valid tile
 *  state.
 *
 *  f e d c b a 9 8 7 6 5 4 3 2 1 0                              Total
 *                          ^-----^ { T_TERRAIN }                   16
 *                        ^         none or S_RIVER                 32     
 *                    ^-^           none or S_SPECIAL_{1,2,3}      128
 *                ^-^               none or S_{MINE,IRRIG,FARM}    512
 *            ^-^                   none or S_{ROAD,RAIL,MAGLEV}  2048
 *          ^                       none or S_POLLUTION           4096
 *        ^                         none or S_FALLOUT             8192
 *                                                               -----
 *    ^-^                           none or S_{HUT,FORT,AIRBASE} 32768
 *
 *  Note: HUT and CITY are a matter of ownership if one wanted S_CITY.
 *
 *  There are thus 8192 distinct food/shield/trade values, and 32768
 *  distinct overlayed tile graphics possibilities.
 *
 *  But there are only 30 distinct names or individual sub-elements
 *  unfortunately neither fully linear nor bitwise orderable, but
 *  grouped into 8 non-exclusive sets of exclusive subelements.
 *
 *  At the moment, this is split into two major subsets, the 16 basic
 *  terrain types (linear), and 14 specials (bitwise) *not* grouped in
 *  ways that show the exclusion between subelements. Unrolling subsets
 *  does allow for efficient bitmask testing however. For now, one 
 *  uses the bit efficiencies fully, i.e. no partial conversion.
 */

/* 
 * Types of terrain (from terrain.h).
 * - suggested changes (order and value dependent)
 * - should T_UNKNOWN be a valid terrain type or "not-a-type" as -1.
 *   if it is a type, is it the "inaccessible" type, or black-tile
 *   used for unreal tiles within the normal set?
 *
 * Generalized access via civterrain.types[T_COUNT]
 */
#ifdef TODO
enum terrain_type {                               .   X_dry(2)
  T_UNKNOWN,    T_TUNDRA,     T_HILLS,      T_ARCTIC,     .  /
  T_SWAMP,      T_PLAINS,     T_FOREST,     T_MOUNTAINS,  . + - Y_high(2)
  T_RIVER,      T_GRASSLAND,  T_JUNGLE,     T_DESERT,     . |
  T_COASTAL,    T_OCEAN,      T_UNUSED0,    T_UNUSED1,    . Z_hot(4)
   ---  or ---                                            
  T_UNKNOWN,    T_OCEAN,      T_COASTAL,    T_SWAMP,      
  T_JUNGLE,     T_FOREST,     T_RIVER,      T_GRASSLAND,  . + - flat(4)
  T_MOUNTAINS,  T_HILLS,      T_DESERT,     T_PLAINS,     . |
  T_ARCTIC,     T_TUNDRA,     T_UNUSED0,    T_UNUSED1,    . cold(4)

  T_COUNT
};

static const char terrain_chars[16] = "uthaspfmrgjdc +x";
static const char terrain_chars[16] = "u csjfrgmhdpat+x";
#endif	/* TODO */

enum terrain_type {
  T_ARCTIC,     T_DESERT,     T_FOREST,     T_GRASSLAND,  
  T_HILLS,      T_JUNGLE,     T_MOUNTAINS,  T_OCEAN,      
  T_PLAINS,     T_RIVER,      T_SWAMP,      T_TUNDRA,     
  T_UNKNOWN,    T_UNUSED0,    T_UNUSED1,    T_UNUSED2,    
  T_COUNT
};
#define T_FIRST (T_ARCTIC)
#define T_LAST	(T_UNUSED0-1)

/* 
 * Matching char codes of terrain (from server/savegame.c).
 */
static const char terrain_chars[16] = "adfghjm prstu012";

enum terrain_river_type {
  R_AS_TERRAIN=1,
  R_AS_SPECIAL=2
};

enum special_river_move {
  RMV_NORMAL=0,
  RMV_FAST_STRICT=1,
  RMV_FAST_RELAXED=2,
  RMV_FAST_ALWAYS=3
};


/* 
 * Types of specials (from terrain.h).
 *
 * Generalized access via civterrain.stypes[S_COUNT]
 */
enum special_type {
  S_NO_SPECIAL_BIT = -1,  S_NO_SPECIAL =    0,
  S_SPECIAL_1_BIT  =  0,  S_SPECIAL_1  =    1,
  S_ROAD_BIT       =  1,  S_ROAD       =    2,
  S_IRRIGATION_BIT =  2,  S_IRRIGATION =    4,
  S_RAILROAD_BIT   =  3,  S_RAILROAD   =    8,
  S_MINE_BIT       =  4,  S_MINE       =   16,
  S_POLLUTION_BIT  =  5,  S_POLLUTION  =   32,
  S_HUT_BIT        =  6,  S_HUT        =   64,
  S_FORTRESS_BIT   =  7,  S_FORTRESS   =  128,
  S_SPECIAL_2_BIT  =  8,  S_SPECIAL_2  =  256,
  S_RIVER_BIT      =  9,  S_RIVER      =  512,
  S_FARMLAND_BIT   = 10,  S_FARMLAND   = 1024,
  S_AIRBASE_BIT    = 11,  S_AIRBASE    = 2048,
  S_FALLOUT_BIT    = 12,  S_FALLOUT    = 4096,
  S_COUNT          = 13
};

#define S_ALL      \
 (  S_SPECIAL_1    \
  | S_ROAD         \
  | S_IRRIGATION   \
  | S_RAILROAD     \
  | S_MINE         \
  | S_POLLUTION    \
  | S_HUT          \
  | S_FORTRESS     \
  | S_SPECIAL_2    \
  | S_RIVER        \
  | S_FARMLAND     \
  | S_AIRBASE      \
  | S_FALLOUT      \
 )

#define S_INFRASTRUCTURE_MASK  \
 (  S_ROAD         \
  | S_RAILROAD     \
  | S_IRRIGATION   \
  | S_FARMLAND     \
  | S_MINE         \
  | S_FORTRESS     \
  | S_AIRBASE      \
 )


/*
 * Complete generalized/ruleset definitions for each special type 
 */
typedef struct civterrain_special
{
  enum special_type spe;
  enum special_type prereq;
  char name[MAX_LEN_NAME];
} CIVTERRAIN_SPECIAL;

/**
 * Hardcoded defaults for the specials above.
 * 
 * Names of specials (from map.c).
 * (These must correspond to enum special_type bit index)
 */
static const CIVTERRAIN_SPECIAL CIVTERRAIN_SPECIAL_DEFAULT[] =
{
  { S_SPECIAL_1,  S_NO_SPECIAL, { N_("Special1") }}, 
  { S_ROAD,       S_NO_SPECIAL, { N_("Road") }}, 
  { S_IRRIGATION, S_NO_SPECIAL, { N_("Irrigation") }}, 
  { S_RAILROAD,   S_ROAD,       { N_("Railroad") }}, 
  { S_MINE,       S_NO_SPECIAL, { N_("Mine") }}, 
  { S_POLLUTION,  S_NO_SPECIAL, { N_("Pollution") }}, 
  { S_HUT,        S_NO_SPECIAL, { N_("Hut") }}, 
  { S_FORTRESS,   S_NO_SPECIAL, { N_("Fortress") }}, 
  { S_SPECIAL_2,  S_NO_SPECIAL, { N_("Special2") }}, 
  { S_RIVER,      S_NO_SPECIAL, { N_("River") }}, 
  { S_FARMLAND,   S_IRRIGATION, { N_("Farmland") }}, 
  { S_AIRBASE,    S_NO_SPECIAL, { N_("Airbase") }}, 
  { S_FALLOUT,    S_NO_SPECIAL, { N_("Fallout") }}, 
};


/*
 * Complete ruleset definitions for each basic terrain type (from map.h)
 */
typedef struct civterrain CIVTERRAIN;
typedef struct civterrain_type CIVTERRAIN_TYPE;

extern const CIVTERRAIN CIVTERRAIN_DEFAULT;
extern const CIVTERRAIN_TYPE CIVTERRAIN_TYPE_DEFAULT[T_COUNT];


/* TODO: increase to 6 for MAP_TYPE_HEX when true hex graphics added */
#ifndef	MAX_TILE_EDGES
  #ifndef MAX_NUM_NSEW
    #define	MAX_TILE_EDGES	4
  #else
    #define	MAX_TILE_EDGES	MAX_NUM_NSEW
  #endif
#endif

struct civterrain_type
{
  /* Basic terrain parameters */
  char name[MAX_LEN_NAME];              /* "" if unused */
  char name_orig[MAX_LEN_NAME];	        /* untranslated copy */
  char graphic_str[MAX_LEN_NAME];
  char graphic_alt[MAX_LEN_NAME];
  union {
    struct Sprite *sprites[MAX_TILE_EDGES];
    struct Sprite *sprite[MAX_TILE_EDGES];
  };
  int food;
  int shield;
  int trade;

  /* Special parameters - basic overrides if special flags are set */
  struct {
    char name[MAX_LEN_NAME];            /* "" if none */
    char name_orig[MAX_LEN_NAME];       /* untranslated copy */
    char graphic_str[MAX_LEN_NAME];
    char graphic_alt[MAX_LEN_NAME];
    union {
      struct Sprite *sprites[MAX_TILE_EDGES]; /* fill (symmetry + future) */
      struct Sprite *sprite;
    };
    int food;
    int shield;
    int trade;
  } special[2];

  /* TODO: put this special data in the above struct too, --dwp */
  /* Done but with temporary backwards compatibility fixes ...  */
  #define special_1_name        special[1].name
  #define special_1_name_orig   special[1].name_orig
  #define food_special_1        special[1].food
  #define shield_special_1      special[1].shield
  #define trade_special_1       special[1].trade

  #define special_2_name        special[2].name
  #define special_2_name_orig   special[2].name_orig
  #define food_special_2        special[2].food
  #define shield_special_2      special[2].shield
  #define trade_special_2       special[2].trade

  /* Common parameters, as opposed to basic/special specific. */

  char *helptext;
  const CIVTERRAIN *global;       /* backlink to global header */

  int movement_cost;
  int defense_bonus;

  /* TODO: symmetrize the next 4 as part of generalized activities */
  enum terrain_type base;
  int road_trade_incr;
  int road_time;

  enum terrain_type irrigation_result;
  int irrigation_food_incr;
  int irrigation_time;

  enum terrain_type mining_result;
  int mining_shield_incr;
  int mining_time;

  enum terrain_type transform_result;
  int transform_pollution_incr;
  int transform_time;
};

/**
 * Hardcoded defaults for the 16 basic tile types ordered as above.
 */

#ifdef FC__CIVTERRAIN_C
const CIVTERRAIN_TYPE CIVTERRAIN_TYPE_DEFAULT[T_COUNT] = \
{
 {{"Glacier"},          /* char name[MAX_LEN_NAME];                  */
  {"Glacier"},          /* char name_orig[MAX_LEN_NAME];             */
  {"t.arctic"},         /* char graphic_str[MAX_LEN_NAME];           */
  {"-"},                /* char graphic_alt[MAX_LEN_NAME];           */
  { {NULL} },           /* struct Sprite *sprite[4];                 */
  0,                    /* int food;                                 */
  0,                    /* int shield;                               */
  0,                    /* int trade;                                */

 {{                     /* special1                                  */
  {"Ivory"},            /*   char name[MAX_LEN_NAME];                */
  {"Ivory"},            /*   char name_orig[MAX_LEN_NAME];           */
  {"ts.arctic_ivory"},  /*   char graphic_str[MAX_LEN_NAME];         */
  {"-"},                /*   char graphic_alt[MAX_LEN_NAME];         */
  { {NULL} },           /*   struct Sprite *sprite[4];               */
  1,                    /*   int food;                               */
  1,                    /*   int shield;                             */
  4,                    /*   int trade;                              */
 },
 {                      /* special2                                  */
  {"Oil"},              /*   char name[MAX_LEN_NAME];                */
  {"Oil"},              /*   char name_orig[MAX_LEN_NAME];           */
  {"ts.arctic_oil"},    /*   char graphic_str[MAX_LEN_NAME];         */
  {"ts.oil"},           /*   char graphic_alt[MAX_LEN_NAME];         */
  { {NULL} },           /*   struct Sprite *sprite[4];               */
  0,                    /*   int food;                               */
  4,                    /*   int shield;                             */
  0,                    /*   int trade;                              */
 }},
  NULL,                 /* char *helptext;                           */
  &CIVTERRAIN_DEFAULT,  /* CIVTERRAIN *global;                       */

  2,                    /* int movement_cost;                        */
  10,                   /* int defense_bonus;                        */

  T_ARCTIC,             /* enum terrain_type base;                   */
  0,                    /* int road_trade_incr;                      */
  4,                    /* int road_time;                            */

  T_LAST,               /* enum terrain_type irrigation_result;      */
  0,                    /* int irrigation_food_incr;                 */
  0,                    /* int irrigation_time;                      */

  T_ARCTIC,             /* enum terrain_type mining_result;          */
  1,                    /* int mining_shield_incr;                   */
  10,                   /* int mining_time;                          */

  T_TUNDRA,             /* enum terrain_type transform_result;       */
  0,                    /* int transform_pollution_incr;             */
  24,                   /* int transform_time;                       */
 },
 {{"Desert"},           /* char name[MAX_LEN_NAME];                  */
  {"Desert"},           /* char name_orig[MAX_LEN_NAME];             */
  {"t.desert"},         /* char graphic_str[MAX_LEN_NAME];           */
  {"-"},                /* char graphic_alt[MAX_LEN_NAME];           */
  { {NULL} },           /* struct Sprite *sprite[4];                 */
  0,                    /* int food;                                 */
  1,                    /* int shield;                               */
  0,                    /* int trade;                                */

 {{                     /* special1                                  */
  {"Oasis"},            /*   char name[MAX_LEN_NAME];                */
  {"Oasis"},            /*   char name_orig[MAX_LEN_NAME];           */
  {"ts.oasis"},         /*   char graphic_str[MAX_LEN_NAME];         */
  {"-"},                /*   char graphic_alt[MAX_LEN_NAME];         */
  { {NULL} },           /*   struct Sprite *sprite[4];               */
  3,                    /*   int food;                               */
  1,                    /*   int shield;                             */
  0,                    /*   int trade;                              */
 },
 {                      /* special2                                  */
  {"Oil"},              /*   char name[MAX_LEN_NAME];                */
  {"Oil"},              /*   char name_orig[MAX_LEN_NAME];           */
  {"ts.oil"},           /*   char graphic_str[MAX_LEN_NAME];         */
  {"-"},                /*   char graphic_alt[MAX_LEN_NAME];         */
  { {NULL} },           /*   struct Sprite *sprite[4];               */
  0,                    /*   int food;                               */
  4,                    /*   int shield;                             */
  0,                    /*   int trade;                              */
 }},
  NULL,                 /* char *helptext;                           */
  &CIVTERRAIN_DEFAULT,  /* CIVTERRAIN *global;                       */

  1,                    /* int movement_cost;                        */
  10,                   /* int defense_bonus;                        */

  T_DESERT,             /* enum terrain_type base;                   */
  1,                    /* int road_trade_incr;                      */
  2,                    /* int road_time;                            */

  T_DESERT,             /* enum terrain_type irrigation_result;      */
  1,                    /* int irrigation_food_incr;                 */
  5,                    /* int irrigation_time;                      */

  T_DESERT,             /* enum terrain_type mining_result;          */
  1,                    /* int mining_shield_incr;                   */
  10,                   /* int mining_time;                          */

  T_PLAINS,             /* enum terrain_type transform_result;       */
  0,                    /* int transform_pollution_incr;             */
  24,                   /* int transform_time;                       */
 },
 {{"Forest"},           /* char name[MAX_LEN_NAME];                  */
  {"Forest"},           /* char name_orig[MAX_LEN_NAME];             */
  {"t.forest"},         /* char graphic_str[MAX_LEN_NAME];           */
  {"-"},                /* char graphic_alt[MAX_LEN_NAME];           */
  { {NULL} },           /* struct Sprite *sprite[4];                 */
  1,                    /* int food;                                 */
  2,                    /* int shield;                               */
  0,                    /* int trade;                                */

 {{                     /* special1                                  */
  {"Pheasant"},         /*   char name[MAX_LEN_NAME];                */
  {"Pheasant"},         /*   char name_orig[MAX_LEN_NAME];           */
  {"ts.pheasant"},      /*   char graphic_str[MAX_LEN_NAME];         */
  {"-"},                /*   char graphic_alt[MAX_LEN_NAME];         */
  { {NULL} },           /*   struct Sprite *sprite[4];               */
  3,                    /*   int food;                               */
  2,                    /*   int shield;                             */
  0,                    /*   int trade;                              */
 },
 {                      /* special2                                  */
  {"Silk"},             /*   char name[MAX_LEN_NAME];                */
  {"Silk"},             /*   char name_orig[MAX_LEN_NAME];           */
  {"ts.silk"},          /*   char graphic_str[MAX_LEN_NAME];         */
  {"-"},                /*   char graphic_alt[MAX_LEN_NAME];         */
  { {NULL} },           /*   struct Sprite *sprite[4];               */
  1,                    /*   int food;                               */
  2,                    /*   int shield;                             */
  3,                    /*   int trade;                              */
 }},
  NULL,                 /* char *helptext;                           */
  &CIVTERRAIN_DEFAULT,  /* CIVTERRAIN *global;                       */

  2,                    /* int movement_cost;                        */
  15,                   /* int defense_bonus;                        */

  T_FOREST,             /* enum terrain_type base;                   */
  0,                    /* int road_trade_incr;                      */
  4,                    /* int road_time;                            */

  T_PLAINS,             /* enum terrain_type irrigation_result;      */
  0,                    /* int irrigation_food_incr;                 */
  6,                    /* int irrigation_time;                      */

  T_LAST,               /* enum terrain_type mining_result;          */
  0,                    /* int mining_shield_incr;                   */
  0,                    /* int mining_time;                          */

  T_GRASSLAND,          /* enum terrain_type transform_result;       */
  0,                    /* int transform_pollution_incr;             */
  24,                   /* int transform_time;                       */
 },
 {{"Grassland"},        /* char name[MAX_LEN_NAME];                  */
  {"Grassland"},        /* char name_orig[MAX_LEN_NAME];             */
  {"t.grassland"},      /* char graphic_str[MAX_LEN_NAME];           */
  {"-"},                /* char graphic_alt[MAX_LEN_NAME];           */
  { {NULL} },           /* struct Sprite *sprite[4];                 */
  2,                    /* int food;                                 */
  0,                    /* int shield;                               */
  0,                    /* int trade;                                */

 {{                     /* special1                                  */
  {"Resources"},        /*   char name[MAX_LEN_NAME];                */
  {"Resources"},        /*   char name_orig[MAX_LEN_NAME];           */
  {"ts.grassland_resources"},  /*   char graphic_str[MAX_LEN_NAME];  */
  {"-"},                /*   char graphic_alt[MAX_LEN_NAME];         */
  { {NULL} },           /*   struct Sprite *sprite[4];               */
  2,                    /*   int food;                               */
  1,                    /*   int shield;                             */
  0,                    /*   int trade;                              */
 },
 {                      /* special2                                  */
  {"Resources"},        /*   char name[MAX_LEN_NAME];                */
  {"Resources"},        /*   char name_orig[MAX_LEN_NAME];           */
  {"ts.grassland_resources"},  /*   char graphic_str[MAX_LEN_NAME];  */
  {"-"},                /*   char graphic_alt[MAX_LEN_NAME];         */
  { {NULL} },           /*   struct Sprite *sprite[4];               */
  2,                    /*   int food;                               */
  1,                    /*   int shield;                             */
  0,                    /*   int trade;                              */
 }},
  NULL,                 /* char *helptext;                           */
  &CIVTERRAIN_DEFAULT,  /* CIVTERRAIN *global;                       */

  1,                    /* int movement_cost;                        */
  10,                   /* int defense_bonus;                        */

  T_GRASSLAND,          /* enum terrain_type base;                   */
  1,                    /* int road_trade_incr;                      */
  2,                    /* int road_time;                            */

  T_GRASSLAND,          /* enum terrain_type irrigation_result;      */
  1,                    /* int irrigation_food_incr;                 */
  6,                    /* int irrigation_time;                      */

  T_FOREST,             /* enum terrain_type mining_result;          */
  0,                    /* int mining_shield_incr;                   */
  10,                   /* int mining_time;                          */

  T_HILLS,              /* enum terrain_type transform_result;       */
  0,                    /* int transform_pollution_incr;             */
  24,                   /* int transform_time;                       */
 },
 {{"Hills"},            /* char name[MAX_LEN_NAME];                  */
  {"Hills"},            /* char name_orig[MAX_LEN_NAME];             */
  {"t.hills"},          /* char graphic_str[MAX_LEN_NAME];           */
  {"-"},                /* char graphic_alt[MAX_LEN_NAME];           */
  { {NULL} },           /* struct Sprite *sprite[4];                 */
  1,                    /* int food;                                 */
  0,                    /* int shield;                               */
  0,                    /* int trade;                                */

 {{                     /* special1                                  */
  {"Coal"},             /*   char name[MAX_LEN_NAME];                */
  {"Coal"},             /*   char name_orig[MAX_LEN_NAME];           */
  {"ts.coal"},          /*   char graphic_str[MAX_LEN_NAME];         */
  {"-"},                /*   char graphic_alt[MAX_LEN_NAME];         */
  { {NULL} },           /*   struct Sprite *sprite[4];               */
  1,                    /*   int food;                               */
  3,                    /*   int shield;                             */
  8,                    /*   int trade;                              */
 },
 {                      /* special2                                  */
  {"Wine"},             /*   char name[MAX_LEN_NAME];                */
  {"Wine"},             /*   char name_orig[MAX_LEN_NAME];           */
  {"ts.Wine"},          /*   char graphic_str[MAX_LEN_NAME];         */
  {"-"},                /*   char graphic_alt[MAX_LEN_NAME];         */
  { {NULL} },           /*   struct Sprite *sprite[4];               */
  0,                    /*   int food;                               */
  4,                    /*   int shield;                             */
  0,                    /*   int trade;                              */
 }},
  NULL,                 /* char *helptext;                           */
  &CIVTERRAIN_DEFAULT,  /* CIVTERRAIN *global;                       */

  2,                    /* int movement_cost;                        */
  20,                   /* int defense_bonus;                        */

  T_HILLS,              /* enum terrain_type base;                   */
  0,                    /* int road_trade_incr;                      */
  4,                    /* int road_time;                            */

  T_HILLS,              /* enum terrain_type irrigation_result;      */
  1,                    /* int irrigation_food_incr;                 */
  8,                    /* int irrigation_time;                      */

  T_HILLS,              /* enum terrain_type mining_result;          */
  3,                    /* int mining_shield_incr;                   */
  8,                    /* int mining_time;                          */

  T_PLAINS,             /* enum terrain_type transform_result;       */
  0,                    /* int transform_pollution_incr;             */
  24,                   /* int transform_time;                       */
 },
 {{"Jungle"},           /* char name[MAX_LEN_NAME];                  */
  {"Jungle"},           /* char name_orig[MAX_LEN_NAME];             */
  {"t.jungle"},         /* char graphic_str[MAX_LEN_NAME];           */
  {"-"},                /* char graphic_alt[MAX_LEN_NAME];           */
  { {NULL} },           /* struct Sprite *sprite[4];                 */
  0,                    /* int food;                                 */
  1,                    /* int shield;                               */
  0,                    /* int trade;                                */

 {{                     /* special1                                  */
  {"Gems"},             /*   char name[MAX_LEN_NAME];                */
  {"Gems"},             /*   char name_orig[MAX_LEN_NAME];           */
  {"ts.Gems"},          /*   char graphic_str[MAX_LEN_NAME];         */
  {"-"},                /*   char graphic_alt[MAX_LEN_NAME];         */
  { {NULL} },           /*   struct Sprite *sprite[4];               */
  0,                    /*   int food;                               */
  1,                    /*   int shield;                             */
  4,                    /*   int trade;                              */
 },
 {                      /* special2                                  */
  {"Fruit"},            /*   char name[MAX_LEN_NAME];                */
  {"Fruit"},            /*   char name_orig[MAX_LEN_NAME];           */
  {"ts.fruit"},         /*   char graphic_str[MAX_LEN_NAME];         */
  {"-"},                /*   char graphic_alt[MAX_LEN_NAME];         */
  { {NULL} },           /*   struct Sprite *sprite[4];               */
  4,                    /*   int food;                               */
  0,                    /*   int shield;                             */
  1,                    /*   int trade;                              */
 }},
  NULL,                 /* char *helptext;                           */
  &CIVTERRAIN_DEFAULT,  /* CIVTERRAIN *global;                       */

  2,                    /* int movement_cost;                        */
  15,                   /* int defense_bonus;                        */

  T_JUNGLE,             /* enum terrain_type base;                   */
  0,                    /* int road_trade_incr;                      */
  4,                    /* int road_time;                            */

  T_SWAMP,              /* enum terrain_type irrigation_result;      */
  0,                    /* int irrigation_food_incr;                 */
  5,                    /* int irrigation_time;                      */

  T_FOREST,             /* enum terrain_type mining_result;          */
  0,                    /* int mining_shield_incr;                   */
  10,                   /* int mining_time;                          */

  T_PLAINS,             /* enum terrain_type transform_result;       */
  0,                    /* int transform_pollution_incr;             */
  24,                   /* int transform_time;                       */
 },
 {{"Mountains"},        /* char name[MAX_LEN_NAME];                  */
  {"Mountains"},        /* char name_orig[MAX_LEN_NAME];             */
  {"t.mountains"},      /* char graphic_str[MAX_LEN_NAME];           */
  {"-"},                /* char graphic_alt[MAX_LEN_NAME];           */
  { {NULL} },           /* struct Sprite *sprite[4];                 */
  0,                    /* int food;                                 */
  1,                    /* int shield;                               */
  0,                    /* int trade;                                */

 {{                     /* special1                                  */
  {"Gold"},             /*   char name[MAX_LEN_NAME];                */
  {"Gold"},             /*   char name_orig[MAX_LEN_NAME];           */
  {"ts.gold"},          /*   char graphic_str[MAX_LEN_NAME];         */
  {"-"},                /*   char graphic_alt[MAX_LEN_NAME];         */
  { {NULL} },           /*   struct Sprite *sprite[4];               */
  0,                    /*   int food;                               */
  1,                    /*   int shield;                             */
  6,                    /*   int trade;                              */
 },
 {                      /* special2                                  */
  {"Iron"},             /*   char name[MAX_LEN_NAME];                */
  {"Iron"},             /*   char name_orig[MAX_LEN_NAME];           */
  {"ts.iron"},          /*   char graphic_str[MAX_LEN_NAME];         */
  {"-"},                /*   char graphic_alt[MAX_LEN_NAME];         */
  { {NULL} },           /*   struct Sprite *sprite[4];               */
  0,                    /*   int food;                               */
  4,                    /*   int shield;                             */
  0,                    /*   int trade;                              */
 }},
  NULL,                 /* char *helptext;                           */
  &CIVTERRAIN_DEFAULT,  /* CIVTERRAIN *global;                       */

  3,                    /* int movement_cost;                        */
  30,                   /* int defense_bonus;                        */

  T_MOUNTAINS,          /* enum terrain_type base;                   */
  0,                    /* int road_trade_incr;                      */
  6,                    /* int road_time;                            */

  T_LAST,               /* enum terrain_type irrigation_result;      */
  0,                    /* int irrigation_food_incr;                 */
  0,                    /* int irrigation_time;                      */

  T_MOUNTAINS,          /* enum terrain_type mining_result;          */
  1,                    /* int mining_shield_incr;                   */
  10,                   /* int mining_time;                          */

  T_HILLS,              /* enum terrain_type transform_result;       */
  0,                    /* int transform_pollution_incr;             */
  24,                   /* int transform_time;                       */
 },
 {{"Ocean"},            /* char name[MAX_LEN_NAME];                  */
  {"Ocean"},            /* char name_orig[MAX_LEN_NAME];             */
  {"t.ocean"},          /* char graphic_str[MAX_LEN_NAME];           */
  {"-"},                /* char graphic_alt[MAX_LEN_NAME];           */
  { {NULL} },           /* struct Sprite *sprite[4];                 */
  1,                    /* int food;                                 */
  0,                    /* int shield;                               */
  2,                    /* int trade;                                */

 {{                     /* special1                                  */
  {"Fish"},             /*   char name[MAX_LEN_NAME];                */
  {"Fish"},             /*   char name_orig[MAX_LEN_NAME];           */
  {"ts.fish"},          /*   char graphic_str[MAX_LEN_NAME];         */
  {"-"},                /*   char graphic_alt[MAX_LEN_NAME];         */
  { {NULL} },           /*   struct Sprite *sprite[4];               */
  3,                    /*   int food;                               */
  0,                    /*   int shield;                             */
  2,                    /*   int trade;                              */
 },
 {                      /* special2                                  */
  {"Whales"},           /*   char name[MAX_LEN_NAME];                */
  {"Whales"},           /*   char name_orig[MAX_LEN_NAME];           */
  {"ts.whales"},        /*   char graphic_str[MAX_LEN_NAME];         */
  {"-"},                /*   char graphic_alt[MAX_LEN_NAME];         */
  { {NULL} },           /*   struct Sprite *sprite[4];               */
  2,                    /*   int food;                               */
  2,                    /*   int shield;                             */
  3,                    /*   int trade;                              */
 }},
  NULL,                 /* char *helptext;                           */
  &CIVTERRAIN_DEFAULT,  /* CIVTERRAIN *global;                       */

  1,                    /* int movement_cost;                        */
  10,                   /* int defense_bonus;                        */

  T_OCEAN,              /* enum terrain_type base;                   */
  0,                    /* int road_trade_incr;                      */
  0,                    /* int road_time;                            */

  T_LAST,               /* enum terrain_type irrigation_result;      */
  0,                    /* int irrigation_food_incr;                 */
  0,                    /* int irrigation_time;                      */

  T_LAST,               /* enum terrain_type mining_result;          */
  0,                    /* int mining_shield_incr;                   */
  0,                    /* int mining_time;                          */

  T_MOUNTAINS,          /* enum terrain_type transform_result;       */
  0,                    /* int transform_pollution_incr;             */
  24,                   /* int transform_time;                       */
 },
 {{"Plains"},           /* char name[MAX_LEN_NAME];                  */
  {"Plains"},           /* char name_orig[MAX_LEN_NAME];             */
  {"t.plains"},         /* char graphic_str[MAX_LEN_NAME];           */
  {"-"},                /* char graphic_alt[MAX_LEN_NAME];           */
  { {NULL} },           /* struct Sprite *sprite[4];                 */
  1,                    /* int food;                                 */
  1,                    /* int shield;                               */
  0,                    /* int trade;                                */

 {{                     /* special1                                  */
  {"Buffalo"},          /*   char name[MAX_LEN_NAME];                */
  {"Buffalo"},          /*   char name_orig[MAX_LEN_NAME];           */
  {"ts.buffalo"},       /*   char graphic_str[MAX_LEN_NAME];         */
  {"-"},                /*   char graphic_alt[MAX_LEN_NAME];         */
  { {NULL} },           /*   struct Sprite *sprite[4];               */
  1,                    /*   int food;                               */
  3,                    /*   int shield;                             */
  0,                    /*   int trade;                              */
 },
 {                      /* special2                                  */
  {"Wheat"},            /*   char name[MAX_LEN_NAME];                */
  {"Wheat"},            /*   char name_orig[MAX_LEN_NAME];           */
  {"ts.wheat"},         /*   char graphic_str[MAX_LEN_NAME];         */
  {"-"},                /*   char graphic_alt[MAX_LEN_NAME];         */
  { {NULL} },           /*   struct Sprite *sprite[4];               */
  3,                    /*   int food;                               */
  1,                    /*   int shield;                             */
  0,                    /*   int trade;                              */
 }},
  NULL,                 /* char *helptext;                           */
  &CIVTERRAIN_DEFAULT,  /* CIVTERRAIN *global;                       */

  1,                    /* int movement_cost;                        */
  10,                   /* int defense_bonus;                        */

  T_PLAINS,             /* enum terrain_type base;                   */
  1,                    /* int road_trade_incr;                      */
  2,                    /* int road_time;                            */

  T_PLAINS,             /* enum terrain_type irrigation_result;      */
  1,                    /* int irrigation_food_incr;                 */
  5,                    /* int irrigation_time;                      */

  T_FOREST,             /* enum terrain_type mining_result;          */
  0,                    /* int mining_shield_incr;                   */
  15,                   /* int mining_time;                          */

  T_GRASSLAND,          /* enum terrain_type transform_result;       */
  0,                    /* int transform_pollution_incr;             */
  24,                   /* int transform_time;                       */
 },
 {{"River"},            /* char name[MAX_LEN_NAME];                  */
  {"River"},            /* char name_orig[MAX_LEN_NAME];             */
  {"t.river"},          /* char graphic_str[MAX_LEN_NAME];           */
  {"-"},                /* char graphic_alt[MAX_LEN_NAME];           */
  { {NULL} },           /* struct Sprite *sprite[4];                 */
  2,                    /* int food;                                 */
  0,                    /* int shield;                               */
  1,                    /* int trade;                                */

 {{                     /* special1                                  */
  {"Resources"},        /*   char name[MAX_LEN_NAME];                */
  {"Resources"},        /*   char name_orig[MAX_LEN_NAME];           */
  {"ts.grassland_resources"},  /*   char graphic_str[MAX_LEN_NAME];  */
  {"-"},                /*   char graphic_alt[MAX_LEN_NAME];         */
  { {NULL} },           /*   struct Sprite *sprite[4];               */
  2,                    /*   int food;                               */
  1,                    /*   int shield;                             */
  0,                    /*   int trade;                              */
 },
 {                      /* special2                                  */
  {"Resources"},        /*   char name[MAX_LEN_NAME];                */
  {"Resources"},        /*   char name_orig[MAX_LEN_NAME];           */
  {"ts.grassland_resources"},  /*   char graphic_str[MAX_LEN_NAME];  */
  {"-"},                /*   char graphic_alt[MAX_LEN_NAME];         */
  { {NULL} },           /*   struct Sprite *sprite[4];               */
  2,                    /*   int food;                               */
  1,                    /*   int shield;                             */
  0,                    /*   int trade;                              */
 }},
  NULL,                 /* char *helptext;                           */
  &CIVTERRAIN_DEFAULT,  /* CIVTERRAIN *global;                       */

  1,                    /* int movement_cost;                        */
  10,                   /* int defense_bonus;                        */

  T_RIVER,              /* enum terrain_type base;                   */
  1,                    /* int road_trade_incr;                      */
  2,                    /* int road_time;                            */

  T_RIVER,              /* enum terrain_type irrigation_result;      */
  1,                    /* int irrigation_food_incr;                 */
  5,                    /* int irrigation_time;                      */

  T_FOREST,             /* enum terrain_type mining_result;          */
  0,                    /* int mining_shield_incr;                   */
  10,                   /* int mining_time;                          */

  T_OCEAN,              /* enum terrain_type transform_result;       */
  0,                    /* int transform_pollution_incr;             */
  24,                   /* int transform_time;                       */
 },
 {{"Swamp"},            /* char name[MAX_LEN_NAME];                  */
  {"Swamp"},            /* char name_orig[MAX_LEN_NAME];             */
  {"t.swamp"},          /* char graphic_str[MAX_LEN_NAME];           */
  {"-"},                /* char graphic_alt[MAX_LEN_NAME];           */
  { {NULL} },           /* struct Sprite *sprite[4];                 */
  1,                    /* int food;                                 */
  0,                    /* int shield;                               */
  0,                    /* int trade;                                */

 {{                     /* special1                                  */
  {"Peat"},             /*   char name[MAX_LEN_NAME];                */
  {"Peat"},             /*   char name_orig[MAX_LEN_NAME];           */
  {"ts.peat"},          /*   char graphic_str[MAX_LEN_NAME];         */
  {"-"},                /*   char graphic_alt[MAX_LEN_NAME];         */
  { {NULL} },           /*   struct Sprite *sprite[4];               */
  1,                    /*   int food;                               */
  4,                    /*   int shield;                             */
  0,                    /*   int trade;                              */
 },
 {                      /* special2                                  */
  {"Spice"},            /*   char name[MAX_LEN_NAME];                */
  {"Spice"},            /*   char name_orig[MAX_LEN_NAME];           */
  {"ts.Spice"},         /*   char graphic_str[MAX_LEN_NAME];         */
  {"-"},                /*   char graphic_alt[MAX_LEN_NAME];         */
  { {NULL} },           /*   struct Sprite *sprite[4];               */
  3,                    /*   int food;                               */
  0,                    /*   int shield;                             */
  4,                    /*   int trade;                              */
 }},
  NULL,                 /* char *helptext;                           */
  &CIVTERRAIN_DEFAULT,  /* CIVTERRAIN *global;                       */

  2,                    /* int movement_cost;                        */
  15,                   /* int defense_bonus;                        */

  T_SWAMP,              /* enum terrain_type base;                   */
  0,                    /* int road_trade_incr;                      */
  4,                    /* int road_time;                            */

  T_GRASSLAND,          /* enum terrain_type irrigation_result;      */
  0,                    /* int irrigation_food_incr;                 */
  10,                   /* int irrigation_time;                      */

  T_JUNGLE,             /* enum terrain_type mining_result;          */
  0,                    /* int mining_shield_incr;                   */
  5,                    /* int mining_time;                          */

  T_OCEAN,              /* enum terrain_type transform_result;       */
  0,                    /* int transform_pollution_incr;             */
  24,                   /* int transform_time;                       */
 },
 {{"Tundra"},           /* char name[MAX_LEN_NAME];                  */
  {"Tundra"},           /* char name_orig[MAX_LEN_NAME];             */
  {"t.tundra"},         /* char graphic_str[MAX_LEN_NAME];           */
  {"-"},                /* char graphic_alt[MAX_LEN_NAME];           */
  { {NULL} },           /* struct Sprite *sprite[4];                 */
  1,                    /* int food;                                 */
  0,                    /* int shield;                               */
  0,                    /* int trade;                                */

 {{                     /* special1                                  */
  {"Game"},             /*   char name[MAX_LEN_NAME];                */
  {"Game"},             /*   char name_orig[MAX_LEN_NAME];           */
  {"ts.game"},          /*   char graphic_str[MAX_LEN_NAME];         */
  {"-"},                /*   char graphic_alt[MAX_LEN_NAME];         */
  { {NULL} },           /*   struct Sprite *sprite[4];               */
  3,                    /*   int food;                               */
  1,                    /*   int shield;                             */
  0,                    /*   int trade;                              */
 },
 {                      /* special2                                  */
  {"Furs"},             /*   char name[MAX_LEN_NAME];                */
  {"Furs"},             /*   char name_orig[MAX_LEN_NAME];           */
  {"ts.furs"},          /*   char graphic_str[MAX_LEN_NAME];         */
  {"-"},                /*   char graphic_alt[MAX_LEN_NAME];         */
  { {NULL} },           /*   struct Sprite *sprite[4];               */
  2,                    /*   int food;                               */
  0,                    /*   int shield;                             */
  3,                    /*   int trade;                              */
 }},
  NULL,                 /* char *helptext;                           */
  &CIVTERRAIN_DEFAULT,  /* CIVTERRAIN *global;                       */

  1,                    /* int movement_cost;                        */
  10,                   /* int defense_bonus;                        */

  T_TUNDRA,             /* enum terrain_type base;                   */
  0,                    /* int road_trade_incr;                      */
  2,                    /* int road_time;                            */

  T_TUNDRA,             /* enum terrain_type irrigation_result;      */
  1,                    /* int irrigation_food_incr;                 */
  5,                    /* int irrigation_time;                      */

  T_LAST,               /* enum terrain_type mining_result;          */
  0,                    /* int mining_shield_incr;                   */
  0,                    /* int mining_time;                          */

  T_DESERT,             /* enum terrain_type transform_result;       */
  0,                    /* int transform_pollution_incr;             */
  24,                   /* int transform_time;                       */
 },
 {{"Unknown"},          /* char name[MAX_LEN_NAME];                  */
  {"Unknown"},          /* char name_orig[MAX_LEN_NAME];             */
  {"t.unknown"},        /* char graphic_str[MAX_LEN_NAME];           */
  {"-"},                /* char graphic_alt[MAX_LEN_NAME];           */
  { {NULL} },           /* struct Sprite *sprite[4];                 */
  0,                    /* int food;                                 */
  0,                    /* int shield;                               */
  0,                    /* int trade;                                */

 {{                     /* special1                                  */
  {"Resources"},        /*   char name[MAX_LEN_NAME];                */
  {"Resources"},        /*   char name_orig[MAX_LEN_NAME];           */
  {"ts.grassland_resources"},  /*   char graphic_str[MAX_LEN_NAME];  */
  {"-"},                /*   char graphic_alt[MAX_LEN_NAME];         */
  { {NULL} },           /*   struct Sprite *sprite[4];               */
  0,                    /*   int food;                               */
  0,                    /*   int shield;                             */
  0,                    /*   int trade;                              */
 },
 {                      /* special2                                  */
  {"Resources"},        /*   char name[MAX_LEN_NAME];                */
  {"Resources"},        /*   char name_orig[MAX_LEN_NAME];           */
  {"ts.grassland_resources"},  /*   char graphic_str[MAX_LEN_NAME];  */
  {"-"},                /*   char graphic_alt[MAX_LEN_NAME];         */
  { {NULL} },           /*   struct Sprite *sprite[4];               */
  0,                    /*   int food;                               */
  0,                    /*   int shield;                             */
  0,                    /*   int trade;                              */
 }},
  NULL,                 /* char *helptext;                           */
  &CIVTERRAIN_DEFAULT,  /* CIVTERRAIN *global;                       */

  255,                  /* int movement_cost;                        */
  10,                   /* int defense_bonus;                        */

  T_UNKNOWN,            /* enum terrain_type base;                   */
  0,                    /* int road_trade_incr;                      */
  0,                    /* int road_time;                            */

  T_LAST,               /* enum terrain_type irrigation_result;      */
  0,                    /* int irrigation_food_incr;                 */
  0,                    /* int irrigation_time;                      */

  T_LAST,               /* enum terrain_type mining_result;          */
  0,                    /* int mining_shield_incr;                   */
  0,                    /* int mining_time;                          */

  T_LAST,               /* enum terrain_type transform_result;       */
  0,                    /* int transform_pollution_incr;             */
  24,                   /* int transform_time;                       */
 },
 /* TODO: rename Ocean <--> Coastal and use this for Ocean           */
 {{"Coastal"},          /* char name[MAX_LEN_NAME];                  */
  {"Coastal"},          /* char name_orig[MAX_LEN_NAME];             */
  {"t.coastal"},        /* char graphic_str[MAX_LEN_NAME];           */
  {"-"},                /* char graphic_alt[MAX_LEN_NAME];           */
  { {NULL} },           /* struct Sprite *sprite[4];                 */
  0,                    /* int food;                                 */
  0,                    /* int shield;                               */
  1,                    /* int trade;                                */

 {{                     /* special1                                  */
  {"Seaweed"},          /*   char name[MAX_LEN_NAME];                */
  {"Seaweed"},          /*   char name_orig[MAX_LEN_NAME];           */
  {"ts.seaweed"},       /*   char graphic_str[MAX_LEN_NAME];         */
  {"-"},                /*   char graphic_alt[MAX_LEN_NAME];         */
  { {NULL} },           /*   struct Sprite *sprite[4];               */
  1,                    /*   int food;                               */
  0,                    /*   int shield;                             */
  2,                    /*   int trade;                              */
 },
 {                      /* special2                                  */
  {"Seamount"},         /*   char name[MAX_LEN_NAME];                */
  {"Seamount"},         /*   char name_orig[MAX_LEN_NAME];           */
  {"ts.Seamount"},      /*   char graphic_str[MAX_LEN_NAME];         */
  {"-"},                /*   char graphic_alt[MAX_LEN_NAME];         */
  { {NULL} },           /*   struct Sprite *sprite[4];               */
  3,                    /*   int food;                               */
  0,                    /*   int shield;                             */
  1,                    /*   int trade;                              */
 }},
  NULL,                 /* char *helptext;                           */
  &CIVTERRAIN_DEFAULT,  /* CIVTERRAIN *global;                       */

  1,                    /* int movement_cost;                        */
  10,                   /* int defense_bonus;                        */

  T_UNUSED0,            /* enum terrain_type base;                   */
  0,                    /* int road_trade_incr;                      */
  0,                    /* int road_time;                            */

  T_LAST,               /* enum terrain_type irrigation_result;      */
  0,                    /* int irrigation_food_incr;                 */
  0,                    /* int irrigation_time;                      */

  T_LAST,               /* enum terrain_type mining_result;          */
  0,                    /* int mining_shield_incr;                   */
  0,                    /* int mining_time;                          */

  T_LAST,               /* enum terrain_type transform_result;       */
  0,                    /* int transform_pollution_incr;             */
  24,                   /* int transform_time;                       */
 },
 {{"Unknown"},          /* char name[MAX_LEN_NAME];                  */
  {"Unknown"},          /* char name_orig[MAX_LEN_NAME];             */
  {"t.unknown"},        /* char graphic_str[MAX_LEN_NAME];           */
  {"-"},                /* char graphic_alt[MAX_LEN_NAME];           */
  { {NULL} },           /* struct Sprite *sprite[4];                 */
  0,                    /* int food;                                 */
  0,                    /* int shield;                               */
  0,                    /* int trade;                                */

 {{                     /* special1                                  */
  {"Resources"},        /*   char name[MAX_LEN_NAME];                */
  {"Resources"},        /*   char name_orig[MAX_LEN_NAME];           */
  {"ts.grassland_resources"},  /*   char graphic_str[MAX_LEN_NAME];  */
  {"-"},                /*   char graphic_alt[MAX_LEN_NAME];         */
  { {NULL} },           /*   struct Sprite *sprite[4];               */
  0,                    /*   int food;                               */
  0,                    /*   int shield;                             */
  0,                    /*   int trade;                              */
 },
 {                      /* special2                                  */
  {"Resources"},        /*   char name[MAX_LEN_NAME];                */
  {"Resources"},        /*   char name_orig[MAX_LEN_NAME];           */
  {"ts.grassland_resources"},  /*   char graphic_str[MAX_LEN_NAME];  */
  {"-"},                /*   char graphic_alt[MAX_LEN_NAME];         */
  { {NULL} },           /*   struct Sprite *sprite[4];               */
  0,                    /*   int food;                               */
  0,                    /*   int shield;                             */
  0,                    /*   int trade;                              */
 }},
  NULL,                 /* char *helptext;                           */
  &CIVTERRAIN_DEFAULT,  /* CIVTERRAIN *global;                       */

  255,                  /* int movement_cost;                        */
  10,                   /* int defense_bonus;                        */

  T_UNUSED1,            /* enum terrain_type base;                   */
  0,                    /* int road_trade_incr;                      */
  0,                    /* int road_time;                            */

  T_LAST,               /* enum terrain_type irrigation_result;      */
  0,                    /* int irrigation_food_incr;                 */
  0,                    /* int irrigation_time;                      */

  T_LAST,               /* enum terrain_type mining_result;          */
  0,                    /* int mining_shield_incr;                   */
  0,                    /* int mining_time;                          */

  T_LAST,               /* enum terrain_type transform_result;       */
  0,                    /* int transform_pollution_incr;             */
  24,                   /* int transform_time;                       */
 },
 {{"Unknown"},          /* char name[MAX_LEN_NAME];                  */
  {"Unknown"},          /* char name_orig[MAX_LEN_NAME];             */
  {"t.unknown"},        /* char graphic_str[MAX_LEN_NAME];           */
  {"-"},                /* char graphic_alt[MAX_LEN_NAME];           */
  { {NULL} },           /* struct Sprite *sprite[4];                 */
  0,                    /* int food;                                 */
  0,                    /* int shield;                               */
  0,                    /* int trade;                                */

 {{                     /* special1                                  */
  {"Resources"},        /*   char name[MAX_LEN_NAME];                */
  {"Resources"},        /*   char name_orig[MAX_LEN_NAME];           */
  {"ts.grassland_resources"},  /*   char graphic_str[MAX_LEN_NAME];  */
  {"-"},                /*   char graphic_alt[MAX_LEN_NAME];         */
  { {NULL} },           /*   struct Sprite *sprite[4];               */
  0,                    /*   int food;                               */
  0,                    /*   int shield;                             */
  0,                    /*   int trade;                              */
 },
 {                      /* special2                                  */
  {"Resources"},        /*   char name[MAX_LEN_NAME];                */
  {"Resources"},        /*   char name_orig[MAX_LEN_NAME];           */
  {"ts.grassland_resources"},  /*   char graphic_str[MAX_LEN_NAME];  */
  {"-"},                /*   char graphic_alt[MAX_LEN_NAME];         */
  { {NULL} },           /*   struct Sprite *sprite[4];               */
  0,                    /*   int food;                               */
  0,                    /*   int shield;                             */
  0,                    /*   int trade;                              */
 }},
  NULL,                 /* char *helptext;                           */
  &CIVTERRAIN_DEFAULT,  /* CIVTERRAIN *global;                       */

  255,                  /* int movement_cost;                        */
  10,                   /* int defense_bonus;                        */

  T_UNUSED2,            /* enum terrain_type base;                   */
  0,                    /* int road_trade_incr;                      */
  0,                    /* int road_time;                            */

  T_LAST,               /* enum terrain_type irrigation_result;      */
  0,                    /* int irrigation_food_incr;                 */
  0,                    /* int irrigation_time;                      */

  T_LAST,               /* enum terrain_type mining_result;          */
  0,                    /* int mining_shield_incr;                   */
  0,                    /* int mining_time;                          */

  T_LAST,               /* enum terrain_type transform_result;       */
  0,                    /* int transform_pollution_incr;             */
  24,                   /* int transform_time;                       */
 },
};
#endif	/* FC__CIVTERRAIN_C */


/*
 * Miscellaneous terrain information (struct terrain_type from map.h)
 * Effectively, the header or generic elements for terrain_types.
 */
struct civterrain
{
  /* options */
  enum terrain_river_type river_style;
  bool may_road;       /* may build roads/railroads       */
  bool may_irrigate;   /* may build irrigation/farmland   */
  bool may_mine;       /* may build mines                 */
  bool may_transform;  /* may transform terrain           */

  /* parameters */
  int ocean_reclaim_requirement;       /* # adjacent land tiles for reclaim  */
  int land_channel_requirement;        /* # adjacent ocean tiles for channel */

  enum special_river_move river_move_mode;
  int river_defense_bonus;             /* % added to defense if Civ2 river   */
  int river_trade_incr;                /* added to trade if Civ2 river       */
  char *river_helptext;                /* help for Civ2-style rivers         */

  int fortress_defense_bonus;          /* % added to defense if fortress     */
  int farmland_supermarket_food_bonus; /* % added to food if farm/s-market   */
  int road_superhighway_trade_bonus;   /* % added to trade if road/s-highway */
  int rail_food_bonus;                 /* % added to food if railroad        */
  int rail_shield_bonus;               /* % added to shield if railroad      */
  int rail_trade_bonus;                /* % added to trade if railroad       */
  int pollution_food_penalty;          /* % subtr. from food if polluted     */
  int pollution_shield_penalty;        /* % subtr. from shield if polluted   */
  int pollution_trade_penalty;         /* % subtr. from trade if polluted    */
  int fallout_food_penalty;            /* % subtr. from food if fallout      */
  int fallout_shield_penalty;          /* % subtr. from shield if fallout    */
  int fallout_trade_penalty;           /* % subtr. from trade if fallout     */

  const CIVTERRAIN_SPECIAL *stypes;    /* current special_types definitions  */
  const CIVTERRAIN_TYPE *types;        /* current terrain_types definitions  */
};


/* Hardcoded defaults for terrain features */
#ifdef FC__CIVTERRAIN_C
const CIVTERRAIN CIVTERRAIN_DEFAULT = \
{
                    /* options */
  R_AS_SPECIAL,     /* enum terrain_river_type river_style;     */
  TRUE,             /* bool may_road;                           */
  TRUE,             /* bool may_irrigate;                       */
  TRUE,             /* bool may_mine;                           */
  TRUE,             /* bool may_transform;                      */

                    /* parameters */
  4,                /* int ocean_reclaim_requirement;           */
  1,                /* int land_channel_requirement;            */

  RMV_FAST_STRICT,  /* enum special_river_move river_move_mode; */
  50,               /* int river_defense_bonus;                 */
  1,                /* int river_trade_incr;                    */
  NULL,             /* char *river_helptext;                    */

  100,              /* int fortress_defense_bonus;              */
  50,               /* int farmland_supermarket_food_bonus;     */
  0,                /* int road_superhighway_trade_bonus;       */
  50,               /* int rail_food_bonus;                     */
  50,               /* int rail_shield_bonus;                   */
  0,                /* int rail_trade_bonus;                    */
  50,               /* int pollution_food_penalty;              */
  0,                /* int pollution_shield_penalty;            */
  50,               /* int pollution_trade_penalty;             */
  0,                /* int fallout_food_penalty;                */
  50,               /* int fallout_shield_penalty;              */
  50,               /* int fallout_trade_penalty;               */
  &CIVTERRAIN_SPECIAL_DEFAULT[0],
                    /* current special_types definitions        */
  &CIVTERRAIN_TYPE_DEFAULT[0],
                    /* current terrain_types definitions        */
};
#endif	/* FC__CIVTERRAIN_C */


/* External definitions */

/* Global singleton initialized from the terrain ruleset (from common/map.c) */
#ifdef FC__CIVTERRAIN_C
CIVTERRAIN      *civterrain_control;
CIVTERRAIN_TYPE *civterrain_types[T_COUNT];
#endif

extern CIVTERRAIN       *civterrain_control;
extern CIVTERRAIN_TYPE  *civterrain_types[T_COUNT];

void civterrain_type_free(CIVTERRAIN_TYPE *ptype);
void civterrain_types_free(CIVTERRAIN_TYPE **ptype, int count);
CIVTERRAIN_TYPE *civterrain_types_init(CIVTERRAIN_TYPE *ptype, int count,
  CIVTERRAIN *pter);

void civterrain_free(CIVTERRAIN **pter);
CIVTERRAIN *civterrain_init(CIVTERRAIN *pter);

const CIVTERRAIN_TYPE *civterrain_get_type(const CIVTERRAIN *pter, 
  enum terrain_type type);
const char *civterrain_get_name(const CIVTERRAIN *pter, 
  enum terrain_type type);
enum terrain_type civterrain_get_type_by_name(const CIVTERRAIN *pter, 
  const char *name);
enum special_type civterrain_get_special_by_name(const CIVTERRAIN *pter,
  const char *name);
const char *civterrain_get_special_name(const CIVTERRAIN *pter, 
  enum special_type spe);
int  civterrain_get_special_prerequisite(const CIVTERRAIN *pter, int spe);

int  civterrain_type_get_food(const CIVTERRAIN_TYPE *ptype, 
  enum special_type spe);
int  civterrain_type_get_food_base(const CIVTERRAIN_TYPE *ptype,
  enum special_type spe);
int  civterrain_type_get_shield(const CIVTERRAIN_TYPE *ptype, 
  enum special_type spe);
int  civterrain_type_get_shield_base(const CIVTERRAIN_TYPE *ptype,
  enum special_type spe);
int  civterrain_type_get_trade(const CIVTERRAIN_TYPE *ptype, 
  enum special_type spe);
int  civterrain_type_get_trade_base(const CIVTERRAIN_TYPE *ptype,
  enum special_type spe);
const char *civterrain_type_get_infrastructure_text(const CIVTERRAIN_TYPE *pter,
  enum special_type spe);
const char *civterrain_type_get_info_text(const CIVTERRAIN_TYPE *pter,
  enum special_type spe);

int civterrain_type_road_time(const CIVTERRAIN_TYPE *pter);
int civterrain_type_railroad_time(const CIVTERRAIN_TYPE *pter);
int civterrain_type_irrigation_time(const CIVTERRAIN_TYPE *pter);
int civterrain_type_mine_time(const CIVTERRAIN_TYPE *pter);
int civterrain_type_transform_time(const CIVTERRAIN_TYPE *pter);
int civterrain_type_fortress_time(const CIVTERRAIN_TYPE *pter);
int civterrain_type_airbase_time(const CIVTERRAIN_TYPE *pter);
int civterrain_type_pollution_time(const CIVTERRAIN_TYPE *pter);
int civterrain_type_fallout_time(const CIVTERRAIN_TYPE *pter);
#ifdef	FC__UNIT_H
int  civterrain_type_activity_time(const CIVTERRAIN_TYPE *pter,
  enum unit_activity activity);
#endif	/* FC__UNIT_H */

int  get_preferred_pillage(int pset);


/* These are used extensively, so are typically optimized to macros. */
#if !defined(FC__NO_CIVTERRAIN_MACROS) && !defined(FC__CIVTERRAIN_C)

#define civterrain_get_types(pter)      ( pter->types )
#define civterrain_get_type(pter, type) ( pter->types + type )

#endif	/* NO_CIVTERRAIN_MACROS */

#endif  /* FC__CIVTERRAIN_H */

