/********************************************************************** 
 civtile.c - Copyright (C) 2002 - R. Wetmore

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
***********************************************************************/

#define FC__CIVTILE_C

/**
 * civtile.c - core civtile operations
 *
 * A civtile is an element which describes an individual map position.
 * Methods in this file manage the core individual tile operations in
 * isolation from all other tiles.
 *
 * Additional detailed terrain operations are handled by civtile extended
 * methods (civtile_ex.c).
 *
 * A map holds a collection of civtiles organized by topology sharing 
 * various common global characteristics (e.g. basic terrain types) and 
 * relationships (e.g. connectivity or adjacency rules). Global tile
 * operations are handled by civmap extended (civmap_ex.c) methods.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "log.h"
#include "capability.h"
#include "registry.h"
#include "civterrain.h"

#include "civmap_main.h"
#include "civmap.h"
#include "civmap_ex.h"
#include "civmapgen.h"

#include "civtile.h"


/* From server/savegame.c */

/*
 * This loops over the entire map to load data. It inputs a line of
 * data using the macro SECFILE_LOOKUP_LINE and then loops using the
 * macro SET_XY_CHAR to load each char into tha map at (x,y). Internal 
 * variables x and y, xn and yn, and ch are allocated within the macro
 * but definable by the user (at the level where first used). User code
 * thus depends only on the macro interface and not macro internals.
 * 
 * ch:     a variable to hold a char
 * x, y:   variables for internal map coordinates
 * xn, yn: variables for output/native coordinates
 * map:    the map instance or context to update
 * SET_XY_CHAR: macro to load the map character at each (x, y).
 * SECFILE_LOOKUP_LINE: macro to input a line (*, yn) for processing.
 *
 * Note: some (but not all) of the code this is replacing used to skip
 * over lines that did not exist for backward-compatibility. We could 
 * add another parameter that specified whether it was OK to skip data,
 * but there's not really much advantage to exiting early in this case.
 * Instead, we let any map data type be empty, and just print an
 * informative warning message about it.
*/
#define LOAD_MAP_DATA(ch, xn, yn, x, y, map,                           \
                      SET_XY_CHAR, SECFILE_LOOKUP_LINE)                \
{                                                                      \
  char ch, *_line;                                                     \
  int x, y, xn, yn;                                                    \
  for (yn = 0; yn < map->ysize; yn++) {                                \
    _line = (SECFILE_LOOKUP_LINE);                                     \
    if (!_line || strlen(_line) != map->xsize) {                       \
      freelog(LOG_ERROR, "Invalid map data at line y=%d.  "            \
        "This can happen with old saved games, or it may "             \
        "indicate an invalid saved game file.  "                       \
        "Proceed at your own risk.", yn);                              \
      if(!_line) {                                                     \
        freelog(LOG_ERROR, _("Reason: line not found"));               \
      } else {                                                         \
        freelog(LOG_ERROR, _("Reason: line too short "                 \
                "(expected %d got %lu"), map->xsize,                   \
                (unsigned long) strlen(_line));                        \
      }                                                                \
      continue;                                                        \
    }                                                                  \
    for(xn = 0; xn < map->xsize; xn++) {                               \
      ch = _line[xn];                                                  \
      native_to_civmap_pos(map, &x, &y, xn, yn);                       \
      if (normalize_civmap_pos(map, &x, &y)) {                         \
        (SET_XY_CHAR);                                                 \
      } else {                                                         \
        /* assert(ch == '#'); */                                       \
      }                                                                \
    }                                                                  \
  }                                                                    \
}

/* 
 * This loops over the entire map to save data. It collects all the
 * data for a line using the macro GET_XY_CHAR and then executes the
 * macro SECFILE_INSERT_LINE. Internal variables x and y, xn and yn,
 * and line are allocated within the macro but definable by the user
 * (at the level where first used in the macros). User code thus
 * depends only on the macro interface and not macro internals.
 * 
 * line:   buffer variable to hold a line of chars
 * xn, yn: variables for output/native coordinates
 * x, y:   variables for internal map coordinates
 * map:    the map instance or context to save
 * GET_XY_CHAR: macro returning the map character for each (x, y).
 * SECFILE_INSERT_LINE: macro to output each processed line (*, yn).
 */

#define SAVE_MAP_DATA(line, xn, yn, x, y, map,                         \
                      GET_XY_CHAR, SECFILE_INSERT_LINE)                \
{                                                                      \
  char line[map->xsize + 1];                                           \
  int x, y, xn, yn;                                                    \
                                                                       \
  for (yn = 0; yn < map->ysize; yn++) {                                \
    for (xn = 0; xn < map->xsize; xn++) {                              \
      native_to_civmap_pos(map, &x, &y, xn, yn);                       \
      if (normalize_civmap_pos(map, &x, &y)) {                         \
        line[xn] = (GET_XY_CHAR);                                      \
        if(!isprint(line[xn] & 0x7f)) {                                \
          freelog(LOG_ERROR, "Trying to write invalid"                 \
                  " map data: '%c' %d", line[xn], line[xn]);           \
          assert(FALSE); /* exit(EXIT_FAILURE); ist fur das Gestapo */ \
        }                                                              \
      } else {                                                         \
        /* skipped over in loading */                                  \
        line[xn] = '#';                                                \
      }                                                                \
    }                                                                  \
    line[map->xsize] = '\0';                                           \
    (SECFILE_INSERT_LINE);                                             \
  }                                                                    \
}

#define SAVE_NORMAL_MAP_DATA(x, y, map,                                \
          GET_XY_CHAR, secfile, secname)                               \
        SAVE_MAP_DATA(line, xn, yn, x, y, map,                         \
          (GET_XY_CHAR), secfile_insert_str(secfile, line, secname, yn))

#define SAVE_PLAYER_MAP_DATA(x, y, map,                                \
          GET_XY_CHAR, secfile, secname, plrno)                        \
        SAVE_MAP_DATA(line, xn, yn, x, y, map,                         \
          (GET_XY_CHAR), secfile_insert_str(secfile, line, secname, plrno, yn))


/***********************************************************************
This returns an ascii hex value of the given half-byte of the binary
integer. See ascii_hex2bin().
  example: bin2ascii_hex(0xa00, 2) == 'a'
***********************************************************************/
#define bin2ascii_hex(value, halfbyte_wanted) \
  hex_chars[((value) >> ((halfbyte_wanted) * 4)) & 0xf]

static const char hex_chars[] = "0123456789abcdef";

/***********************************************************************
This returns a binary integer value of the ascii hex char, offset by
the given number of half-bytes. See bin2ascii_hex().
  example: ascii_hex2bin('a', 2) == 0xa00
This is only used in loading games, and it requires some error
checking so it's done as a function.
***********************************************************************/
static int ascii_hex2bin(char ch, int halfbyte)
{
  char *pch;

  if (ch == ' ') {
    /* 
     * Sane value. It is unknow if there are savegames out there which
     * need this fix. Savegame.c doesn't write such savegames
     * (anymore) since the inclusion into CVS (2000-08-25).
     */
    return 0;
  }
  
  pch = strchr(hex_chars, ch);

  if (!pch || ch == '\0') {
    freelog(LOG_FATAL, "Unknown hex value: '%c' %d", ch, ch);
    exit(EXIT_FAILURE);
  }
  return (pch - hex_chars) << (halfbyte * 4);
}

/***************************************************************
Dereferences the terrain character.  See terrain_chars[].
  example: char2terrain('a') == 0
***************************************************************/
static int char2terrain(char ch)
{
  char *pch = strchr(terrain_chars, ch);

  if (!pch || ch == '\0') {
    freelog(LOG_FATAL, "Unknown terrain type: '%c' %d", ch, ch);
    exit(EXIT_FAILURE);
  }
  return pch - terrain_chars;
}

/**
 * Load a backing store map tiles section into the civmap structure 
 *
 * @return status 0-success, nn-error code
 *
 * Note: loosely synced with server/savegame.c
 */
int civtile_load(CIVMAP *map, struct section_file *fp)
{
  int  game_version;
  char *savefile_options;
  enum server_states server_state;

  /* Key control options */
  game_version = secfile_lookup_int_default(fp, 0, "game.version");
  savefile_options = secfile_lookup_str_default(fp, "", "savefile.options");
  server_state = (enum server_states)
    secfile_lookup_int_default(fp, PRE_GAME_STATE, "game.server_state");

  /* TODO: should we always run civmap_load() here??? */
  /* (Re)initialize map->tiles */
  civmap_tiles_init(map);

  /* get the terrain type */
  LOAD_MAP_DATA(ch, xn, yn, x, y, map, 
    civmap_get_tile(map, x, y)->terrain = char2terrain(ch),
    secfile_lookup_str(fp, "map.t%03d", yn));

  if (has_capability("specials", savefile_options)) {

    /* get 4-bit segments of 16-bit "special" field. */
    LOAD_MAP_DATA(ch, xn, yn, x, y, map, 
      civmap_get_tile(map, x, y)->special = ascii_hex2bin(ch, 0),
      secfile_lookup_str(fp, "map.l%03d", yn));
    LOAD_MAP_DATA(ch, xn, yn, x, y, map, 
      civmap_get_tile(map, x, y)->special |= ascii_hex2bin(ch, 1),
      secfile_lookup_str(fp, "map.u%03d", yn));
    LOAD_MAP_DATA(ch, xn, yn, x, y, map, 
      civmap_get_tile(map, x, y)->special |= ascii_hex2bin(ch, 2),
      secfile_lookup_str(fp, "map.n%03d", yn));
    LOAD_MAP_DATA(ch, xn, yn, x, y, map, 
      civmap_get_tile(map, x, y)->special |= ascii_hex2bin(ch, 3),
      secfile_lookup_str(fp, "map.f%03d", yn));

    map->has_specials = TRUE;
  } else
  if (has_capability("riversoverlay", savefile_options)) {

    /* Get the bits of the special flags which contain the river special
       and extract the rivers overlay from them. */
    LOAD_MAP_DATA(ch, xn, yn, x, y, map, 
      civmap_get_tile(map, x, y)->special |= (ascii_hex2bin(ch, 2) & S_RIVER),
      secfile_lookup_str_default(fp, NULL, "map.n%03d", yn));

    map->has_rivers_overlay = TRUE;
  }

  /* secfile_lookup_bool_default(fp, TRUE, "game.save_known") */
  if (FALSE) {
    LOAD_MAP_DATA(ch, xn, yn, x, y, map, 
      civmap_get_tile(map, x, y)->known |= ascii_hex2bin(ch, 0),
      secfile_lookup_str(fp, "map.a%03d", yn));
    LOAD_MAP_DATA(ch, xn, yn, x, y, map, 
      civmap_get_tile(map, x, y)->known |= ascii_hex2bin(ch, 1),
      secfile_lookup_str(fp, "map.b%03d", yn));
    LOAD_MAP_DATA(ch, xn, yn, x, y, map, 
      civmap_get_tile(map, x, y)->known |= ascii_hex2bin(ch, 2),
      secfile_lookup_str(fp, "map.c%03d", yn));
    LOAD_MAP_DATA(ch, xn, yn, x, y, map, 
      civmap_get_tile(map, x, y)->known |= ascii_hex2bin(ch, 3),
      secfile_lookup_str(fp, "map.d%03d", yn));

    if (has_capability("known32fix", savefile_options)) {
      LOAD_MAP_DATA(ch, xn, yn, x, y, map, 
        civmap_get_tile(map, x, y)->known |= ascii_hex2bin(ch, 4),
        secfile_lookup_str(fp, "map.e%03d", yn));
      LOAD_MAP_DATA(ch, xn, yn, x, y, map, 
        civmap_get_tile(map, x, y)->known |= ascii_hex2bin(ch, 5),
        secfile_lookup_str(fp, "map.g%03d", yn));
      LOAD_MAP_DATA(ch, xn, yn, x, y, map, 
        civmap_get_tile(map, x, y)->known |= ascii_hex2bin(ch, 6),
        secfile_lookup_str(fp, "map.h%03d", yn));
      LOAD_MAP_DATA(ch, xn, yn, x, y, map, 
        civmap_get_tile(map, x, y)->known |= ascii_hex2bin(ch, 7),
        secfile_lookup_str(fp, "map.i%03d", yn));
    }
  }

  /* TODO: Force algorithms to generate consistent mappings or save.
   * assign_continent_numbers();
   * assign_ocean_numbers();
  */
  return 0;
}

/**
 * Dump the civmap working tiles into a backing store map section 
 *
 * @return status 0-success, nn-error code
 *
 * Note: loosely synced with server/savegame.c
 */
int civtile_save(const CIVMAP *map, struct section_file *fp)
{
  /* put the terrain type */
  SAVE_NORMAL_MAP_DATA(x, y, map,
    terrain_chars[(civmap_get_tile(map, x, y)->terrain & 15)], fp, "map.t%03d");

  /* These must agree with save_options capabilities. */
  if (map->has_specials) {
    /*
     * put 4-bit segments of 12-bit "special flags" field
     */
    SAVE_NORMAL_MAP_DATA(x, y, map,
      bin2ascii_hex(civmap_get_tile(map, x, y)->special, 0), fp, "map.l%03d");
    SAVE_NORMAL_MAP_DATA(x, y, map,
      bin2ascii_hex(civmap_get_tile(map, x, y)->special, 1), fp, "map.u%03d");
    SAVE_NORMAL_MAP_DATA(x, y, map,
      bin2ascii_hex(civmap_get_tile(map, x, y)->special, 2), fp, "map.n%03d");
    SAVE_NORMAL_MAP_DATA(x, y, map,
      bin2ascii_hex(civmap_get_tile(map, x, y)->special, 3), fp, "map.f%03d");
  } else
  if (map->has_rivers_overlay) {
    /* 
     * Save the rivers overlay map; this is a special case to allow
     * re-saving scenarios which have rivers overlay data. 
     * This only applies if don't have the rest of the specials.
     * 
     * bits 8-11 of special flags field
     */
    SAVE_NORMAL_MAP_DATA(x, y, map,
      bin2ascii_hex(civmap_get_tile(map, x, y)->special, 2), fp, "map.n%03d");
  }

  /* game.save_known */
  secfile_insert_bool(fp, TRUE, "game.save_known");
  if (TRUE) {
    /* put 4-bit segments of the 32-bit "known" field */
    SAVE_NORMAL_MAP_DATA(x, y, map,
      bin2ascii_hex(civmap_get_tile(map, x, y)->known, 0), fp, "map.a%03d");
    SAVE_NORMAL_MAP_DATA(x, y, map,
      bin2ascii_hex(civmap_get_tile(map, x, y)->known, 1), fp, "map.b%03d");
    SAVE_NORMAL_MAP_DATA(x, y, map,
      bin2ascii_hex(civmap_get_tile(map, x, y)->known, 2), fp, "map.c%03d");
    SAVE_NORMAL_MAP_DATA(x, y, map,
      bin2ascii_hex(civmap_get_tile(map, x, y)->known, 3), fp, "map.d%03d");
    SAVE_NORMAL_MAP_DATA(x, y, map,
      bin2ascii_hex(civmap_get_tile(map, x, y)->known, 4), fp, "map.e%03d");
    SAVE_NORMAL_MAP_DATA(x, y, map,
      bin2ascii_hex(civmap_get_tile(map, x, y)->known, 5), fp, "map.g%03d");
    SAVE_NORMAL_MAP_DATA(x, y, map,
      bin2ascii_hex(civmap_get_tile(map, x, y)->known, 6), fp, "map.h%03d");
    SAVE_NORMAL_MAP_DATA(x, y, map,
      bin2ascii_hex(civmap_get_tile(map, x, y)->known, 7), fp, "map.i%03d");
  }

  return 0;
}


/**
 * Remove a vector of working tiles cleaning up all memory references
 */
void civtile_free(CIVTILE **ptile, int nn)
{
  if (ptile && *ptile) {
    // This is not needed for mapgeneration save games.
    //  while (--nn >= 0)
    //    unit_list_unlink_all(&((*ptile)[nn].units));
    free(*ptile);
    *ptile = NULL;
  }
}

/**
 * Initialize or reinitialize a working tile of a civmap in memory
 */
CIVTILE *civtile_init(CIVTILE *ptile, int nn, CIVMAP *pmap)
{
  /* Cleanout any stale tile data - overkill for pure mapgeneration */
  if (NULL != ptile)
    civtile_free(&ptile, nn);

  ptile = fc_malloc(nn * sizeof(CIVTILE_DEFAULT));
  while (--nn >= 0) {
    ptile[nn] = CIVTILE_DEFAULT;
    ptile[nn].global = pmap;
    // This is not needed for mapgeneration save games.
    // unit_list_init(&ptile->units);
  }

  return ptile;
}


/***********************************************************************
  Returns a resource weighted goodness for the tile.
  Range 0 ... 5 , 2 standard.
***********************************************************************/
int civtile_goodness(CIVTILE *ptile)
{
  switch (ptile->terrain) {
  case T_FOREST:
    return (ptile->special == S_NO_SPECIAL) ? 2 : 5;
  case T_GRASSLAND:
  case T_RIVER:
    return (ptile->special == S_NO_SPECIAL) ? 3 : 4;
  case T_PLAINS:
  case T_HILLS:
    return (ptile->special == S_NO_SPECIAL) ? 2 : 4;
  case T_DESERT:
  case T_OCEAN: /* must be called with usable seas */
    return (ptile->special == S_NO_SPECIAL) ? 1 : 3;
  case T_SWAMP:
  case T_JUNGLE:
  case T_MOUNTAINS:
    return (ptile->special == S_NO_SPECIAL) ? 0 : 3;
  /* case T_ARCTIC: */
  /* case T_TUNDRA: */
  default:
    return (ptile->special == S_NO_SPECIAL) ? 0 : 1;
  }
}


/***********************************************************************
  Returns the given special set from the given tile.
***********************************************************************/
enum special_type civtile_get_specials(CIVTILE *ptile)
{
  return ptile->special;
}

/***********************************************************************
  Returns true is the tile has any of the given special set.
***********************************************************************/
bool civtile_has_specials(CIVTILE *ptile, enum special_type spe)
{
  if ((spe & S_RIVER) != 0 && ptile->terrain == T_RIVER)
    return TRUE;
  return (ptile->special & spe) != 0;
}

/***********************************************************************
  Assigns the given special set to the given tile.
***********************************************************************/
void civtile_set_specials(CIVTILE *ptile, enum special_type spe)
{
  if ((spe & S_RIVER) != 0 
   && ptile->global->terrain->river_style == R_AS_TERRAIN) {
    ptile->terrain = T_RIVER;
    spe &= ~S_RIVER;
  }
  ptile->special = spe;
}

/***********************************************************************
  Assign the given special to the given tile.
***********************************************************************/
void civtile_special_add(CIVTILE *ptile, enum special_type spe)
{
  if ((spe & S_RIVER) != 0 
   && ptile->global->terrain->river_style == R_AS_TERRAIN) {
    ptile->terrain = T_RIVER;
    spe &= ~S_RIVER;
  }
  ptile->special |= spe;
}

/***********************************************************************
  Delete the given special from the given tile.
***********************************************************************/
void civtile_special_delete(CIVTILE *ptile, enum special_type spe)
{
  if ((spe & S_RIVER) != 0 
   && ptile->global->terrain->river_style == R_AS_TERRAIN) {
    ptile->terrain = T_GRASSLAND;
    spe &= ~S_RIVER;
  }
  ptile->special &= ~spe;
}

/***********************************************************************
  Deletes all infrastructure specials at the given tile.
***********************************************************************/
void civtile_infrastructure_delete(CIVTILE *ptile)
{
  ptile->special &= ~S_INFRASTRUCTURE_MASK;
}

/***********************************************************************
  Deletes the infrastructure specials at the given tile.
***********************************************************************/
void civtile_dirtiness_delete(CIVTILE *ptile)
{
  ptile->special &= ~(S_POLLUTION | S_FALLOUT);
}


/***********************************************************************
  Returns the given terrain set from the given tile.
***********************************************************************/
enum terrain_type civtile_get_terrain(CIVTILE *ptile)
{
  return ptile->terrain;
}

/***********************************************************************
  Returns true is the tile has any of the given terrain set.
***********************************************************************/
bool civtile_has_terrain(CIVTILE *ptile, enum terrain_type tt)
{
  if (T_RIVER == tt && (ptile->special & S_RIVER) != 0)
    return TRUE;
  return (ptile->terrain == tt);
}

/***********************************************************************
  Assigns the given terrain set to the given tile.
***********************************************************************/
void civtile_set_terrain(CIVTILE *ptile, enum terrain_type tt)
{
  if (T_RIVER == tt
   && ptile->global->terrain->river_style == R_AS_SPECIAL) {
    ptile->special |= S_RIVER;
    tt = T_GRASSLAND;
  }
  ptile->terrain = tt;
}

