/**********************************************************************
 Freeciv - Copyright (C) 1996 - A Kjeldberg, L Gregersen, P Unold
 civtile.h - Copyright (C) 2002 - R. Wetmore

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
***********************************************************************/

#ifndef FC__CIVTILE_H
#define FC__CIVTILE_H

/**
 * Civtile.h - basic Freeciv tile header definitions and methods
 *
 *   This is loosely synced with FREECIV map.h. The hard compatiblity
 * constraint is the ability to load/save each other's savegame files.
 *
 *   Tiles are individual map positions. The tile structure holds data
 * such as graphics elements or features, terrain type (including user
 * modifiable elements as in specials {S_IRRIGATION, S_MINE, S_ROAD})
 * or play elements such as a city or list of units. Tile data contains 
 * dynamic state as opposed to purely static or const data.
 *
 *   Tiles are stored and organized in maps. Thus in normal use the
 * global backreference can be used to access any needed global 
 * properies from the map. Static terrain information is stored in
 * a global terrain structure and indexed by type since there are 
 * typically far fewer terrain types than tiles.
 *
 *   As with CIVMAPs, commonly used accessor functions as for terrain
 * information are found in civtile_ex.h. They wrap terrain functions
 * hiding  the basic indexed lookup step. Thus
 *   civterrain_get_food(CIVTERRAIN_TYPE *ptype, enum special_type spe)
 * becomes
 *   civtile_get_food(CIVTILE *ptile)
 * where expanding the steps gives
 *   civterrain_get_food(civtile_get_terrain(CIVTILE *ptile),
 *     enum special_type ptile->special)
 *   civterrain_get_food((ptile->global->terrain->types + ptile->terrain),
 *     enum special_type ptile->special)
 */

/* required dependencies */
#include <assert.h>

#include "genlist.h"
#include "civterrain.h"
#include "civmap.h"

struct city;
struct unit_list;

/* States used in fog-of-war (FoW) operations. */
enum known_type { TILE_UNKNOWN, TILE_KNOWN_FOGGED, TILE_KNOWN };

/**
 * A tile defines a single map element.
 */
typedef struct tile {
  enum terrain_type terrain;
                        /* one of the 16 basic terrain types    */
  enum special_type special;
                        /* S_RIVER | {S_SPECIAL1-3}             */
  short continent;      /* the continent number for land tiles. */
                        /* ocean tiles == 0                     */
  short ocean;          /* the ocean/watershed for water tiles. */
                        /* for OCEAN, COASTAL/LAKES, RIVERS     */

  /* Runtime summary and working fields.  */
  unsigned int known;   /* An enum known_type client side.      */
                        /* a bitvector server side with         */
                        /* Player_no as index                   */
  unsigned int sent;    /* Not used client side.                */
                        /* A bitvector like known server side.  */
                        /* Indicates if the client knows the    */
                        /* tile as TILE_KNOWN_NODRAW.           */
  int assigned;
  char movecost[8];     /* Cached basic move cost for exitting  */
                        /* the tile in the given diretion.      */

  /* Runtime hooks for attaching objects  */
  CIVMAP *global;       /* back link to the global map defn's   */
  struct city *city;    /* city occupying this tile, or NULL    */
  struct city *worked;  /* city working tile, or NULL if none   */
  struct genlist units; /* list of units currently in this tile */
} CIVTILE;


/* Hardcoded tile defaults */

static const struct tile CIVTILE_DEFAULT =
{
  T_UNKNOWN,    /* enum terrain_type terrain;                   */
  S_NO_SPECIAL, /* enum special_type special;                   */
  0,            /* short continent;                             */
  0,            /* short ocean;                                 */
  TILE_UNKNOWN, /* unsigned int known;                          */
  0,            /* unsigned int sent;                           */
  0,            /* int assigned;                                */
  { 0 },        /* char move_cost[8];                           */
  NULL,         /* CIVMAP *global;                              */
  NULL,         /* struct city *city;                           */
  NULL,         /* struct city *worked;                         */
  { 0 },        /* struct unit_list units;                      */
};


/* External methods */

void civtile_free(CIVTILE **ptile, int nn);
CIVTILE *civtile_init(CIVTILE *ptile, int nn, CIVMAP *pmap);

int civtile_load(CIVMAP *map, struct section_file *fp);
int civtile_save(const CIVMAP *map, struct section_file *fp);

const CIVTERRAIN_TYPE *civtile_get_terrain_type(const CIVTILE *ptile);
const CIVTERRAIN_TYPE *civtile_get_terrain_types(const CIVTILE *ptile);

int  civtile_goodness(CIVTILE *ptile);

enum special_type civtile_get_specials(CIVTILE *ptile);
bool civtile_has_specials(CIVTILE *ptile, enum special_type spe);
void civtile_set_specials(CIVTILE *ptile, enum special_type spe);
void civtile_special_add(CIVTILE *ptile, enum special_type spe);
void civtile_special_delete(CIVTILE *ptile, enum special_type spe);
void civtile_infrastructure_delete(CIVTILE *ptile);
void civtile_dirtiness_delete(CIVTILE *ptile);

enum terrain_type civtile_get_terrain(CIVTILE *ptile);
bool civtile_has_terrain(CIVTILE *ptile, enum terrain_type tt);
void civtile_set_terrain(CIVTILE *ptile, enum terrain_type tt);


/* These are used extensively, so are typically optimized to macros. */
#if !defined(FC__NO_CIVTILE_MACROS) && !defined(FC__CIVTILE_C)

#define civtile_get_terrain_types(ptile) \
( (ptile)->global->terrain->types )
#define civtile_get_terrain_type(ptile)  \
( civtile_get_terrain_types(ptile) + (ptile)->terrain )

#define	civtile_get_specials(ptile)      ( (ptile)->special )
#define	civtile_has_specials(ptile, spe) \
( 0 != (spe & ((ptile)->special | ((ptile)->terrain == T_RIVER ? S_RIVER:0))))

#define	civtile_get_terrain(ptile)       ( (ptile)->terrain )
#define	civtile_has_terrain(ptile, tt)   \
(   ((ptile)->terrain == tt)             \
 || ((ptile)->terrain == T_RIVER && 0 != ((ptile)->special & S_RIVER)) )

#define	civtile_get_continent(ptile)     ( (ptile)->continent )
#define	civtile_set_continent(ptile, nn) ( (ptile)->continent = (nn) )
#define	civtile_get_ocean(ptile)         ( (ptile)->ocean )
#define	civtile_set_ocean(ptile, nn)     ( (ptile)->ocean = (nn) )
#define	civtile_get_known(ptile)         ( (ptile)->known )

#endif	/* FC__NO_CIVTILE_MACROS */

#endif  /* FC__CIVTILE_H */

